// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;

namespace Twig
{
    /// <summary>
    /// Represents whether an action is running and if not how it terminated.
    /// Note: the status of an action that has not yet been started is undefined.
    /// </summary>
    public enum ActionStatus { 
        /// <summary>
        /// The action is in the process of running and has not yet completed.
        /// </summary>
        Running, 
        /// <summary>
        /// The action completed successfully.
        /// </summary>
        Succeed,
        /// <summary>
        /// The action failed and is no longer running.
        /// </summary>
        Fail };

    //public delegate void ActionStatusChangedHandler(ActionStatus newStatus);

    /// <summary>
    /// Represents an instance of a discrete action, i.e. a step in a plan.
    /// </summary>
    public abstract class Action
    {
        /// <summary>
        /// Starts the action
        /// </summary>
        public virtual void Start()
        {
        }
        /// <summary>
        /// Terminates the action
        /// </summary>
        public virtual void Abort()
        {
        }

        /// <summary>
        /// Returns the current state of the action (running, succeed, failed)
        /// </summary>
        public abstract ActionStatus Status { get; }
        /// <summary>
        /// Performs any necessary update for the action (e.g. checking for termination).
        /// Only called if the action is actually running.
        /// </summary>
        public virtual void Update() { }

        /// <summary>
        /// RequestID of the Action, if running in a server.
        /// </summary>
        public int requestID;

        /// <summary>
        /// Procedure to start/end status updates to client
        /// </summary>
        public delegate System.IO.TextWriter MessageControl(Action self, bool start);

        /// <summary>
        /// Called to start/end message.
        /// </summary>
        public MessageControl messageControl;

        /// <summary>
        /// Starts and update message to the RPC client.  Returns a stream you can write to.
        /// IMPORTANT: Don't write a newline to the stream.
        /// IMPORTANT: finish all writing and call EndMessage() before returning from your update routine.
        /// </summary>
        /// <returns>Stream to write the message to.</returns>
        protected System.IO.TextWriter StartMessage()
        {
            return messageControl(this, true);
        }

        /// <summary>
        /// Ends writing of status update message.
        /// </summary>
        protected void EndMessage()
        {
            messageControl(this, false);
        }

        /// <summary>
        /// Produces a simple string as a status message.
        /// </summary>
        protected void Message(string message)
        {
            System.IO.TextWriter t = StartMessage();
            t.Write(message);
            EndMessage();
        }
    }

    /// <summary>
    /// A compound action consisting a sequence of subactions to be executed in order.
    /// A failure of any subaction fails the whole sequence.
    /// </summary>
    public class Sequence : Action
    {
        int currentStep=-1;
        bool failed = false;
        Action[] steps;

        /// <summary>
        /// Creates a new action composed of the specified series of subactions.
        /// </summary>
        /// <param name="steps"></param>
        public Sequence(params Action[] steps)
        {
            this.steps = steps;
        }

        /// <summary>
        /// Starts execution of the sequence beginning with the first action
        /// </summary>
        public override void Start()
        {
            Debug.Assert(currentStep == -1, "Start called on sequence already in progress");
            NextStep();
        }

        void NextStep()
        {
            currentStep++;
            if (currentStep == steps.Length)
            {
                currentStep = -1;
                failed = false;
            } else
                steps[currentStep].Start();
        }

        /// <summary>
        /// Checks for completion/failure of the current action and responds accordingly.
        /// </summary>
        public override void Update()
        {
            if (currentStep >= 0)
            {
                // We're running
                steps[currentStep].Update();
                switch (steps[currentStep].Status)
                {
                    case ActionStatus.Running:
                        break;

                    case ActionStatus.Succeed:
                        NextStep();
                        break;

                    case ActionStatus.Fail:
                        Abort();
                        failed = true;
                        break;

                    default:
                        Debug.Assert(false, "Action returned invalid state");
                        break;
                }
            }
        }

        /// <summary>
        /// Aborts the current action and terminates execution of the sequence.
        /// </summary>
        public override void Abort()
        {
            currentStep = -1;
        }

        /// <summary>
        /// Current status of the sequence's execution.
        /// </summary>
        public override ActionStatus Status
        {
            get
            {
                if (currentStep >= 0)
                    return ActionStatus.Running;
                else if (failed)
                    return ActionStatus.Fail;
                else
                    return ActionStatus.Succeed;
            }
        }
    }

    /// <summary>
    /// Represents a pause in an action sequence.
    /// </summary>
    public class Sleep : Action
    {
        int wakeUpTick = -1;
        int duration;

        /// <summary>
        /// Represents a pause in an action sequence.
        /// </summary>
        /// <param name="seconds"></param>
        public Sleep(float seconds)
        {
            duration = Physics.SecondsToTicks(seconds);
        }

        /// <summary>
        /// Starts the action
        /// </summary>
        public override void Start()
        {
            wakeUpTick = Physics.tickCount+duration;
        }

        /// <summary>
        /// Aborts the action
        /// </summary>
        public override void Abort()
        {
            wakeUpTick = -1;
        }

        /// <summary>
        /// Running if still waiting, otherwise Succeeded
        /// </summary>
        public override ActionStatus Status
        {
            get { return (wakeUpTick<Physics.tickCount)?ActionStatus.Succeed:ActionStatus.Running; }
        }
    }

    /// <summary>
    /// An action that asserts a predicate should be achieved on a given object and terminates when it's achieved.
    /// </summary>
    public class GoalAchievementAction : Action
    {
        Predicate predicate;
        float strength;
        TwigObject target;

        /// <summary>
        /// Creates an action that, when run asserts that the achievement of the predicate on the object
        /// has the specified strength.  Terminates when the goal is achieved.
        /// </summary>
        public GoalAchievementAction(Predicate p, TwigObject o, float strength)
        {
            predicate = p;
            target = o;
            this.strength = strength;
        }

        /// <summary>
        /// Creates an action that, when run, asserts that the achievement of the predicate on the object
        /// is a goal of unit strength.
        /// </summary>
        public GoalAchievementAction(Predicate p, TwigObject o)
            : this(p, o, 1f)
        { }

        /// <summary>
        /// Asserts the goal
        /// </summary>
        public override void Update()
        {
            predicate.Goal.Assert(target, strength);
        }

        /// <summary>
        /// Returns success if the predicate is achieved, else running.
        /// </summary>
        public override ActionStatus Status
        {
            get
            {
                if (predicate[target])
                    return ActionStatus.Succeed;
                else
                    return ActionStatus.Running;
            }
        }
    }

    /// <summary>
    /// And action that always fails.
    /// </summary>
    public class FailAction : Action
    {
        /// <summary>
        /// Always returns fail.
        /// </summary>
        public override ActionStatus Status
        {
            get { return ActionStatus.Fail; }
        }
    }
}