// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;

#pragma warning disable 1591

namespace Twig
{
    partial class Mammal
    {
        // Gratch and Marsella appraise situations rather than objects.
        // Their dimensions are desirability, likelihood, urgency, expectedness, causal attribution,
        // controllability, and changeability (last two sometimes grouped as coping potential)

        /// <summary>
        /// Information about the interest of an object relative to the creature.
        /// </summary>
        public struct ObjectAppraisal
        {
            public ObjectAppraisal(float valence, float monitoringPriority) {
                this.valence = valence;
                this.monitoringPriority = monitoringPriority;
            }
            public ObjectAppraisal(float valence) : this(valence, Math.Abs(valence)) { }
            /// <summary>
            /// The valence of the object.  Positive is good, negative bad, 0 neutral.
            /// </summary>
            public float valence;
            /// <summary>
            /// The importance of keeping track of this object.
            /// </summary>
            public float monitoringPriority;
            /// <summary>
            /// The relative interest level of this object.
            /// </summary>
            public float Salience {
                get {
                    return Math.Abs(valence);
                }
            }
        }

        /// <summary>
        /// Table of appraisals of specific objects.
        /// This overrides the GenericAppraisal() process.
        /// </summary>
        public Dictionary<TwigObject,ObjectAppraisal> specificObjectAppraisal = new Dictionary<TwigObject,ObjectAppraisal>();

        class TypeAppraisal
        {
            public Type type;
            public ObjectAppraisal appraisal;
            public TypeAppraisal(Type type, ObjectAppraisal appraisal)
            {
                this.type = type;
                this.appraisal = appraisal;
            }
        }

        List<TypeAppraisal> typeAppraisals = new List<TypeAppraisal>();

        /// <summary>
        /// Find this character's appraisal of this specific type, ignoring type inheritance
        /// </summary>
        TypeAppraisal SpecificTypeAppraisalOfType(Type t)
        {
            for (int i = 0; i < typeAppraisals.Count; i++)
                if (typeAppraisals[i].type == t)
                    return typeAppraisals[i];
            return null;
        }

        /// <summary>
        /// Return this character's appraisal of objects of this type,
        /// taking into account type inheritance.
        /// </summary>
        TypeAppraisal TypeAppraisalOfType(Type t)
        {
            while (t != null)
            {
                TypeAppraisal a = SpecificTypeAppraisalOfType(t);
                if (a != null)
                    return a;
                else t = t.BaseType;
            }
            return null;
        }

        public void DeclareTypeAppraisal(Type t, ObjectAppraisal appraisal)
        {
            TypeAppraisal a = SpecificTypeAppraisalOfType(t);
            if (a == null)
                typeAppraisals.Add(new TypeAppraisal(t, appraisal));
            else
                a.appraisal = appraisal;
        }

        /// <summary>
        /// Computes appraisal of specified object
        /// </summary>
        public virtual ObjectAppraisal GenericAppraisal(TwigObject o)
        {
            if (IsCareGiver(o))
                return new ObjectAppraisal(attachmentBehavior.Activation*((o==primaryCareGiver)?2f:1f), 1+Anxiety);

            TypeAppraisal a = TypeAppraisalOfType(o.GetType());
            if (a == null)
                return new ObjectAppraisal(0f);
            else
                return a.appraisal;
        }

        /// <summary>
        /// Appraises object.  Uses caching to speed up appraisal process.
        /// </summary>
        /// <param name="o">object to appraise</param>
        /// <returns>appraisal</returns>
        public ObjectAppraisal Appraisal(TwigObject o)
        {
            if (o == null)
                return new ObjectAppraisal(0f);
            else
                return Appraise(o);
        }

        /// <summary>
        /// Forcibly computes appraisal of object without caching.
        /// </summary>
        /// <param name="o"></param>
        /// <returns></returns>
        ObjectAppraisal Appraise(TwigObject o)
        {
            ObjectAppraisal appraisal;
            if (specificObjectAppraisal.ContainsKey(o))
                appraisal = specificObjectAppraisal[o];
            else
                appraisal = GenericAppraisal(o);

            if (appraisal.valence > 0)
                appraisal.valence *= (1+Security)/2f;
            else
                appraisal.valence *= (1+Anxiety)/2f;

            return appraisal;
        }
    }
}
