// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;

namespace Twig
{
    partial class Mammal
    {
        /// <summary>
        /// True when potential field navigation is stuck in a local minimum.
        /// </summary>
        public bool blocked;
        /// <summary>
        /// True if there's a nearby obstacle.
        /// </summary>
        public bool crowded;

        /// <summary>
        /// Locomotion behavior that generates walk vectors so as to approach a specified object
        /// while avoiding obstacles.
        /// </summary>
        public MaxBehavior<ApproachParameters> approachController;

        /// <summary>
        /// Locomotion behavior that generates walk vectors so as to approach a specified object
        /// while avoiding obstacles.
        /// </summary>
        public class ApproachParameters
        {
            /// <summary>
            /// Object to approach
            /// </summary>
            public TwigObject target;
            /// <summary>
            /// Desired distance from target.
            /// </summary>
            public float approachDistance;
            /// <summary>
            /// Desired offset vector from target.
            /// </summary>
            public Vector3 approachOffset;
            /// <summary>
            /// True if should reach for target.
            /// </summary>
            public bool reach = false;
            /// <summary>
            /// True if should hug target upon arrival.
            /// </summary>
            public bool hug = false;
            /// <summary>
            /// True if should punch target.
            /// </summary>
            public bool punch = false;
            /// <summary>
            /// True if should grapple with target.
            /// </summary>
            public bool grapple = false;

            /// <summary>
            /// Sets fields to indicate the specified object should be approached.
            /// </summary>
            /// <param name="o">Approach target</param>
            public void Approach(TwigObject o)
            {
                Approach(o, 0);
            }

            /// <summary>
            /// Sets fields to indicate the specified object should be approached.
            /// </summary>
            /// <param name="o">Approach target</param>
            /// <param name="distance">Distance to maintain from target</param>
            public void Approach(TwigObject o, float distance)
            {
                target = o;
                approachDistance = distance;
                approachOffset = Vector3.Zero;
            }

            /// <summary>
            /// Sets fields to indicate the specified object should be approached.
            /// </summary>
            /// <param name="o">Approach target</param>
            /// <param name="distance">Distance to maintain from target</param>
            /// <param name="direction">Direction from which to approach the target</param>
            public void Approach(TwigObject o, float distance, Vector3 direction)
            {
                target = o;
                approachDistance = distance;
                approachOffset = direction;
                //approachOffset.Normalize();
            }
        }

        /// <summary>
        /// True if object is this creature's approach target.
        /// </summary>
        public override bool IsApproachTarget(TwigObject o)
        {
            return approach.target == o || o.IsSupporting(approach.target);
        }

        /// <summary>
        /// Idle behavior for the approach system.
        /// Initialized in Mammal constructor.
        /// </summary>
        ApproachNothing approachIdle;

        /// <summary>
        /// Idle behavior for the approach system.
        /// </summary>
        class ApproachNothing : Behavior<ApproachParameters>
        {
            public override string Name
            {
                get
                {
                    return "Idle";
                }
            }
            protected override void Update()
            {
            }
        }

        /// <summary>
        /// Returns an action to approach the specified object
        /// </summary>
        public Action Goto(TwigObject target, float distance, Vector3 direction)
        {
            ApproachParameters p = new ApproachParameters();
            p.Approach(target, distance, direction);
            return approachActionBehavior.MakeAction(p);
        }

        /// <summary>
        /// Returns an action to approach the specified object
        /// </summary>
        public Action Goto(TwigObject target, float distance)
        {
            ApproachParameters p = new ApproachParameters();
            p.Approach(target, distance);
            return approachActionBehavior.MakeAction(p);
        }

        /// <summary>
        /// Returns an action to approach the specified object
        /// </summary>
        public Action Goto(TwigObject target)
        {
            ApproachParameters p = new ApproachParameters();
            p.Approach(target, target.CollisionRadius+1);
            return approachActionBehavior.MakeAction(p);
        }

        ApproachActionBehavior approachActionBehavior;

        /// <summary>
        /// Supports the execution of Approach actions.
        /// When an Approach action is executed in a sequence, it stores itself in this behavior's
        /// currentAction slot, activating this behavior, and causing its ApproachParameters to be
        /// fed to the main Approach behavior.  When the character moves within the specified range of
        /// the target, this behavior clears its currentAction field, signaling the action as being
        /// completed.
        /// </summary>
        class ApproachActionBehavior : Behavior<ApproachParameters>
        {
            /// <summary>
            /// The behavior's name
            /// </summary>
            public override string Name
            {
                get
                {
                    return "ApproachAction";
                }
            }
            Mammal me;

            /// <summary>
            /// A behavior controller for Approach actions, as opposed to the Approach action itself (which
            /// represents an instance of approaching an object within a plan), or the Approach behavior itself
            /// (which steers the body toward the current target, regardless of what behavior or action has specified
            /// the target).
            /// </summary>
            public ApproachActionBehavior(Mammal me)
            {
                this.me = me;
            }

            /// <summary>
            /// Checks whether an approach action is active and updates Activation and MotorOutput accordingly.
            /// </summary>
            protected override void Update()
            {
                if (currentAction == null)
                    mActivation = 0;
                else
                {
                    mActivation = 100;
                    mOutput = currentAction.parameters;
                    //me.Attention.Focus = mOutput.target;
                    if (mOutput.target==null
                        || MathUtil.Distance(me.Position, mOutput.target.Position) < mOutput.approachDistance)
                    {
                        mActivation = 0;
                        currentAction = null;
                    }
                }
            }

            /// <summary>
            /// Makes a new Action object representing approach to a specified object
            /// </summary>
            public Action MakeAction(ApproachParameters p)
            {
                return new ApproachAction(this, p);
            }

            ApproachAction currentAction = null;

            class ApproachAction : Action
            {
                ApproachActionBehavior behavior;
                public ApproachParameters parameters;

                public ApproachAction(ApproachActionBehavior b, ApproachParameters p)
                {
                    behavior = b;
                    parameters = p;
                }

                public override void Start()
                {
                    behavior.currentAction = this;
                }

                public override void Abort()
                {
                    if (behavior.currentAction==this)
                        behavior.currentAction = null;
                }

                public override ActionStatus Status
                {
                    get {
                        if (behavior.currentAction == this)
                            return ActionStatus.Running;
                        else
                            return ActionStatus.Succeed;
                    }
                }
            }
        }

        /// <summary>
        /// Drives locomotion system to approach object selected by approachController.
        /// Initialized in constructor for Mammal.
        /// </summary>
        ApproachBehavior approach;

        /// <summary>
        /// Drives locomotion system to approach object selected by approachController.
        /// </summary>
        class ApproachBehavior : Behavior<Vector3>
        {
            /// <summary>
            /// Pointer back to the mammal object we're controlling.
            /// </summary>
            Mammal me;
            public ApproachBehavior(Mammal me)
            {
                this.me = me;
            }

            /// <summary>
            /// Object to be approached
            /// </summary>
            public TwigObject target;
            /// <summary>
            /// Distance from target at which to stop
            /// </summary>
            public float approachDistance;
            /// <summary>
            /// Direction to approach target from
            /// </summary>
            public Vector3 approachOffset;
            /// <summary>
            /// Afferent.  Current heading of target.  CAN BE ZERO!
            /// </summary>
            public Vector3 targetDirection;
            /// <summary>
            /// Afferent.  True when character has arrived at target.
            /// </summary>
            public bool approachComplete;

            /// <summary>
            /// Repulsive force generated by obstacles
            /// </summary>
            public Vector3 repulsion;

            protected virtual Vector3 TargetVector()
            {
                Vector3 targetLocation;
                Adult a = target as Adult;
                if (me.hug && a != null && me is Child)
                    targetLocation = MathUtil.ProjectToGroundPlane(a.legs.left.joint.Position);
                else
                    targetLocation = target.Position;

                return targetLocation + approachOffset - me.Position
;
            }

            protected virtual Vector3 ObstacleRepulsion(Vector3 here, float distanceToTarget, out bool obstacleInWay, out bool crowded)
            {
                Vector3 repulsion = Vector3.Zero;
                obstacleInWay = false;
                crowded = false;

                foreach (TwigObject o in TwigObject.AllTwigObjects)
                {
                    if (o != me && (!me.IsApproachTarget(o) || approachOffset != Vector3.Zero) 
                        && !o.IsGrasped
                        && !o.IsApproachTarget(me)
                        && o.IsObstacle)
                    {
                        Vector3 offset = (here - o.Position);
                        float dist = offset.Length();

                        if (dist < distanceToTarget)
                        {
                            obstacleInWay = true;
                            if (!me.IsApproachTarget(o))
                                crowded = true;
                        }

                        if (dist < 10f)
                        {
                            dist -= o.CollisionRadius + me.CollisionRadius + me.AvoidDistance(o);
                            offset *= 0.4f / Math.Max(0.1f, dist * dist);
                            repulsion += offset
                                + Vector3.Cross(offset, Vector3.Dot(offset, me.pelvisRight) > 0
                                         ? -Vector3.Up : Vector3.Up);
                        }
                    }
                }
                return repulsion;
            }

            //
            // This is the real steering logic ofr the system.
            //
            protected override void Update()
            {
                // Find out what our destination is.
                ApproachParameters winner = me.approachController.MotorOutput;
                target = winner.target;
                approachOffset = winner.approachOffset;
                approachDistance = winner.approachDistance;
                //me.Chart("Winner reach", winner.reach ? 1 : -1, 2f);
                me.reach = winner.reach;
                me.hug = winner.hug;
                me.punch = winner.punch;
                me.grapple = winner.grapple;

                // Don't try to approach something we're already holding
                if (me.IsHolding(target))
                    target = null;

                // Figure out where the target is relative to us.
                Vector3 targetVector;
                Vector3 approachVector;
                targetDirection = targetVector = approachVector = Vector3.Zero;
                Vector3 here = me.Position;

                if (target != null)
                {
                    targetDirection = targetVector = approachVector = TargetVector();
                    MathUtil.SafeNormalize(ref targetDirection);
                }

                me.reachDirection = targetDirection;

                // Limit approachVector to length 5.
                if (approachVector.Length() > 5f)
                    approachVector *= 5f / approachVector.Length();

                float distanceToTarget = targetVector.Length();
                repulsion = Vector3.Zero;

                // Find nearby obstacles and compute a repulsive force from them.
                me.blocked = false;

                bool obstacleInWay;

                if (target != null)
                {
                    repulsion = ObstacleRepulsion(here, distanceToTarget, out obstacleInWay, out me.crowded);
                    if (obstacleInWay)
                    {
                        if (approachVector != Vector3.Zero && (approachVector + repulsion).LengthSquared() < 0.005)
                            me.blocked = true;
                        approachVector += repulsion;
                    }

                    approachComplete = (target != null) && distanceToTarget <= approachDistance;
                    if (approachComplete)
                        approachVector = Vector3.Zero;
                    else
                        me.WaitingFor("Approach: target in range");
                }
                //me.Chart("blocked", me.blocked ? 1 : -1, 2.5f);
                //me.Chart("crowded", me.crowded ? 1 : -1, 1.5f);

                MotorOutput = approachVector + ((target==null)?Vector3.Zero:target.Velocity);
                Activation = approachComplete ? 0 : 1;
                //System.Diagnostics.Debug.Assert(!(Activation > 0 && MotorOutput == Vector3.Zero));
            }
        }
    }
}