// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;

#pragma warning disable 1591

namespace Twig
{
    partial class Mammal
    {
        /// <summary>
        /// Salience of the specified object.
        /// </summary>
        /// <param name="o">Object</param>
        /// <returns>Salience</returns>
        public abstract float Salience(TwigObject o);
        /// <summary>
        /// Enumerates objects in view of the creature.
        /// </summary>
        public abstract IEnumerable<TwigObject> ObjectsInView { get; }
        /// <summary>
        /// Tracks information about objects in short-term memory.
        /// </summary>
        public AttentionBuffer Attention;

        /// <summary>
        /// Tracks information about objects in short-term memory.
        /// </summary>
        [System.Diagnostics.DebuggerDisplay("Focus={Focus}")]
        public class AttentionBuffer
        {
            /// <summary>
            /// Number of objects that can be stored in attention buffer.
            /// WARNING: If this becomes larger than 8, you need to change SlotSet to be an unsigned short.
            /// </summary>
            public const int AttentionCapacity = 7;

            /// <summary>
            /// A slot represents the position of an object within the AttentionBuffer.
            /// Any object within the AttentionBuffer has a slot associated with it.
            /// </summary>
            public struct Slot
            {
                /// <summary>
                /// Makes a slot from its slotnumber
                /// </summary>
                /// <param name="slotNumber"></param>
                public Slot(byte slotNumber)
                {
                    number = slotNumber;
                }

                /// <summary>
                /// The actual position in memory
                /// </summary>
                byte number;

                /// <summary>
                /// The actual position within the AttentionBuffer.  It doesn't have
                /// any significance other than which slot of the array to look in.
                /// </summary>
                public byte Number
                {
                    get
                    {
                        return number;
                    }
                }

                /// <summary>
                /// Returns a bitmask with the relevant bit set.
                /// </summary>
                public byte Mask
                {
                    get
                    {
                        return (byte)(1 << number);
                    }
                }
            }

            /// <summary>
            /// Represents a set of slots
            /// </summary>
            public struct SlotSet
            {
                /// <summary>
                /// The empty SlotSet
                /// </summary>
                public static readonly SlotSet Empty = new SlotSet((byte)0);
                byte mask;

                /// <summary>
                /// Bitmask of slotset.
                /// </summary>
                public byte Mask
                {
                    get { return mask; }
                }

                /// <summary>
                /// True if the slot is included in the set.
                /// </summary>
                public bool this[Slot s]
                {
                    get
                    {
                        return (mask & s.Mask) > 0;
                    }
                }

                public bool this[byte i]
                {
                    get
                    {
                        return (mask & (1 << i)) > 0;
                    }
                }

                /// <summary>
                /// True if the set is the empty set.
                /// </summary>
                public bool IsEmpty
                {
                    get
                    {
                        return mask == 0;
                    }
                }

                /// <summary>
                /// Creates a slot set from a bitmask.
                /// </summary>
                public SlotSet(byte mask)
                {
                    this.mask = mask;
                }

                /// <summary>
                /// Creates a singleton slot set.
                /// </summary>
                public SlotSet(Slot s)
                {
                    mask = s.Mask;
                }

                public static SlotSet operator &(SlotSet a, SlotSet b)
                {
                    return new SlotSet((byte)(a.mask & b.mask));
                }
                public static SlotSet operator |(SlotSet a, SlotSet b)
                {
                    return new SlotSet((byte)(a.mask | b.mask));
                }
                public static SlotSet operator ~(SlotSet a)
                {
                    return new SlotSet(((byte) ~a.mask));
                }
            }

            public SlotSet TestSlots(Predicate.TestProcedure t)
            {
                byte set = 0;
                for (int i = 0; i < AttentionCapacity; i++)
                {
                    TwigObject o = recent[i];
                    if (o!=null & t(o))
                        set |= (byte)(1 << i);
                }
                return new SlotSet(set);
            }

            /// <summary>
            /// Index of focus object within attention buffer
            /// </summary>
            int focusIndex = 0;
            TwigObject[] recent = new TwigObject[AttentionCapacity];
            double[] timeStamp = new double[AttentionCapacity];
            double[] lastLook = new double[AttentionCapacity];
            /// <summary>
            /// TwigObject to which this AttentionBuffer belongs
            /// </summary>
            Mammal me;
            double updateTimeout;
            /// <summary>
            /// Minimum period (in seconds) for which an object should be the focus of attention.
            /// </summary>
            public float minimumAttentionPeriod = 1f;

            public AttentionBuffer(Mammal me)
            {
                this.me = me;
            }

            /// <summary>
            /// Shifts Focus to maximal salience object in view
            /// </summary>
            public void Update()
            {
                // Update looking information
                int lookSlot = FastObjectSlotNumber(me.gazeTarget);
                if (me.gazeTarget!= null && lookSlot >= 0)
                    lastLook[lookSlot] = Physics.time;

                // Don't bother shifting attention unless
                // we've been attending to the focus for
                // a certain period of time
                if (Physics.time < updateTimeout)
                    return;

                TwigObject focus = Focus;
                float focusSalience = me.Salience(focus);

                foreach (TwigObject o in me.ObjectsInView)
                {
                    if (o.Cloaked)
                        continue;
                    float salience = me.Salience(o);
                    if (salience > focusSalience)
                    {
                        focus = o;
                        focusSalience = salience;
                    }
                }
                foreach (TwigObject o in AttendedObjects)
                {
                    if (o.Cloaked)
                        continue;
                    float salience = me.Salience(o);
                    if (salience > focusSalience)
                    {
                        focus = o;
                        focusSalience = salience;
                    }
                }
                // Now do the approach target just for good measure
                float tsalience = me.Salience(me.approach.target);
                if (tsalience > focusSalience)
                {
                    focus = me.approach.target;
                    focusSalience = tsalience;
                }

                if (Focus != focus)
                {
                    updateTimeout = Physics.time + minimumAttentionPeriod;
                    Focus = focus;
                }
            }

            /// <summary>
            /// The creature's current focus of attention.
            /// </summary>
            public TwigObject Focus
            {
                get
                {
                    return recent[focusIndex];
                }
                set
                {
                    focusIndex = ObjectSlotNumber(value);
                }
            }

            /// <summary>
            /// Find a slot for o in attention buffer.
            /// If object is already present, return its current slot
            /// and update that slot's timestamp.  If no room and not
            /// already in buffer, kick out oldest entry
            /// </summary>
            public Slot ObjectSlot(TwigObject o)
            {
                return new Slot(ObjectSlotNumber(o));
            }

            /// <summary>
            /// Find a slot for o in attention buffer.
            /// If object is already present, return its current slot
            /// and update that slot's timestamp.  If no room and not
            /// already in buffer, kick out oldest entry
            /// </summary>
            public byte ObjectSlotNumber(TwigObject o)
            {
                byte oldest = 0;
                double oldestTS = timeStamp[oldest];

                // We actually want it to start at 0, not 1
                // to force the check for whether slot 0 already
                // holds the object
                for (byte i = 0; i < AttentionCapacity; i++)
                {
                    if (recent[i] == o)
                    {
                        // Already in buffer
                        timeStamp[i] = Physics.time;   // refresh it
                        return i;
                    }
                    else if (recent[i] == null)   // Empty slots always win
                    {
                        oldest = i;
                        oldestTS = 0;
                    }
                    else if (timeStamp[i] < oldestTS)
                    {
                        oldest = i;
                        oldestTS = timeStamp[i];
                    }
                }
                // Wasn't already in buffer; kick out oldest
                recent[oldest] = o;
                timeStamp[oldest] = Physics.time;
                return oldest;
            }

            /// <summary>
            /// Finds slot number of object, but does not allocate space for it or update its timestamp.
            /// </summary>
            public int FastObjectSlotNumber(TwigObject o)
            {
                return Array.IndexOf<TwigObject>(recent, o);
            }

            /// <summary>
            /// Returns the actual object associated with a slot number.
            /// </summary>
            public TwigObject SlotObject(Slot s)
            {
                return recent[s.Number];
            }

            /// <summary>
            /// Enumerates objects in short term memory.
            /// </summary>
            public IEnumerable<TwigObject> AttendedObjects
            {
                get
                {
                    for (int i = 0; i < AttentionCapacity; i++)
                    {
                        if (recent[i] != null)
                            yield return recent[i];
                    }
                }
            }

            [System.Diagnostics.DebuggerDisplay("{obj}")]
            internal class AttentionInfo
            {
                public TwigObject obj;
                public double timeSinceLastLook;
            }

            /// <summary>
            /// Enumerates attention information for objects in short term memory.
            /// WARNING! Iterator recycles the same AttentionInfo object.
            /// </summary>
            internal IEnumerable<AttentionInfo> AttendedObjectsInfo
            {
                get
                {
                    AttentionInfo info = new AttentionInfo();
                    for (int i = 0; i < AttentionCapacity; i++)
                    {
                        if (recent[i] != null)
                        {
                            info.obj = recent[i];
                            info.timeSinceLastLook = Physics.time - lastLook[i];
                            yield return info;
                        }
                    }
                }
            }
        }
    }
}
