// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Twig
{
    /// <summary>
    /// Type of camera shot to use
    /// </summary>
    public enum CameraShot
    {
        /// <summary>
        /// A closeup, focusing on the head
        /// </summary>
        CloseUp,
        /// <summary>
        /// A medium shot, showing most of the torso
        /// </summary>
        MediumShot,
        /// <summary>
        /// Full body shot
        /// </summary>
        LongShot
    }

    /// <summary>
    /// Determines viewpoint from which subject is shot
    /// </summary>
    public enum CameraHeight
    {
        /// <summary>
        /// Camera looks up at subject
        /// </summary>
        WormsEye,
        /// <summary>
        /// Camera at chest level
        /// </summary>
        ChestLevel,
        /// <summary>
        /// Camera is at subject's eye level
        /// </summary>
        EyeLevel,
        /// <summary>
        /// Camera looks down at subject
        /// </summary>
        BirdsEye
    }

    /// <summary>
    /// A scriptable camera that characters can (optionally) respond to
    /// </summary>
    public class Camera : TwigObject
    {
        /// <summary>
        /// Type of shot to use
        /// </summary>
        public CameraHeight height = CameraHeight.ChestLevel;
        /// <summary>
        /// Type of shot to use
        /// </summary>
        public CameraShot shot;
        /// <summary>
        /// Direction from which camera should shoot the subject
        /// </summary>
        public Vector3 CameraAngle = Vector3.UnitZ;
        /// <summary>
        /// Distance to subject normalized to subject's height.
        /// </summary>
        public float subjectDistance = 3.0f;

        float[] cameraHeights = new float[] { 0.15f, 0.7f, 1f, 3f };

        TwigObject subject;

        /// <summary>
        /// Focus of attention of camera
        /// </summary>
        public TwigObject Subject
        {
            get { return subject; }
            set { subject = value; }
        }

        /// <summary>
        /// A very lame camera controller
        /// </summary>
        public Camera(TwigGame g, string name)
            : base(g, name, true)
        {
            shot = CameraShot.LongShot;
            Cloaked = true;
#if PILOTING
            g.DocumentKey(Keys.Left, "Circle camera to the left around subject");
            g.DocumentKey(Keys.Right, "Circle camera to the right around subject");
            g.DocumentKey(Keys.Up, "Truck camera forward");
            g.DocumentKey(Keys.Down, "Truck camera backward");
            g.BindKey(KeyModifiers.Control, Keys.Up, "Switche between chest level, eye level, bird's eye, and worm's eye",
                delegate(Keys ignore)
                {
                    height++;
                    if (height > CameraHeight.BirdsEye)
                        height = CameraHeight.BirdsEye;
                }
            );
            g.BindKey(KeyModifiers.Control, Keys.Down, "Switche between chest level, eye level, bird's eye, and worm's eye",
               delegate(Keys ignore)
               {
                    height--;
                    if ((int)height < 0)
                        height = (CameraHeight)0;
               }
            );
#endif
        }

        /// <summary>
        /// Move the game's viewpoint as appropriate
        /// </summary>
        public override void Update(GameTime gameTime)
        {
#if PILOTING
            if (TwigGame.KeyModifiers == KeyModifiers.None)
            {
                if (TwigGame.IsKeyDown(Keys.Left))
                    CameraAngle = Vector3.Transform(CameraAngle, Matrix.CreateRotationY(-(float)Physics.deltaT));
                else if (TwigGame.IsKeyDown(Keys.Right))
                    CameraAngle = Vector3.Transform(CameraAngle, Matrix.CreateRotationY((float)Physics.deltaT));
                CameraAngle.Normalize();

                if (TwigGame.IsKeyDown(Keys.Up))
                    subjectDistance -= 0.05f;
                else if (TwigGame.IsKeyDown(Keys.Down))
                    subjectDistance += 0.05f;
            }

            if (TwigGame.DebugObject != null)
                subject = TwigGame.DebugObject;
#endif

            if (subject == null)
                return;
            float cameraDistance;
            float cameraHeight;

            Mammal m = subject as Mammal;
            if (m != null)
            {
                cameraDistance = m.Height * subjectDistance;
                cameraHeight = m.EyeLevel*cameraHeights[(int)height];
                TwigGame.CameraTargetLocation = 0.9f*TwigGame.CameraTargetLocation+ 0.1f*(m.Position + ((height==CameraHeight.ChestLevel)?cameraHeight:m.EyeLevel) * Vector3.Up);
            }
            else
            {
                cameraDistance = subjectDistance;
                cameraHeight = subject.LookPoint.Y * cameraHeights[(int)height];
                TwigGame.CameraTargetLocation = subject.LookPoint;
            }

            TwigGame.CameraPosition =
                0.9f * TwigGame.CameraPosition
                + 0.1f * (subject.Position+CameraAngle*cameraDistance+Vector3.Up*cameraHeight);
            base.Update(gameTime);
        }

        /// <summary>
        /// Change camera fixation to LookPoint of subject
        /// </summary>
        public void LookAt(TwigObject subject, float distance)
        {
            this.subject = subject;
            this.subjectDistance = distance;
            this.height = CameraHeight.BirdsEye;
        }

        /// <summary>
        /// Nothing to constrain
        /// </summary>
        public override void Constrain()
        { }

        /// <summary>
        /// Cameras are modeled as points, so CoM is the camera's point of projection.
        /// </summary>
        public override Vector3 CenterOfMass
        {
            get { return LookPoint; }
        }

        /// <summary>
        /// Location of the camera's point of projection.
        /// </summary>
        public override Vector3 LookPoint
        {
            get { return TwigGame.CameraPosition; }
        }

        /// <summary>
        /// Position of camera (in particular, of its point of projection)
        /// </summary>
        public override Vector3 Position
        {
            get { return LookPoint; }
        }

        /// <summary>
        /// Cameras are modeled as massless.
        /// </summary>
        public override float Mass
        {
            get { return 0; }
        }

        /// <summary>
        /// Cameras are invisible to collisions when cloaked, otherwise have a 1m collision radius
        /// </summary>
        public override float CollisionRadius
        {
            get { return Cloaked?0:1; }
        }
    }
}
