// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Graphics;
using System.Reflection;

namespace Twig
{
    /// <summary>
    /// Lets you click on objects to select them as the debugObject
    /// </summary>
    public class DebugPicker : HUD
    {
        TwigGame twigGame;

        abstract class DebugSwitch
        {
            public string description;
            public abstract bool State
            {
                get;
                set;
            }
        }

        class FieldDebugSwitch : DebugSwitch
        {
            public FieldDebugSwitch(string desc, FieldInfo field) {
                description = desc;
                this.field = field;
            }
            FieldInfo field;
            public override bool State
            {
                get
                {
                    return (bool)field.GetValue(null);
                }
                set
                {
                    field.SetValue(null, value);
                }
            }
        }

        static List<DebugSwitch> debugSwitches = new List<DebugSwitch>();
        static DebugPicker()
        {
            foreach (Assembly a in AppDomain.CurrentDomain.GetAssemblies())
                foreach (Type t in a.GetTypes())
                    foreach (FieldInfo field in t.GetFields())
                    {
                        object[] attrs = field.GetCustomAttributes(typeof(DebugSwitchAttribute), false);
                        if (attrs.Length > 0)
                            debugSwitches.Add(new FieldDebugSwitch(((DebugSwitchAttribute)attrs[0]).description, field));
                    }

            debugSwitches.Sort(delegate(DebugSwitch a, DebugSwitch b)
            {
                return String.Compare(a.description, b.description);
            });
        }

        /// <summary>
        /// Lets you click on objects to select them as the debugObject
        /// </summary>
        public DebugPicker(TwigGame g, Keys displaySwitchesKey, Keys nextKey)
            : base(g, displaySwitchesKey, "Toggle display of debug switches", false)
        {
            twigGame = g;
            TwigGame.BindKey(nextKey, "Switch to next character", delegate(Keys ignore)
            {
                NextObject();
            });
        }

        /// <summary>
        /// Checks for mouse hit if left button is initially pressed
        /// </summary>
        /// <param name="gameTime"></param>
        public override void Update(GameTime gameTime)
        {
            if (Visible)
                HandleSwitchPicking();
            else
                HandleDebugObjectPicking();
        }

        static bool mouseUpLastTick = true;
        void HandleSwitchPicking()
        {
            int lineHeight = twigGame.HUDFont.LineSpacing;
            MouseState mouse = Mouse.GetState();
            int chosenItem = mouse.Y / lineHeight;

            if (mouse.LeftButton==ButtonState.Pressed && mouseUpLastTick && chosenItem<debugSwitches.Count) {
                debugSwitches[chosenItem].State = !debugSwitches[chosenItem].State;
            } else
            mouseUpLastTick = mouse.LeftButton==ButtonState.Released;
        }

        void HandleDebugObjectPicking() {
            MouseState current = Mouse.GetState();

            if (current.LeftButton == ButtonState.Released)
                return;

#if oldUnproject
            Viewport viewport = device.Viewport;
            Matrix proj = device.Transform.Projection;
            Matrix viewInv = Matrix.Invert(proj);
            Vector3 ic = new Vector3(((2f * point.X) / viewport.Width - 1f) / proj.M11,
                (1f - (2f * point.Y) / viewport.Height) / proj.M22,
                1);

            Vector3 rayCenter = new Vector3(viewInv.M41, viewInv.M42, viewInv.M43);
            Vector3 rayDirection = new Vector3(Vector3.Dot(ic, new Vector3(viewInv.M11, viewInv.M21, viewInv.M31)),
                Vector3.Dot(ic, new Vector3(viewInv.M12, viewInv.M22, viewInv.M32)),
                Vector3.Dot(ic, new Vector3(viewInv.M13, viewInv.M23, viewInv.M33)));
            rayDirection.Normalize();
            mouseRayCenter = rayCenter;
            mouseRayDirection = rayDirection;
#endif
            Matrix world = Matrix.Identity;
            Vector3 near = twigGame.Device.Viewport.Unproject(new Vector3(current.X, current.Y, 0),
                                                                twigGame.ProjectionMatrix, twigGame.ViewMatrix, world);
            Vector3 far = twigGame.Device.Viewport.Unproject(new Vector3(current.X, current.Y, 1),
                                                                twigGame.ProjectionMatrix, twigGame.ViewMatrix, world);
            Vector3 rayCenter = near;
            Vector3 rayDirection = far - near;
            rayDirection.Normalize();
            //Program.protagonist.UnHang();
            //Program.protagonist.HangAt(rayCenter + 10 * rayDirection);

            float nearestDepth = 9999999f;
            TwigObject nearest = null;
            Node nearestNode = null;
            foreach (TwigObject o in TwigObject.AllTwigObjects)
            {
                Mammal m = o as Mammal;
                if (m != null)
                {
                    foreach (Link l in m.Links)
                    {
                        float depth = Vector3.Dot(l.node1.PositionForDisplay - rayCenter, rayDirection);
                        if (depth < nearestDepth)
                        {
                            Vector3 projected = (l.node1.PositionForDisplay - rayDirection * depth) - rayCenter;
                            //System.Diagnostics.Debug.Assert(Vector3.Dot(projected, rayDirection) < 0.001f);
                            if (projected.LengthSquared() < 0.5f)
                            {
                                // Hit
                                nearestDepth = depth;
                                nearest = o;
                                nearestNode = l.node1;
                            }
                        }
                        depth = Vector3.Dot(l.node2.PositionForDisplay - rayCenter, rayDirection);
                        if (depth < nearestDepth)
                        {
                            Vector3 projected = (l.node2.PositionForDisplay - rayDirection * depth) - rayCenter;
                            //System.Diagnostics.Debug.Assert(Vector3.Dot(projected, rayDirection) < 0.001f);
                            if (projected.LengthSquared() < 0.1f)
                            {
                                // Hit
                                nearestDepth = depth;
                                nearest = o;
                                nearestNode = l.node2;
                            }
                        }
                    }
                }
            }
            if (nearest != null)
                twigGame.DebugObject = nearest;
        }

        void NextObject()
        {
            int oldIndex = (twigGame.DebugObject==null)?0:twigGame.Components.IndexOf(twigGame.DebugObject);
            for (int index = (oldIndex + 1) % twigGame.Components.Count;
                index != oldIndex;
                index = (index + 1) % twigGame.Components.Count)
            {
                IGameComponent g = twigGame.Components[index];
                if (g is Mammal)
                {
                    twigGame.DebugObject = (TwigObject)g;
                    break;
                }
            }

        }

        /// <summary>
        /// Draws the debugSwitches, if showSwitches is set.
        /// </summary>
        /// <param name="gameTime"></param>
        public override void Draw(GameTime gameTime)
        {
            int lineHeight = twigGame.HUDFont.LineSpacing;
            MouseState mouse = Mouse.GetState();
            int chosenItem = mouse.Y / lineHeight;

            twigGame.HUDSprites.Begin();
            for (int i = 0; i < debugSwitches.Count; i++)
            {
                twigGame.HUDSprites.DrawString(twigGame.HUDFont, debugSwitches[i].description,
                    new Vector2(0, lineHeight * i),
                    (i == chosenItem) ? Color.White : debugSwitches[i].State ? Color.Red : Color.Gray);
            }
            twigGame.HUDSprites.End();
        }
    }
}
