// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Twig
{

    /// <summary>
    /// Represents a character utterance to be displayed on screen.
    /// </summary>
    public class DialogBalloon
    {
        string[] lines;
        Rectangle bbox;
        Mammal speaker;
        const int horizontalPadding = 10;
        static VertexPositionColor[] connectorLines = new VertexPositionColor[100];
        int expirationTime;

        static int[] seedHeights = new int[] { 40, 70, 10 };
        /// <summary>
        /// 
        /// </summary>
        public static float speed = 1f;

        DialogBalloon(DialogBalloonManager manager, Mammal speaker, string[] lines)
        {
            int totalChars=0;
            this.lines = lines;
            this.speaker = speaker;
            SpriteFont f = speaker.TwigGame.DialogFont;
            foreach (string s in lines) {
                Vector2 size = f.MeasureString(s);
                bbox.Height += (int)(size.Y+0.5f);
                bbox.Width = Math.Max(bbox.Width, (int)(size.X+0.5f));
                totalChars += s.Length;
            }
            bbox.Width += 2*horizontalPadding;
            bbox.X = speaker.ScreenPosition.X - bbox.Width / 2;
            foreach (int h in seedHeights)
            {
                bool collision = false;
                bbox.Y = h;
                foreach (DialogBalloon b in manager.balloons)
                {
                    if (IntervalsIntersect(h, h + bbox.Height, b.bbox.Top, b.bbox.Bottom))
                        collision = true;
                }
                if (!collision)
                    break;
            }

            expirationTime = Physics.tickCount + (int)Math.Round(Math.Max(50, 3 * totalChars)/speed);
        }

        static bool IntervalsIntersect(int aStart, int aWidth, int bStart, int bWidth)
        {
            if (aStart < bStart)
                return bStart < aStart + aWidth;
            else
                return aStart < bStart + bWidth;
        }

        /// <summary>
        /// Controls the display of dialog balloons.
        /// </summary>
        public class DialogBalloonManager : GameComponent
        {
            bool pauseOnDialog = false;

            /// <summary>
            /// Force pausing before removing a dialog balloon
            /// </summary>
            public bool PauseOnDialog
            {
                get { return pauseOnDialog; }
                set { pauseOnDialog = value; }
            }

            TwigGame twigGame;

            /// <summary>
            ///  A manager for the dialog balloons in the game.
            /// </summary>

            public DialogBalloonManager(TwigGame g) : base(g)
            {
                twigGame = g;
            }

            /// <summary>
            /// All currently active dialog balloons
            /// </summary>
            public readonly List<DialogBalloon> balloons = new List<DialogBalloon>();

            /// <summary>
            /// This is a little 2D collision avoidance system for the dialog balloons
            /// </summary>
            void UpdateBalloonPositions()
            {
                if (pauseOnDialog)
                {
                    for (int i = 0; i < balloons.Count; i++)
                        if (balloons[i].expirationTime < Physics.tickCount)
                            twigGame.Pause = true;
                }
                else
                    balloons.RemoveAll(delegate(DialogBalloon b)
                    {
                        return b.expirationTime < Physics.tickCount;
                    });

                for (int iter = 0; iter < 10; iter++)
                {
                    bool changed = false;
                    int screenWidth = twigGame.GraphicsDevice.Viewport.Width;

                    for (int i = 0; i < balloons.Count; i++)
                    {
                        DialogBalloon a = balloons[i];
                        if (a.speaker.Visible)
                            a.bbox.X = a.speaker.ScreenPosition.X - a.bbox.Width / 2;
                        for (int j = i + 1; j < balloons.Count; j++)
                        {
                            DialogBalloon b = balloons[j];
                            if (a.bbox.Intersects(b.bbox))
                            {
                                //changed = true;
                                if (IntervalsIntersect(a.bbox.Left, a.bbox.Width, b.bbox.Left, b.bbox.Width))
                                {
                                    // Amount required to shift a to the left of b
                                    int aLeft = a.bbox.Right - b.bbox.Left;
                                    int bLeft = b.bbox.Right - a.bbox.Left;
                                    if (aLeft < bLeft)
                                    {
                                        if ((Physics.tickCount & 1) == 0)
                                            a.bbox.Offset(-1, 0);
                                        else
                                            b.bbox.Offset(1, 0);
                                    }
                                    else
                                    {
                                        if ((Physics.tickCount & 1) == 0)
                                            a.bbox.Offset(1, 0);
                                        else
                                            b.bbox.Offset(-1, 0);
                                    }
                                }
                                if (IntervalsIntersect(a.bbox.Top, a.bbox.Height, b.bbox.Top, b.bbox.Height))
                                {
                                    // Amount required to shift a to the left of b
                                    int aTop = a.bbox.Bottom - b.bbox.Top;
                                    int bTop = b.bbox.Bottom - a.bbox.Top;
                                    if (aTop < bTop)
                                    {
                                        if ((Physics.tickCount & 1) == 0)
                                            a.bbox.Offset(0, -1);
                                        else
                                            b.bbox.Offset(0, 1);
                                    }
                                    else
                                    {
                                        if ((Physics.tickCount & 1) == 0)
                                            a.bbox.Offset(0, 1);
                                        else
                                            b.bbox.Offset(0, -1);
                                    }
                                }
                            }
                        }
                        if (a.bbox.Y < 0)
                            a.bbox.Y = 0;
                        a.bbox.X = Math.Min(screenWidth-a.bbox.Width, Math.Max(0, a.bbox.X));
                    }
                    // We've reached fixed-point
                    if (!changed)
                        return;
                }
                // Bad: give up looking for fixed-point.
            }

            /// <summary>
            /// Make a new dialog ballon with the specified text
            /// </summary>
            public void Say(Mammal speaker, string[] lines)
            {
                balloons.Add(new DialogBalloon(this, speaker, lines));
            }

            /// <summary>
            /// Remove any dialog balloon(s) for this speaker.
            /// </summary>
            /// <param name="speaker"></param>
            public void RemoveDialog(TwigObject speaker)
            {
                balloons.RemoveAll(delegate(DialogBalloon b)
                {
                    return b.speaker == speaker;
                });
            }

            /// <summary>
            /// Remove the oldest balloon currently on the screen.
            /// </summary>
            public void RemoveLastDialogBalloon()
            {
                if (balloons.Count > 0)
                    balloons.RemoveAt(0);
            }

            /// <summary>
            /// True if speaker is speaking
            /// </summary>
            public bool IsSpeaking(TwigObject speaker)
            {
                return null != balloons.Find(delegate(DialogBalloon b)
                {
                    return b.speaker == speaker;
                });
            }

            /// <summary>
            /// True if nobody is speaking
            /// </summary>
            public bool NobodySpeaking
            {
                get
                {
                    return balloons.Count == 0;
                }
            }

            /// <summary>
            /// Draw all dialog balloons
            /// </summary>
            public void DrawBalloons(SpriteBatch batch, SpriteFont font, BasicEffect effect)
            {
                effect.Begin();
                foreach (EffectPass pass in effect.CurrentTechnique.Passes)
                {
                    pass.Begin();
                    UpdateBalloonPositions();
                    batch.Begin();
                    for (int i = 0; i < balloons.Count; i++)
                    {
                        batch.DrawString(font, balloons[i].lines[0],
                            new Vector2(balloons[i].bbox.Left + horizontalPadding, balloons[i].bbox.Top),
                            Color.White);
                    }

                    int lines = Math.Min(balloons.Count, connectorLines.Length / 2);
                    int actualLines = 0;
                    if (lines > 0)
                    {
                        for (int i = 0; i < lines; i++)
                        {
                            if (balloons[i].speaker.Visible)
                            {
                                Point head = balloons[i].speaker.HeadScreenPosition;
                                connectorLines[2 * actualLines] = new VertexPositionColor(new Vector3(head.X, head.Y - 10, 0),
                                    Color.White);
                                connectorLines[2 * actualLines + 1] =
                                    new VertexPositionColor(new Vector3(balloons[i].bbox.Left + balloons[i].bbox.Width / 2, balloons[i].bbox.Bottom, 0),
                                    Color.White);
                                actualLines++;
                            }
                        }
                        effect.GraphicsDevice.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.LineList,
                            connectorLines, 0, actualLines);
                    }

                    batch.End();
                    pass.End();
                }
                effect.End();
            }
        }

    }
}
