// Copyright 2007, 2008, 2009 Ian Horswill
// Much of the code in this file is based on the book Real-Time
// Collision Detection, by Christer Ericson, published by
// Morgan Kaufmann.
//
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Diagnostics;

//
// This code has been adapted from Real-Time Collision Detection by Christer Ericson,
// published by Morgan Kaufmann Publishers, Copyright (c) 2005 Elsevier Inc
// 
// I've ported the code from C++ to C#/XNA.  Not all code is ported, in part because some
// of it is sort of pseudocode with references to undefined functions.  I'll port that
// if/when there's need for it.
// 
// Point and Vector are now both the XNA Vector3 datatype.  And the Point2D is the XNA
// Vector2.  Because C# doesn't normally have pointers, I've had to modify the calling
// conventions to use out and ref variables.  In general, output parameters are out
// parameters unless the code is written so that it doesn't always return a value.  In
// that case, I use a ref parameter.
//

namespace Twig
{
    /// <summary>
    /// Contains static methods used for testing bounding volumes.
    /// </summary>
    public class GeometryPrimitives
    {
        /// <summary>
        /// A bounding sphere
        /// </summary>
        [DebuggerDisplay("Sphere({c}, {r})")]
        public struct Sphere
        {
            /// <summary>
            /// Creates a sphere of the specified center and radius
            /// </summary>
            public Sphere(Vector3 center, float radius)
            {
                c = center;
                r = radius;
            }
            /// <summary>
            /// Center of the sphere
            /// </summary>
            public Vector3 c;
            /// <summary>
            /// Radius of the sphere
            /// </summary>
            public float r;
        }
        /// <summary>
        /// Represents a plane in 3-space
        /// </summary>
        public struct Plane
        {
            /// <summary>
            /// Creates a plane from a normal (i.e. unit) vector and a distance from the origin.
            /// </summary>
            /// <param name="n">Normal specifying orientation of plane.  MUST BE A UNIT VECTOR!</param>
            /// <param name="d">Distance from origin</param>
            public Plane(Vector3 n, float d)
            {
                this.n = n;
                Debug.Assert(Math.Abs(n.LengthSquared() - 1) < 0.001f, "Plane normal is not a unit vector");
                this.d = d;
            }


            /// <summary>
            /// Given three noncollinear Vector3s (ordered ccw), compute the plane equation
            /// </summary>
            public Plane(Vector3 a, Vector3 b, Vector3 c)
            {
                n = Vector3.Normalize(Vector3.Cross(b - a, c - a));
                d = Vector3.Dot(n, a);
            }

            /// <summary>
            /// Plane normal. Vector3s x on the plane satisfy Vector3.Dot(n,x) = d
            /// MUST BE A UNIT VECTOR
            /// </summary>
            public Vector3 n;
            /// <summary>
            /// Distance of the plane from the origin.
            /// d = Vector3.Dot(n,p) for a given Vector3 p on the plane
            /// </summary>
            public float d;
        }

        /// <summary>
        /// Axis-Aligned Bounding Box
        /// </summary>
        public struct AABB
        {
            /// <summary>
            /// Creates an axis-aligned bounding box
            /// </summary>
            /// <param name="min">Vector of minimum coordinates of the box</param>
            /// <param name="max">Vector of minimum coordinates of the box</param>
            public AABB(Vector3 min, Vector3 max)
            {
                this.min = min;
                this.max = max;
            }
            /// <summary>
            /// Creates an axis-aligned bounding-box for a list of nodes.
            /// </summary>
            /// <param name="nodes"></param>
            public AABB(Node[] nodes)
            {
                min = max = nodes[0].Position;
                for (int i = 1; i < nodes.Length; i++)
                {
                    Vector3 p = nodes[i].Position;
                    if (p.X < min.X)
                        min.X = p.X;
                    if (p.X > max.X)
                        max.X = p.X;

                    if (p.Y < min.Y)
                        min.Y = p.Y;
                    if (p.Y > max.Y)
                        max.Y = p.Y;

                    if (p.Z < min.Z)
                        min.Z = p.Z;
                    if (p.Z > max.Z)
                        max.Z = p.Z;
                }
            }
            /// <summary>
            /// Vector of minimum coordinates of the box
            /// </summary>
            public Vector3 min;
            /// <summary>
            /// Vector of maximum coordinates of the box
            /// </summary>
            public Vector3 max;

            /// <summary>
            /// Center of the bounding box
            /// </summary>
            public Vector3 Center
            {
                get
                {
                    return 0.5f * (min + max);
                }
            }

            /// <summary>
            /// Width of the bounding box
            /// </summary>
            public float Width
            {
                get
                {
                    return max.X - min.X;
                }
            }

            /// <summary>
            /// Height of the bounding box
            /// </summary>
            public float Height
            {
                get
                {
                    return max.Y - min.Y;
                }
            }

            /// <summary>
            /// Depth of the bounding box
            /// </summary>
            public float Depth
            {
                get
                {
                    return max.Z - min.Z;
                }
            }

            void AdjoinPoint(Vector3 p)
            {
                if (p.X < min.X)
                    min.X = p.X;
                if (p.X > max.X)
                    max.X = p.X;

                if (p.Y < min.Y)
                    min.Y = p.Y;
                if (p.Y > max.Y)
                    max.Y = p.Y;

                if (p.Z < min.Z)
                    min.Z = p.Z;
                if (p.Z > max.Z)
                    max.Z = p.Z;
            }
        }

        /// <summary>
        /// Oriented Bounding Box
        /// </summary>
        public struct OBB
        {
            /// <summary>
            /// Creates an OBB give its center, axes, and half-widths along each axis
            /// </summary>
            /// <param name="center">Center of OBB</param>
            /// <param name="axes">Axes of OBB</param>
            /// <param name="extents">Half-widths of OBB along each axis</param>
            public OBB(Vector3 center, Vector3[] axes, float[] extents)
            {
                Debug.Assert(axes.Length == 3, "Wrong number of axes for bounding box");
                Debug.Assert(extents.Length == 3, "Wrong number of extents for bounding box");
                c = center;
                u1 = axes[0]; u2 = axes[1]; u3 = axes[2];
                e1 = extents[0]; e2 = extents[1]; e3 = extents[2];
            }

            /// <summary>
            /// Creates an OBB give its center, axes, and half-widths along each axis
            /// </summary>
            public OBB(Vector3 center, Vector3 u1, Vector3 u2, Vector3 u3, float e1, float e2, float e3)
            {
                Debug.AssertUnitVector(u1, "OBB axis not a unit vector");
                Debug.AssertUnitVector(u2, "OBB axis not a unit vector");
                Debug.AssertUnitVector(u3, "OBB axis not a unit vector");
                //Debug.Assert(Vector3.Dot(u1, u2) > 0.1f, "Axes of OBB are not perpendicular");
                //Debug.Assert(Vector3.Dot(u1, u3) > 0.1f, "Axes of OBB are not perpendicular");
                //Debug.Assert(Vector3.Dot(u3, u2) > 0.1f, "Axes of OBB are not perpendicular");
                c = center;
                this.u1 = u1;
                this.u2 = u2;
                this.u3 = u3;
                this.e1 = e1;
                this.e2 = e2;
                this.e3 = e3;
            }

            /// <summary>
            /// center point
            /// </summary>
            public Vector3 c;
            /// <summary>
            /// Local x-, y-, and z- axes
            /// </summary>
            public Vector3 u1, u2, u3;
            /// <summary>
            /// Positive half-width extents of OBB along each axis
            /// </summary>
            public float e1, e2, e3;

            /// <summary>
            /// True if point is contained within AABB
            /// </summary>
            public bool ContainsPoint(Vector3 p)
            {
                Vector3 offset = p-c;
                return Math.Abs(Vector3.Dot(offset, u1)) < e1
                    && Math.Abs(Vector3.Dot(offset, u2)) < e2
                    && Math.Abs(Vector3.Dot(offset, u3)) < e3;
            }

            /// <summary>
            /// True if point is contained within AABB.  If true, outputs contact normal and penetration depth;
            /// </summary>
            public bool TestPenetrationPoint(Vector3 p, ref Vector3 contactNormal, ref float penetrationDepth)
            {
                Vector3 offset = p - c;
                float p1 = Vector3.Dot(offset, u1);
                float p2 = Vector3.Dot(offset, u2);
                float p3 = Vector3.Dot(offset, u3);
                float a1 = Math.Abs(p1);
                float a2 = Math.Abs(p2);
                float a3 = Math.Abs(p3);

                if (a1 >= e1 || a2 >= e2 || a3 >= e3)
                    // Fast path - no intersection
                    return false;

                if (p1 >= 0)
                {
                    contactNormal = u1;
                    penetrationDepth = e1 - p1;
                }
                else
                {
                    contactNormal = -u1;
                    penetrationDepth = e1 + p1;
                }

                if (e2 - a2 < penetrationDepth)
                {
                    if (p2 >= 0)
                    {
                        contactNormal = u2;
                        penetrationDepth = e2 - p2;
                    }
                    else
                    {
                        contactNormal = -u2;
                        penetrationDepth = e2 + p2;
                    }
                }

                if (e3 - a3 < penetrationDepth)
                {
                    if (p3 >= 0)
                    {
                        contactNormal = u3;
                        penetrationDepth = e3 - p3;
                    }
                    else
                    {
                        contactNormal = -u3;
                        penetrationDepth = e3 + p3;
                    }
                }

                return true;
            }

            /// <summary>
            /// True if sphere is contained within AABB.  If true, outputs contact normal and penetration depth;
            /// </summary>
            public bool TestPenetrationSphere(Sphere s, ref Vector3 contactNormal, ref float penetrationDepth)
            {
                Vector3 offset = s.c - c;               // Offset sphere center to OBB center
                float p1 = Vector3.Dot(offset, u1);     // Offset along u1
                float p2 = Vector3.Dot(offset, u2);     // Offset along u2
                float p3 = Vector3.Dot(offset, u3);     // Offset along u3
                float a1 = Math.Abs(p1);                // Dist between centers along u1
                float a2 = Math.Abs(p2);                // Dist between centers along u2
                float a3 = Math.Abs(p3);                // Dist between centers along u2
                float radius = s.r;

                // Now do a separating axis test
                bool u1Separates = a1 >= e1 + radius;
                bool u2Separates = a2 >= e2 + radius;
                bool u3Separates = a3 >= e3 + radius;

                // If there's some axis along which they don't intersect, return
                if (u1Separates || u2Separates || u3Separates)
                    // Fast path - no intersection
                    return false;

                //
                // Separating axis test failed, so we want to find the axis
                // along which the penetration occured.  This is the axis
                // with *minimal* penetration because that's the axis along 
                // which we want to move the objects to produce a separation.
                //

                // Start with u1
                if (p1 >= 0)
                {
                    contactNormal = u1;
                    penetrationDepth = e1 - p1;
                }
                else
                {
                    contactNormal = -u1;
                    penetrationDepth = e1 + p1;
                }

                // Now check if u2 had *smaller* penetration
                if (e2-a2 < penetrationDepth)
                {
                    if (p2 >= 0)
                    {
                        contactNormal = u2;
                        penetrationDepth = e2 - p2;
                    }
                    else
                    {
                        contactNormal = -u2;
                        penetrationDepth = e2 + p2;
                    }
                }

                // And now u3
                if (e3-a3 < penetrationDepth)
                {
                    if (p3 >= 0)
                    {
                        contactNormal = u3;
                        penetrationDepth = e3 - p3;
                    }
                    else
                    {
                        contactNormal = -u3;
                        penetrationDepth = e3 + p3;
                    }
                }

                penetrationDepth += radius;

                return true;
            }        
        }

        /// <summary>
        /// Rectangle (i.e. planar surface) in 3-space.
        /// </summary>
        public struct Rect
        {
            /// <summary>
            /// Creates a 3D rectangle from its center, axes, and half-widths
            /// </summary>
            public Rect(Vector3 center, Vector3[] axes, float[] extents)
            {
                Debug.Assert(axes.Length == 2, "Wrong number of axes");
                Debug.Assert(extents.Length == 2, "Wrong number of extents");
                c = center;
                u = axes;
                e = extents;
            }
            /// <summary>
            /// center point of rectangle
            /// </summary>
            public Vector3 c;
            /// <summary>
            /// unit vectors determining local x and y axes for the rectangle
            /// </summary>
            public Vector3[] u;
            /// <summary>
            /// the halfwidth extents of the rectangle along the axes
            /// </summary>
            public float[] e;
        }

        /// <summary>
        ///  A sphere-capped cylinder
        /// </summary>
        [DebuggerDisplay("Capsule({a}, {b}, {r})")]
        public struct Capsule
        {
            /// <summary>
            /// Creates a capsule with specified radius and medial line segment endpoints
            /// </summary>
            public Capsule(Vector3 a, Vector3 b, float radius)
            {
                this.a = a;
                this.b = b;
                r = radius;
            }
            /// <summary>
            /// Center of one of the spheres
            /// </summary>
            public Vector3 a;      // Medial line segment start point
            /// <summary>
            /// Center of the other sphere
            /// </summary>
            public Vector3 b;      // Medial line segment end point
            /// <summary>
            /// Radius of the spheres/cylinder
            /// </summary>
            public float r;      // Radius
        };

#if NotYetPorted
/// <summary>
/// Region R = { x | (x - [a + u[0]*s + u[1]*t])^2 <= r }, 0 <= s,t <= 1
/// </summary>
public struct Lozenge {
    public Vector3 a;      // Origin
    public Vector3[] u;  // The two edges axes of the rectangle
    public float r;      // Radius
};
        /// <summary>
        /// A 3-space polygon
        /// </summary>
        public class Polygon
        {
            public Polygon(Vector3[] vertices)
            {
                v = vertices;
            }
            /// <summary>
            /// Vertices of the polygon
            /// </summary>
            public Vector3[] v;
            /// <summary>
            /// Number of vertices in the polygon
            /// </summary>
            public int numVerts
            {
                get
                {
                    return v.Length;
                }
            }
        }
#endif

        /// <summary>
        /// 
        /// </summary>
        public struct Triangle
        {
            /// <summary>
            /// Plane equation for triangle plane
            /// </summary>
            public Plane p;
            /// <summary>
            /// When evaluated gives barycentric weight u (for vertex A)
            /// </summary>
            public Plane edgePlaneBC;
            /// <summary>
            /// When evaluated gives barycentric weight v (for vertex B)
            /// </summary>
            public Plane edgePlaneCA;
        }


#if notdef
        Vector3 ClosestPtPointPlaneUnnormalized(Vector3 q, Plane p)
        {
            float t = (Vector3.Dot(p.n, q) - p.d) / Vector3.Dot(p.n, p.n);
            return q - t * p.n;
        }
#endif

        /// <summary>
        /// Returns the point on the specified plane that is closest to the specified point.
        /// </summary>
        public static Vector3 ClosestPtPointPlane(Vector3 q, Plane p)
        {
            float t = Vector3.Dot(p.n, q) - p.d;
            return q - t * p.n;
        }

        /// <summary>
        /// Distance from the point to the plane.
        /// </summary>
        public static float DistPointPlane(Vector3 q, Plane p)
        {
            // return Vector3.Dot(q, p.n) - p.d; if plane equation normalized (||p.n||==1)
            return (Vector3.Dot(p.n, q) - p.d) / Vector3.Dot(p.n, p.n);
        }

        /// <summary>
        /// Given segment ab and Vector3 c, computes closest Vector3 d on ab.
        /// Also returns t for the position of d, d(t) = a + t*(b - a)
        /// </summary>
        public static void ClosestPtPointSegment(Vector3 c, Vector3 a, Vector3 b, out float t, out Vector3 d)
        {
            Vector3 ab = b - a;
            // Project c onto ab, computing parameterized position d(t) = a + t*(b - a)
            t = Vector3.Dot(c - a, ab) / Vector3.Dot(ab, ab);
            // If outside segment, clamp t (and therefore d) to the closest endVector3
            if (t < 0.0f) t = 0.0f;
            if (t > 1.0f) t = 1.0f;
            // Compute projected position from the clamped t
            d = a + t * ab;
        }

#if notdef
        // Given segment ab and Vector3 c, computes closest Vector3 d on ab.
        // Also returns t for the parametric position of d, d(t) = a + t*(b - a)
        void ClosestPtPointSegment(Vector3 c, Vector3 a, Vector3 b, out float t, out Vector3 d)
        {
            Vector ab = b - a;
            // Project c onto ab, but deferring divide by Vector3.Dot(ab, ab)
            t = Vector3.Dot(c - a, ab);
            if (t <= 0.0f)
            {
                // c projects outside the [a,b] interval, on the a side; clamp to a
                t = 0.0f;
                d = a;
            }
            else
            {
                float denom = Vector3.Dot(ab, ab); // Always nonnegative since denom = ||ab||^2
                if (t >= denom)
                {
                    // c projects outside the [a,b] interval, on the b side; clamp to b
                    t = 1.0f;
                    d = b;
                }
                else
                {
                    // c projects inside the [a,b] interval; must do deferred divide now
                    t = t / denom;
                    d = a + t * ab;
                }
            }
        }
#endif

        /// <summary>
        /// Returns the squared distance between Vector3 c and segment ab
        /// </summary>
        public static float SqDistPointSegment(Vector3 a, Vector3 b, Vector3 c)
        {
            Vector3 ab = b - a, ac = c - a, bc = c - b;
            float e = Vector3.Dot(ac, ab);
            // Handle cases where c projects outside ab
            if (e <= 0.0f) return Vector3.Dot(ac, ac);
            float f = Vector3.Dot(ab, ab);
            if (e >= f) return Vector3.Dot(bc, bc);
            // Handle case where c projects onto ab
            return Vector3.Dot(ac, ac) - e * e / f;
        }

        /// <summary>
        /// Given Vector3 p, return the Vector3 q on or in AABB b, that is closest to p
        /// </summary>
        public static void ClosestPtPointAABB(Vector3 p, AABB b, out Vector3 q)
        {
            float v = p.X;
            if (v < b.min.X) v = b.min.X; // v = max(v, b.min[i])
            if (v > b.max.X) v = b.max.X; // v = min(v, b.max[i])
            q.X = v;

            v = p.Y;
            if (v < b.min.Y) v = b.min.Y; // v = max(v, b.min[i])
            if (v > b.max.Y) v = b.max.Y; // v = min(v, b.max[i])
            q.Y = v;

            v = p.Z;
            if (v < b.min.Z) v = b.min.Z; // v = max(v, b.min.Z)
            if (v > b.max.Z) v = b.max.Z; // v = min(v, b.max.Z)
            q.Z = v;
        }

        /// <summary>
        /// Squared distance between a Vector3 p and an AABB b
        /// </summary>
        public static float SqDistPointAABB(Vector3 p, AABB b)
        {
            float sqDist = 0.0f;
            // For each axis count any excess distance outside box extents
            float v = p.X;
            if (v < b.min.X) sqDist += (b.min.X - v) * (b.min.X - v);
            if (v > b.max.X) sqDist += (v - b.max.X) * (v - b.max.X);

            v = p.Y;
            if (v < b.min.Y) sqDist += (b.min.Y - v) * (b.min.Y - v);
            if (v > b.max.Y) sqDist += (v - b.max.Y) * (v - b.max.Y);

            v = p.Z;
            if (v < b.min.Z) sqDist += (b.min.Z - v) * (b.min.Z - v);
            if (v > b.max.Z) sqDist += (v - b.max.Z) * (v - b.max.Z);

            return sqDist;
        }

        /// <summary>
        /// Given Vector3 p, return Vector3 q on (or in) OBB b, closest to p
        /// </summary>
        public static void ClosestPtPointOBB(Vector3 p, OBB b, out Vector3 q)
        {
            Vector3 d = p - b.c;
            // Start result at center of box; make steps from there
            q = b.c;
            // For each OBB axis...
                // ...project d onto that axis to get the distance
                // along the axis of d from the box center
                float dist = Vector3.Dot(d, b.u1);
                // If distance farther than the box extents, clamp to the box
                if (dist > b.e1) dist = b.e1;
                if (dist < -b.e1) dist = -b.e1;
                // Step that distance along the axis to get world coordinate
                q += dist * b.u1;

                // ...project d onto that axis to get the distance
                // along the axis of d from the box center
                dist = Vector3.Dot(d, b.u2);
                // If distance farther than the box extents, clamp to the box
                if (dist > b.e2) dist = b.e2;
                if (dist < -b.e2) dist = -b.e2;
                // Step that distance along the axis to get world coordinate
                q += dist * b.u2;

                // ...project d onto that axis to get the distance
                // along the axis of d from the box center
                dist = Vector3.Dot(d, b.u3);
                // If distance farther than the box extents, clamp to the box
                if (dist > b.e3) dist = b.e3;
                if (dist < -b.e3) dist = -b.e3;
                // Step that distance along the axis to get world coordinate
                q += dist * b.u3;
        }

#if notdef
        /// <summary>
        /// Computes the square distance between Vector3 p and OBB b
        /// </summary>
        public float SqDistPointOBB(Vector3 p, OBB b)
        {
            Vector3 closest;
            ClosestPtVector3OBB(p, b, closest);
            float sqDist = Vector3.Dot(closest - p, closest - p);
            return sqDist;
        }

        /// <summary>
        /// Computes the square distance between Vector3 p and OBB b
        /// </summary>
        public static float SqDistPointOBB(Vector3 p, OBB b)
        {
            Vector3 v = p - b.c;
            float sqDist = 0.0f;
            for (int i = 0; i < 3; i++)
            {
                // Project vector from box center to p on each axis, getting the distance
                // of p along that axis, and count any excess distance outside box extents
                float d = Vector3.Dot(v, b.u[i]), excess = 0.0f;
                if (d < -b.e[i])
                    excess = d + b.e[i];
                else if (d > b.e[i])
                    excess = d - b.e[i];
                sqDist += excess * excess;
            }
            return sqDist;
        }
#endif

        /// <summary>
        /// Given Vector3 p, return Vector3 q on (or in) Rect r, closest to p
        /// </summary>
        static void ClosestPtPointRect(Vector3 p, Rect r, out Vector3 q)
        {
            Vector3 d = p - r.c;
            // Start result at center of rect; make steps from there
            q = r.c;
            // For each rect axis...
            for (int i = 0; i < 2; i++)
            {
                // ...project d onto that axis to get the distance
                // along the axis of d from the rect center
                float dist = Vector3.Dot(d, r.u[i]);
                // If distance farther than the rect extents, clamp to the rect
                if (dist > r.e[i]) dist = r.e[i];
                if (dist < -r.e[i]) dist = -r.e[i];
                // Step that distance along the axis to get world coordinate
                q += dist * r.u[i];
            }
        }

        /// <summary>
        /// Return Vector3 q on (or in) rect (specified by a, b, and c), closest to given Vector3 p
        /// </summary>
        public static void ClosestPtPointRect(Vector3 p, Vector3 a, Vector3 b, Vector3 c, out Vector3 q)
        {
            Vector3 ab = b - a; // vector aVector3.Cross rect
            Vector3 ac = c - a; // vector down rect
            Vector3 d = p - a;
            // Start result at top-left corner of rect; make steps from there
            q = a;
            // Clamp p (projection of p to plane of r) to rectangle in the aVector3.Cross direction
            float dist = Vector3.Dot(d, ab);
            float maxdist = Vector3.Dot(ab, ab);
            if (dist >= maxdist)
                q += ab;
            else if (dist > 0.0f)
                q += (dist / maxdist) * ab;
            // Clamp p (projection of p to plane of r) to rectangle in the down direction
            dist = Vector3.Dot(d, ac);
            maxdist = Vector3.Dot(ac, ac);
            if (dist >= maxdist)
                q += ac;
            else if (dist > 0.0f)
                q += (dist / maxdist) * ac;
        }

        /// <summary>
        /// Returns the closest point to p lying on the triangle a, b, c.
        /// </summary>
        public static Vector3 ClosestPtPointTriangle(Vector3 p, Vector3 a, Vector3 b, Vector3 c)
        {
            // Check if P in vertex region outside A
            Vector3 ab = b - a;
            Vector3 ac = c - a;
            Vector3 ap = p - a;
            float d1 = Vector3.Dot(ab, ap);
            float d2 = Vector3.Dot(ac, ap);
            if (d1 <= 0.0f && d2 <= 0.0f) return a; // barycentric coordinates (1,0,0)

            // Check if P in vertex region outside B
            Vector3 bp = p - b;
            float d3 = Vector3.Dot(ab, bp);
            float d4 = Vector3.Dot(ac, bp);
            if (d3 >= 0.0f && d4 <= d3) return b; // barycentric coordinates (0,1,0)

            // Check if P in edge region of AB, if so return projection of P onto AB
            float vc = d1 * d4 - d3 * d2;
            if (vc <= 0.0f && d1 >= 0.0f && d3 <= 0.0f)
            {
                float v = d1 / (d1 - d3);
                return a + v * ab; // barycentric coordinates (1-v,v,0)
            }

            // Check if P in vertex region outside C
            Vector3 cp = p - c;
            float d5 = Vector3.Dot(ab, cp);
            float d6 = Vector3.Dot(ac, cp);
            if (d6 >= 0.0f && d5 <= d6) return c; // barycentric coordinates (0,0,1)

            // Check if P in edge region of AC, if so return projection of P onto AC
            float vb = d5 * d2 - d1 * d6;
            if (vb <= 0.0f && d2 >= 0.0f && d6 <= 0.0f)
            {
                float w = d2 / (d2 - d6);
                return a + w * ac; // barycentric coordinates (1-w,0,w)
            }

            // Check if P in edge region of BC, if so return projection of P onto BC
            float va = d3 * d6 - d5 * d4;
            if (va <= 0.0f && (d4 - d3) >= 0.0f && (d5 - d6) >= 0.0f)
            {
                float w = (d4 - d3) / ((d4 - d3) + (d5 - d6));
                return b + w * (c - b); // barycentric coordinates (0,1-w,w)
            }

            // P inside face region. Compute Q through its barycentric coordinates (u,v,w)
            float denom = 1.0f / (va + vb + vc);
            float v1 = vb * denom;
            float w1 = vc * denom;
            return a + ab * v1 + ac * w1; // = u*a + v*b + w*c, u = va * denom = 1.0f - v - w
        }

        /// <summary>
        /// Returns the closest point to p on the tetrahedron ABCD 
        /// </summary>
        public static Vector3 ClosestPtPtTetrahedron(Vector3 p, Vector3 a, Vector3 b, Vector3 c, Vector3 d)
        {
            // Start out assuming Vector3 inside all halfspaces, so closest to itself
            Vector3 closestPt = p;
            float bestSqDist = Single.MaxValue;
            // If Vector3 outside face abc then compute closest Vector3 on abc
            if (PointOutsideOfPlane(p, a, b, c))
            {
                Vector3 q = ClosestPtPointTriangle(p, a, b, c);
                float sqDist = Vector3.Dot(q - p, q - p);
                // Update best closest Vector3 if (squared) distance is less than current best
                if (sqDist < bestSqDist) { bestSqDist = sqDist; closestPt = q; }
            }
            // Repeat test for face acd
            if (PointOutsideOfPlane(p, a, c, d))
            {
                Vector3 q = ClosestPtPointTriangle(p, a, c, d);
                float sqDist = Vector3.Dot(q - p, q - p);
                if (sqDist < bestSqDist) { bestSqDist = sqDist; closestPt = q; }
            }
            // Repeat test for face adb
            if (PointOutsideOfPlane(p, a, d, b))
            {
                Vector3 q = ClosestPtPointTriangle(p, a, d, b);
                float sqDist = Vector3.Dot(q - p, q - p);
                if (sqDist < bestSqDist) { bestSqDist = sqDist; closestPt = q; }
            }
            // Repeat test for face bdc
            if (PointOutsideOfPlane(p, b, d, c))
            {
                Vector3 q = ClosestPtPointTriangle(p, b, d, c);
                float sqDist = Vector3.Dot(q - p, q - p);
                if (sqDist < bestSqDist) { bestSqDist = sqDist; closestPt = q; }
            }
            return closestPt;
        }

        /// <summary>
        /// Test if Vector3 p lies outside plane through abc
        /// </summary>
        public static bool PointOutsideOfPlane(Vector3 p, Vector3 a, Vector3 b, Vector3 c)
        {
            return Vector3.Dot(p - a, Vector3.Cross(b - a, c - a)) >= 0.0f; // [AP AB AC] >= 0
        }

        /// <summary>
        /// Test if Vector3 p and d lie on opposite sides of plane through abc
        /// </summary>
        public static bool PointOutsideOfPlane(Vector3 p, Vector3 a, Vector3 b, Vector3 c, Vector3 d)
        {
            float signp = Vector3.Dot(p - a, Vector3.Cross(b - a, c - a)); // [AP AB AC]
            float signd = Vector3.Dot(d - a, Vector3.Cross(b - a, c - a)); // [AD AB AC]
            // Vector3s on opposite sides if expression signs are opposite
            return signp * signd < 0.0f;
        }

        /// <summary>
        /// Clamp n to lie within the range [min, max]
        /// </summary>
        static float Clamp(float n, float min, float max)
        {
            if (n < min) return min;
            if (n > max) return max;
            return n;
        }

        const float EPSILON = 0.00001f;

        /// <summary>
        /// Computes closest Vector3s C1 and C2 of S1(s)=P1+s*(Q1-P1) and
        /// S2(t)=P2+t*(Q2-P2), returning s and t. Function result is squared
        /// distance between between S1(s) and S2(t)
        /// </summary>
        public static float ClosestPtSegmentSegment(Vector3 p1, Vector3 q1, Vector3 p2, Vector3 q2,
                                      out float s, out float t, out Vector3 c1, out Vector3 c2)
        {
            Vector3 d1 = q1 - p1; // Direction vector of segment S1
            Vector3 d2 = q2 - p2; // Direction vector of segment S2
            Vector3 r = p1 - p2;
            float a = Vector3.Dot(d1, d1); // Squared length of segment S1, always nonnegative
            float e = Vector3.Dot(d2, d2); // Squared length of segment S2, always nonnegative
            float f = Vector3.Dot(d2, r);

            // Check if either or both segments degenerate into Vector3s
            if (a <= EPSILON && e <= EPSILON)
            {
                // Both segments degenerate into Vector3s
                s = t = 0.0f;
                c1 = p1;
                c2 = p2;
                return Vector3.Dot(c1 - c2, c1 - c2);
            }
            if (a <= EPSILON)
            {
                // First segment degenerates into a Vector3
                s = 0.0f;
                t = f / e; // s = 0 => t = (b*s + f) / e = f / e
                t = Clamp(t, 0.0f, 1.0f);
            }
            else
            {
                float c = Vector3.Dot(d1, r);
                if (e <= EPSILON)
                {
                    // Second segment degenerates into a Vector3
                    t = 0.0f;
                    s = Clamp(-c / a, 0.0f, 1.0f); // t = 0 => s = (b*t - c) / a = -c / a
                }
                else
                {
                    // The general nondegenerate case starts here
                    float b = Vector3.Dot(d1, d2);
                    float denom = a * e - b * b; // Always nonnegative

                    // If segments not parallel, compute closest Vector3 on L1 to L2, and
                    // clamp to segment S1. Else pick arbitrary s (here 0)
                    if (denom != 0.0f)
                    {
                        s = Clamp((b * f - c * e) / denom, 0.0f, 1.0f);
                    }
                    else s = 0.0f;

                    float tnom = b * s + f;
                    if (tnom < 0.0f)
                    {
                        t = 0.0f;
                        s = Clamp(-c / a, 0.0f, 1.0f);
                    }
                    else if (tnom > e)
                    {
                        t = 1.0f;
                        s = Clamp((b - c) / a, 0.0f, 1.0f);
                    }
                    else
                    {
                        t = tnom / e;
                    }
                }
            }

            c1 = p1 + d1 * s;
            c2 = p2 + d2 * t;
            return Vector3.Dot(c1 - c2, c1 - c2);
        }

        /// <summary>
        /// Returns 2 times the signed triangle area. The result is positive if
        /// abc is ccw, negative if abc is cw, zero if abc is degenerate.
        /// </summary>
        static float Signed2DTriArea(Vector2 a, Vector2 b, Vector2 c)
        {
            return (a.X - c.X) * (b.Y - c.Y) - (a.Y - c.Y) * (b.X - c.X);
        }

        /// <summary>
        /// Test if segments ab and cd overlap. If they do, compute and return
        /// intersection t value along ab and intersection position p
        /// </summary>
        static public bool Test2DSegmentSegment(Vector2 a, Vector2 b, Vector2 c, Vector2 d, ref float t, ref Vector2 p)
        {
            // Sign of areas correspond to which side of ab Vector3s c and d are
            float a1 = Signed2DTriArea(a, b, d); // Compute winding of abd (+ or -)
            float a2 = Signed2DTriArea(a, b, c); // To intersect, must have sign opposite of a1

            // If c and d are on different sides of ab, areas have different signs
            if (a1 * a2 < 0.0f)
            {
                // Compute signs for a and b with respect to segment cd
                float a3 = Signed2DTriArea(c, d, a); // Compute winding of cda (+ or -)
                // Since area is constant a1-a2 = a3-a4, or a4=a3+a2-a1
                //      float a4 = Signed2DTriArea(c, d, b); // Must have opposite sign of a3
                float a4 = a3 + a2 - a1;
                // Vector3s a and b on different sides of cd if areas have different signs
                if (a3 * a4 < 0.0f)
                {
                    // Segments intersect. Find intersection Vector3 along L(t)=a+t*(b-a).
                    // Given height h1 of a over cd and height h2 of b over cd,
                    // t = h1 / (h1 - h2) = (b*h1/2) / (b*h1/2 - b*h2/2) = a3 / (a3 - a4),
                    // where b (the base of the triangles cda and cdb, i.e., the length
                    // of cd) cancels out.
                    t = a3 / (a3 - a4);
                    p = a + t * (b - a);
                    return true;
                }
            }

            // Segments not intersecting (or collinear)
            return false;
        }

        /// <summary>
        /// Determine whether plane p intersects sphere s
        /// </summary>
        static public bool TestSpherePlane(Sphere s, Plane p)
        {
            // For a normalized plane (|p.n| = 1), evaluating the plane equation
            // for a Vector3 gives the signed distance of the Vector3 to the plane
            float dist = Vector3.Dot(s.c, p.n) - p.d;
            // If sphere center within +/-radius from plane, plane intersects sphere
            return Math.Abs(dist) <= s.r;
        }

        /// <summary>
        /// Determine whether sphere s fully behind (inside negative halfspace of) plane p
        /// </summary>
        static public bool InsideSpherePlane(Sphere s, Plane p)
        {
            float dist = Vector3.Dot(s.c, p.n) - p.d;
            return dist < -s.r;
        }

        /// <summary>
        /// Determine whether sphere s intersects negative halfspace of plane p
        /// </summary>
        static public bool TestSphereHalfspace(Sphere s, Plane p)
        {
            float dist = Vector3.Dot(s.c, p.n) - p.d;
            return dist <= s.r;
        }

#if notdef
        /// <summary>
        /// Test if OBB b intersects plane p
        /// </summary>
        static public bool TestOBBPlane(OBB b, Plane p)
        {
            // Compute the projection interval radius of b onto L(t) = b.c + t * p.n
            float r = b.e[0] * Math.Abs(Vector3.Dot(p.n, b.u[0])) +
                      b.e[1] * Math.Abs(Vector3.Dot(p.n, b.u[1])) +
                      b.e[2] * Math.Abs(Vector3.Dot(p.n, b.u[2]));
            // Compute distance of box center from plane
            float s = Vector3.Dot(p.n, b.c) - p.d;
            // Intersection occurs when distance s falls within [-r,+r] interval
            return Math.Abs(s) <= r;
        }
#endif

        /// <summary>
        /// Test if AABB b intersects plane p
        /// </summary>
        static public bool TestAABBPlane(AABB b, Plane p)
        {
            // These two lines not necessary with a (center, extents) AABB representation
            Vector3 c = (b.max + b.min) * 0.5f; // Compute AABB center
            Vector3 e = b.max - c; // Compute positive extents

            // Compute the projection interval radius of b onto L(t) = b.c + t * p.n
            float r = e.X * Math.Abs(p.n.X) + e.Y * Math.Abs(p.n.Y) + e.Z * Math.Abs(p.n.Z);
            // Compute distance of box center from plane
            float s = Vector3.Dot(p.n, c) - p.d;
            // Intersection occurs when distance s falls within [-r,+r] interval
            return Math.Abs(s) <= r;
        }

        /// <summary>
        /// Test whether two axis-aligned bounding boxes intersect.
        /// </summary>
        static public bool TestAABBAABB(AABB a, AABB b)
        {
            // Exit with no intersection if separated along an axis
            if (a.max.Z < b.min.Z || a.min.Z > b.max.Z) return false;
            if (a.max.Y < b.min.Y || a.min.Y > b.max.Y) return false;
            if (a.max.Z < b.min.Z || a.min.Z > b.max.Z) return false;
            // Overlapping on all axes means AABBs are intersecting
            return true;
        }

        /// <summary>
        /// Returns true if sphere s intersects AABB b, false otherwise
        /// </summary>
        static public bool TestSphereAABB(Sphere s, AABB b)
        {
            // Compute squared distance between sphere center and AABB
            float sqDist = SqDistPointAABB(s.c, b);

            // Sphere and AABB intersect if the (squared) distance
            // between them is less than the (squared) sphere radius
            return sqDist <= s.r * s.r;
        }


        /// <summary>
        /// Returns true if sphere s intersects AABB b, false otherwise.
        /// The Vector3 p on the AABB closest to the sphere center is also returned
        /// </summary>
        static public bool TestSphereAABB(Sphere s, AABB b, out Vector3 p)
        {
            // Find Vector3 p on AABB closest to sphere center
            ClosestPtPointAABB(s.c, b, out p);

            // Sphere and AABB intersect if the (squared) distance from sphere
            // center to Vector3 p is less than the (squared) sphere radius
            Vector3 v = p - s.c;
            return Vector3.Dot(v, v) <= s.r * s.r;
        }

        /// <summary>
        /// Returns true if sphere s intersects OBB b, false otherwise.
        /// The Vector3 p on the OBB closest to the sphere center is also returned
        /// </summary>
        public static bool TestSphereOBB(Sphere s, OBB b, out Vector3 p)
        {
            // Find Vector3 p on OBB closest to sphere center
            ClosestPtPointOBB(s.c, b, out p);

            // Sphere and OBB intersect if the (squared) distance from sphere
            // center to Vector3 p is less than the (squared) sphere radius
            Vector3 v = p - s.c;
            return Vector3.Dot(v, v) <= s.r * s.r;
        }


        /// <summary>
        /// Returns true if sphere s intersects triangle ABC, false otherwise.
        /// The Vector3 p on abc closest to the sphere center is also returned
        /// </summary>
        public static bool TestSphereTriangle(Sphere s, Vector3 a, Vector3 b, Vector3 c, out Vector3 p)
        {
            // Find Vector3 P on triangle ABC closest to sphere center
            p = ClosestPtPointTriangle(s.c, a, b, c);

            // Sphere and triangle intersect if the (squared) distance from sphere
            // center to Vector3 p is less than the (squared) sphere radius
            Vector3 v = p - s.c;
            return Vector3.Dot(v, v) <= s.r * s.r;
        }

#if NotPortedYet
        /// <summary>
        /// Test whether sphere s intersects polygon p
        /// </summary>
        public bool TestSpherePolygon(Sphere s, Polygon p)
        {
            // Compute normal for the plane of the polygon
            Vector3 n = Vector3.Normalize(Vector3.Cross(p.v[1] - p.v[0], p.v[2] - p.v[0]));
            // Compute the plane equation for p
            Plane m; m.n = n; m.d = -Vector3.Dot(n, p.v[0]);
            // No intersection if sphere not intersecting plane of polygon
            if (!TestSpherePlane(s, m)) return false;
            // Test to see if any one of the polygon edges pierces the sphere   
            for (int k = p.numVerts, i = 0, j = k - 1; i < k; j = i, i++)
            {
                float t;
                Vector3 q;
                // Test if edge (p.v[j], p.v[i]) intersects s
                if (IntersectRaySphere(p.v[j], p.v[i] - p.v[j], s, ref t, ref q) && t <= 1.0f)
                    return true;
            }
            // Test if the orthogonal projection q of the sphere center onto m is inside p
            Vector3 q1 = ClosestPtPointPlane(s.c, m);
            return PointInPolygon(q1, p);
        }


        public float TestTriangleAABB(Vector3 v0, Vector3 v1, Vector3 v2, AABB b)
        {
            float p0, p1, p2, r;

            // Compute box center and extents (if not already given in that format)
            Vector3 c = (b.min + b.max) * 0.5f;
            float e0 = (b.max.x - b.min.x) * 0.5f;
            float e1 = (b.max.y - b.min.y) * 0.5f;
            float e2 = (b.max.z - b.min.z) * 0.5f;

            // Translate triangle as conceptually moving AABB to origin
            v0 = v0 - c;
            v1 = v1 - c;
            v2 = v2 - c;

            // Compute edge vectors for triangle
            Vector3 f0 = v1 - v0, f1 = v2 - v1, f2 = v0 - v2;

            // Test axes a00..a22 (category 3)
            // Test axis a00
            p0 = v0.z * v1.y - v0.y * v1.z;
            p2 = v2.z * (v1.y - v0.y) - v2.y * (v1.z - v0.z);
            r = e1 * Math.Abs(f0.z) + e2 * Math.Abs(f0.y);
            if (Math.Max(-Math.Max(p0, p2), Math.Min(p0, p2)) > r) return false; // Axis is a separating axis

            // Repeat similar tests for remaining axes a01..a22
            ...


            // Test the three axes corresponding to the face normals of AABB b (category 1).
            // Exit if...
            // ... [-e0, e0] and [min(v0.x,v1.x,v2.x), max(v0.x,v1.x,v2.x)] do not overlap
            if (Max(v0.x, v1.x, v2.x) < -e0 || Min(v0.x, v1.x, v2.x) > e0) return false;
            // ... [-e1, e1] and [min(v0.y,v1.y,v2.y), max(v0.y,v1.y,v2.y)] do not overlap
            if (Max(v0.y, v1.y, v2.y) < -e1 || Min(v0.y, v1.y, v2.y) > e1) return false;
            // ... [-e2, e2] and [min(v0.z,v1.z,v2.z), max(v0.z,v1.z,v2.z)] do not overlap
            if (Max(v0.z, v1.z, v2.z) < -e2 || Min(v0.z, v1.z, v2.z) > e2) return false;

            // Test separating axis corresponding to triangle face normal (category 2)
            Plane p;
            p.n = Vector3.Cross(f0, f1);
            p.d = Vector3.Dot(p.n, v0);
            return TestAABBPlane(b, p);
        }
#endif

        /// <summary>
        /// Test whether the line segment AB intersects the plane P
        /// </summary>
        public static bool IntersectSegmentPlane(Vector3 a, Vector3 b, Plane p, ref float t, ref Vector3 q)
        {
            // Compute the t value for the directed line ab intersecting the plane
            Vector3 ab = b - a;
            t = (p.d - Vector3.Dot(p.n, a)) / Vector3.Dot(p.n, ab);

            // If t in [0..1] compute and return intersection Vector3
            if (t >= 0.0f && t <= 1.0f)
            {
                q = a + t * ab;
                return true;
            }
            // Else no intersection
            return false;
        }


        /// <summary>
        /// Intersect segment ab against plane of triangle def. If intersecting,
        /// return t value and position q of intersection
        /// </summary>
        public static bool IntersectSegmentPlane(Vector3 a, Vector3 b, Vector3 d, Vector3 e, Vector3 f,
                                  ref float t, ref Vector3 q)
        {
            Plane p;
            p.n = Vector3.Cross(e - d, f - d);
            p.d = Vector3.Dot(p.n, d);
            return IntersectSegmentPlane(a, b, p, ref t, ref q);
        }

        /// <summary>
        /// Intersects ray r = p + td, |d| = 1, with sphere s and, if intersecting,
        /// returns t value of intersection and intersection Vector3 q
        /// </summary>
        public static bool IntersectRaySphere(Vector3 p, Vector3 d, Sphere s, ref float t, ref Vector3 q)
        {
            Vector3 m = p - s.c;
            float b = Vector3.Dot(m, d);
            float c = Vector3.Dot(m, m) - s.r * s.r;
            // Exit if rs origin outside s (c > 0)and r Vector3ing away from s (b > 0)
            if (c > 0.0f && b > 0.0f) return false;
            float discr = b * b - c;
            // A negative discriminant corresponds to ray missing sphere
            if (discr < 0.0f) return false;
            // Ray now found to intersect sphere, compute smallest t value of intersection
            t = -b - (float)Math.Sqrt(discr);
            // If t is negative, ray started inside sphere so clamp t to zero
            if (t < 0.0f) t = 0.0f;
            q = p + t * d;
            return true;
        }

        /// <summary>
        /// Test if ray r = p + td intersects sphere s
        /// </summary>
        static bool TestRaySphere(Vector3 p, Vector3 d, Sphere s)
        {
            Vector3 m = p - s.c;
            float c = Vector3.Dot(m, m) - s.r * s.r;
            // If there is definitely at least one real root, there must be an intersection
            if (c <= 0.0f) return true;
            float b = Vector3.Dot(m, d);
            // Early exit if ray origin outside sphere and ray Vector3ing away from sphere
            if (b > 0.0f) return false;
            float disc = b * b - c;
            // A negative discriminant corresponds to ray missing sphere
            if (disc < 0.0f) return false;
            // Now ray must hit sphere
            return true;
        }


        /// <summary>
        /// Intersect ray R(t) = p + t*d against AABB a. When intersecting,
        /// return intersection distance tmin and Vector3 q of intersection
        /// </summary>
        public static bool IntersectRayAABB(Vector3 p, Vector3 d, AABB a, ref float tmin, ref Vector3 q)
        {
            tmin = 0.0f;          // set to -FLT_MAX to get first hit on line
            float tmax = Single.MaxValue; // set to max distance ray can travel (for segment)

            // For all three slabs
            if (Math.Abs(d.X) < EPSILON)
            {
                // Ray is parallel to slab. No hit if origin not within slab
                if (p.X < a.min.X || p.X > a.max.X) return false;
            }
            else
            {
                // Compute intersection t value of ray with near and far plane of slab
                float ood = 1.0f / d.X;
                float t1 = (a.min.X - p.X) * ood;
                float t2 = (a.max.X - p.X) * ood;
                // Make t1 be intersection with near plane, t2 with far plane
                if (t1 > t2)
                {
                    float temp = t1;
                    t1 = t2;
                    t2 = temp;
                }
                // Compute the intersection of slab intersections intervals
                tmin = Math.Max(tmin, t1);
                tmax = Math.Min(tmax, t2);
                // Exit with no collision as soon as slab intersection becomes empty
                if (tmin > tmax) return false;
            }

            if (Math.Abs(d.Y) < EPSILON)
            {
                // Ray is parallel to slab. No hit if origin not within slab
                if (p.Y < a.min.Y || p.Y > a.max.Y) return false;
            }
            else
            {
                // Compute intersection t value of ray with near and far plane of slab
                float ood = 1.0f / d.Y;
                float t1 = (a.min.Y - p.Y) * ood;
                float t2 = (a.max.Y - p.Y) * ood;
                // Make t1 be intersection with near plane, t2 with far plane
                if (t1 > t2)
                {
                    float temp = t1;
                    t1 = t2;
                    t2 = temp;
                }
                // Compute the intersection of slab intersections intervals
                tmin = Math.Max(tmin, t1);
                tmax = Math.Min(tmax, t2);
                // Exit with no collision as soon as slab intersection becomes empty
                if (tmin > tmax) return false;
            }

            if (Math.Abs(d.Z) < EPSILON)
            {
                // Ray is parallel to slab. No hit if origin not within slab
                if (p.Z < a.min.Z || p.Z > a.max.Z) return false;
            }
            else
            {
                // Compute intersection t value of ray with near and far plane of slab
                float ood = 1.0f / d.Z;
                float t1 = (a.min.Z - p.Z) * ood;
                float t2 = (a.max.Z - p.Z) * ood;
                // Make t1 be intersection with near plane, t2 with far plane
                if (t1 > t2)
                {
                    float temp = t1;
                    t1 = t2;
                    t2 = temp;
                }
                // Compute the intersection of slab intersections intervals
                tmin = Math.Max(tmin, t1);
                tmax = Math.Min(tmax, t2);
                // Exit with no collision as soon as slab intersection becomes empty
                if (tmin > tmax) return false;
            }


            // Ray intersects all 3 slabs. Return Vector3 (q) and intersection t value (tmin) 
            q = p + d * tmin;
            return true;
        }

        /// <summary>
        /// Test if segment specified by Vector3s p0 and p1 intersects AABB b
        /// </summary>
        public static bool TestSegmentAABB(Vector3 p0, Vector3 p1, AABB b)
        {
            Vector3 c = (b.min + b.max) * 0.5f; // Box center-Vector3
            Vector3 e = b.max - c;             // Box halflength extents
            Vector3 m = (p0 + p1) * 0.5f;       // Segment midVector3
            Vector3 d = p1 - m;                // Segment halflength vector
            m = m - c;                        // Translate box and segment to origin

            // Try world coordinate axes as separating axes
            float adx = Math.Abs(d.X);
            if (Math.Abs(m.X) > e.X + adx) return false;
            float ady = Math.Abs(d.Y);
            if (Math.Abs(m.Y) > e.Y + ady) return false;
            float adz = Math.Abs(d.Z);
            if (Math.Abs(m.Z) > e.Z + adz) return false;

            // Add in an epsilon term to counteract arithmetic errors when segment is
            // (near) parallel to a coordinate axis (see text for detail)
            adx += EPSILON; ady += EPSILON; adz += EPSILON;

            // Try Vector3.Cross products of segment direction vector with coordinate axes
            if (Math.Abs(m.Y * d.Z - m.Z * d.Y) > e.Y * adz + e.Z * ady) return false;
            if (Math.Abs(m.Z * d.X - m.X * d.Z) > e.X * adz + e.Z * adx) return false;
            if (Math.Abs(m.X * d.Y - m.Y * d.X) > e.X * ady + e.Y * adx) return false;

            // No separating axis found; segment must be overlapping AABB
            return true;
        }

        /// <summary>
        /// Given line pq and ccw triangle abc, return whether line pierces triangle. If
        /// so, also return the barycentric coordinates (u,v,w) of the intersection Vector3
        /// </summary>
        public static bool IntersectLineTriangle(Vector3 p, Vector3 q, Vector3 a, Vector3 b, Vector3 c,
                                             ref float u, ref float v, ref float w)
        {
            Vector3 pq = q - p;
            Vector3 pa = a - p;
            Vector3 pb = b - p;
            Vector3 pc = c - p;
            // Test if pq is inside the edges bc, ca and ab. Done by testing
            // that the signed tetrahedral volumes, computed using scalar triple
            // products, are all positive
            u = MathUtil.STriple(pq, pc, pb);
            if (u < 0.0f) return false;
            v = MathUtil.STriple(pq, pa, pc);
            if (v < 0.0f) return false;
            w = MathUtil.STriple(pq, pb, pa);
            if (w < 0.0f) return false;

            // Compute the barycentric coordinates (u, v, w) determining the
            // intersection Vector3 r, r = u*a + v*b + w*c
            float denom = 1.0f / (u + v + w);
            u *= denom;
            v *= denom;
            w *= denom; // w = 1.0f - u - v;
            return true;
        }

        /// <summary>
        /// Given line pq and ccw quadrilateral abcd, return whether the line
        /// pierces the triangle. If so, also return the Vector3 r of intersection
        /// </summary>
        public static bool IntersectLineQuad(Vector3 p, Vector3 q, Vector3 a, Vector3 b, Vector3 c, Vector3 d, ref Vector3 r)
        {
            Vector3 pq = q - p;
            Vector3 pa = a - p;
            Vector3 pb = b - p;
            Vector3 pc = c - p;
            // Determine which triangle to test against by testing against diagonal first
            Vector3 m = Vector3.Cross(pc, pq);
            float v = Vector3.Dot(pa, m); // MathUtil.STriple(pq, pa, pc);
            if (v >= 0.0f)
            {
                // Test intersection against triangle abc
                float u = -Vector3.Dot(pb, m); // MathUtil.STriple(pq, pc, pb);
                if (u < 0.0f) return false;
                float w = MathUtil.STriple(pq, pb, pa);
                if (w < 0.0f) return false;
                // Compute r, r = u*a + v*b + w*c, from barycentric coordinates (u, v, w)
                float denom = 1.0f / (u + v + w);
                u *= denom;
                v *= denom;
                w *= denom; // w = 1.0f - u - v;
                r = u * a + v * b + w * c;
            }
            else
            {
                // Test intersection against triangle dac
                Vector3 pd = d - p;
                float u = Vector3.Dot(pd, m); // MathUtil.STriple(pq, pd, pc);
                if (u < 0.0f) return false;
                float w = MathUtil.STriple(pq, pa, pd);
                if (w < 0.0f) return false;
                v = -v;
                // Compute r, r = u*a + v*d + w*c, from barycentric coordinates (u, v, w)
                float denom = 1.0f / (u + v + w);
                u *= denom;
                v *= denom;
                w *= denom; // w = 1.0f - u - v;
                r = u * a + v * d + w * c;
            }
            return true;
        }

        // Given segment pq and triangle abc, returns whether segment intersects
        // triangle and if so, also returns the barycentric coordinates (u,v,w)
        // of the intersection Vector3
        static bool IntersectSegmentTriangle(Vector3 p, Vector3 q, Vector3 a, Vector3 b, Vector3 c,
                                     ref float u, ref float v, ref float w, ref float t)
        {
            Vector3 ab = b - a;
            Vector3 ac = c - a;
            Vector3 qp = p - q;

            // Compute triangle normal. Can be precalculated or cached if
            // intersecting multiple segments against the same triangle
            Vector3 n = Vector3.Cross(ab, ac);

            // Compute denominator d. If d <= 0, segment is parallel to or Vector3s
            // away from triangle, so exit early
            float d = Vector3.Dot(qp, n);
            if (d <= 0.0f) return false;

            // Compute intersection t value of pq with plane of triangle. A ray
            // intersects iff 0 <= t. Segment intersects iff 0 <= t <= 1. Delay
            // dividing by d until intersection has been found to pierce triangle
            Vector3 ap = p - a;
            t = Vector3.Dot(ap, n);
            if (t < 0.0f) return false;
            if (t > d) return false; // For segment; exclude this code line for a ray test

            // Compute barycentric coordinate components and test if within bounds
            Vector3 e = Vector3.Cross(qp, ap);
            v = Vector3.Dot(ac, e);
            if (v < 0.0f || v > d) return false;
            w = -Vector3.Dot(ab, e);
            if (w < 0.0f || v + w > d) return false;

            // Segment/ray intersects triangle. Perform delayed division and
            // compute the last barycentric coordinate component
            float ood = 1.0f / d;
            t *= ood;
            v *= ood;
            w *= ood;
            u = 1.0f - v - w;
            return true;
        }

        /// <summary>
        /// Given segment pq and precomputed triangle tri, returns whether segment intersects
        /// triangle. If so, also returns the barycentric coordinates (u,v,w) of the
        /// intersection Vector3 s, and the parameterized intersection t value
        /// </summary>
        public static bool IntersectSegmentTriangle(Vector3 p, Vector3 q, Triangle tri,
                                     ref float u, ref float v, ref float w, ref float t, ref Vector3 s)
        {
            // Compute distance of p to triangle plane. Exit if p lies behind plane
            float distp = Vector3.Dot(p, tri.p.n) - tri.p.d;
            if (distp < 0.0f) return false;

            // Compute distance of q to triangle plane. Exit if q lies in front of plane
            float distq = Vector3.Dot(q, tri.p.n) - tri.p.d;
            if (distq >= 0.0f) return false;

            // Compute t value and Vector3 s of intersection with triangle plane
            float denom = distp - distq;
            t = distp / denom;
            s = p + t * (q - p);

            // Compute the barycentric coordinate u; exit if outside 0..1 range
            u = Vector3.Dot(s, tri.edgePlaneBC.n) - tri.edgePlaneBC.d;
            if (u < 0.0f || u > 1.0f) return false;
            // Compute the barycentric coordinate v; exit if negative
            v = Vector3.Dot(s, tri.edgePlaneCA.n) - tri.edgePlaneCA.d;
            if (v < 0.0f) return false;
            // Compute the barycentric coordinate w; exit if negative
            w = 1.0f - u - v;
            if (w < 0.0f) return false;

            // Segment intersects tri at distance t in position s (s = u*A + v*B + w*C)
            return true;
        }


        /// <summary>
        /// Intersect segment S(t)=sa+t(sb-sa), 0&lt;=t&lt;=1 against cylinder specified by p, q and r
        /// </summary>
        public static bool IntersectSegmentCylinder(Vector3 sa, Vector3 sb, Vector3 p, Vector3 q, float r, ref float t)
        {
            Vector3 d = q - p, m = sa - p, n = sb - sa;
            float md = Vector3.Dot(m, d);
            float nd = Vector3.Dot(n, d);
            float dd = Vector3.Dot(d, d);
            // Test if segment fully outside either endcap of cylinder
            if (md < 0.0f && md + nd < 0.0f) return false; // Segment outside p side of cylinder
            if (md > dd && md + nd > dd) return false;     // Segment outside q side of cylinder
            float nn = Vector3.Dot(n, n);
            float mn = Vector3.Dot(m, n);
            float a = dd * nn - nd * nd;
            float k = Vector3.Dot(m, m) - r * r;
            float c = dd * k - md * md;
            if (Math.Abs(a) < EPSILON)
            {
                // Segment runs parallel to cylinder axis
                if (c > 0.0f) return false; // a and thus the segment lie outside cylinder
                // Now known that segment intersects cylinder; figure out how it intersects
                if (md < 0.0f) t = -mn / nn; // Intersect segment against p endcap
                else if (md > dd) t = (nd - mn) / nn; // Intersect segment against q endcap
                else t = 0.0f; // a lies inside cylinder
                return true;
            }
            float b = dd * mn - nd * md;
            float discr = b * b - a * c;
            if (discr < 0.0f) return false; // No real roots; no intersection
            t = (-b - (float)Math.Sqrt(discr)) / a;
            if (t < 0.0f || t > 1.0f) return false; // Intersection lies outside segment
            if (md + t * nd < 0.0f)
            {
                // Intersection outside cylinder on p side
                if (nd <= 0.0f) return false; // Segment Vector3ing away from endcap
                t = -md / nd;
                // Keep intersection if Vector3.Dot(S(t) - p, S(t) - p) <= r^2
                return k + 2 * t * (mn + t * nn) <= 0.0f;
            }
            else if (md + t * nd > dd)
            {
                // Intersection outside cylinder on q side
                if (nd >= 0.0f) return false; // Segment Vector3ing away from endcap
                t = (dd - md) / nd;
                // Keep intersection if Vector3.Dot(S(t) - q, S(t) - q) <= r^2
                return k + dd - 2 * md + t * (2 * (mn - nd) + t * nn) <= 0.0f;
            }
            // Segment intersects cylinder between the end-caps; t is correct
            return true;
        }


        /// <summary>
        /// Intersect segment S(t)=A+t(B-A), 0&lt;=t&lt;=1 against convex polyhedron specified
        /// by the n halfspaces defined by the planes p[]. On exit tfirst and tlast
        /// define the intersection, if any
        /// </summary>
        public static bool IntersectSegmentPolyhedron(Vector3 a, Vector3 b, Plane[] p, int n,
                                       out float tfirst, out float tlast)
        {
            // Compute direction vector for the segment
            Vector3 d = b - a;
            // Set initial interval to being the whole segment. For a ray, tlast should be
            // set to +FLT_MAX. For a line, additionally tfirst should be set to -FLT_MAX
            tfirst = 0.0f;
            tlast = 1.0f;
            // Intersect segment against each plane 
            for (int i = 0; i < n; i++)
            {
                float denom = Vector3.Dot(p[i].n, d);
                float dist = p[i].d - Vector3.Dot(p[i].n, a);
                // Test if segment runs parallel to the plane
                if (denom == 0.0f)
                {
                    // If so, return no intersection if segment lies outside plane
                    if (dist > 0.0f) return false;
                }
                else
                {
                    // Compute parameterized t value for intersection with current plane
                    float t = dist / denom;
                    if (denom < 0.0f)
                    {
                        // When entering halfspace, update tfirst if t is larger
                        if (t > tfirst) tfirst = t;
                    }
                    else
                    {
                        // When exiting halfspace, update tlast if t is smaller
                        if (t < tlast) tlast = t;
                    }
                    // Exit with no intersection if intersection becomes empty
                    if (tfirst > tlast) return false;
                }
            }
            // A nonzero logical intersection, so the segment intersects the polyhedron
            return true;
        }

#if NotYetPorted
/// <summary>
/// Test if Vector3 p lies inside ccw-specified convex n-gon given by vertices v[]
/// </summary>
public bool PointInConvexPolygon(Vector3 p, int n, Vector3[] v)
{
    // Do binary search over polygon vertices to find the fan triangle
    // (v[0], v[low], v[high]) the Vector3 p lies within the near sides of
    int low = 0, high = n;
    do {
        int mid = (low + high) / 2;
        if (TriangleIsCCW(v[0], v[mid], p))
            low = mid;
        else
            high = mid;
    } while (low + 1 < high);

    // If Vector3 outside last (or first) edge, then it is not inside the n-gon
    if (low == 0 || high == n) return false;

    // p is inside the polygon if it is left of
    // the directed edge from v[low] to v[high]
    return TriangleIsCCW(v[low], v[high], p);
}

/// <summary>
/// Test if Vector3 P lies inside the counterclockwise triangle ABC
/// </summary>
bool PointInTriangle(Vector3 p, Vector3 a, Vector3 b, Vector3 c)
{
    // Translate Vector3 and triangle so that Vector3 lies at origin
    a -= p; b -= p; c -= p;
    // Compute normal vectors for triangles pab and pbc
    Vector3 u = Vector3.Cross(b, c);
    Vector3 v = Vector3.Cross(c, a);
    // Make sure they are both Vector3ing in the same direction
    if (Vector3.Dot(u, v) < 0.0f) return false;
    // Compute normal vector for triangle pca
    Vector3 w = Vector3.Cross(a, b);
    // Make sure it Vector3s in the same direction as the first two
    if (Vector3.Dot(u, w) < 0.0f) return false;
    // Otherwise P must be in (or on) the triangle
    return true;
}
#endif

        /// <summary>
        /// Test if Vector3 P lies inside the counterclockwise 3D triangle ABC
        /// </summary>
        public static bool PointInTriangle(Vector3 p, Vector3 a, Vector3 b, Vector3 c)
        {
            // Translate Vector3 and triangle so that Vector3 lies at origin
            a -= p; b -= p; c -= p;

            float ab = Vector3.Dot(a, b);
            float ac = Vector3.Dot(a, c);
            float bc = Vector3.Dot(b, c);
            float cc = Vector3.Dot(c, c);
            // Make sure plane normals for pab and pbc Vector3 in the same direction
            if (bc * ac - cc * ab < 0.0f) return false;
            // Make sure plane normals for pab and pca Vector3 in the same direction
            float bb = Vector3.Dot(b, b);
            if (ab * bc - ac * bb < 0.0f) return false;
            // Otherwise P must be in (or on) the triangle
            return true;
        }

        /// <summary>
        /// Compute the 2D pseudo Vector3.Cross product Vector3.Dot(Perp(u), v)
        /// </summary>
        static float Cross2D(Vector2 u, Vector2 v)
        {
            return u.Y * v.X - u.X * v.Y;
        }

        /// <summary>
        /// Test if 2D Vector3 P lies inside the counterclockwise 2D triangle ABC
        /// </summary>
        public static bool PointInTriangle(Vector2 p, Vector2 a, Vector2 b, Vector2 c)
        {
            // If P to the right of AB then outside triangle
            if (Cross2D(p - a, b - a) < 0.0f) return false;
            // If P to the right of BC then outside triangle
            if (Cross2D(p - b, c - b) < 0.0f) return false;
            // If P to the right of CA then outside triangle
            if (Cross2D(p - c, a - c) < 0.0f) return false;
            // Otherwise P must be in (or on) the triangle
            return true;
        }

#if NotYetPorted
/// <summary>
/// Test if 2D Vector3 P lies inside 2D triangle ABC
/// </summary>
public bool PointInTriangle2D(Vector2 p, Vector2 a, Vector2 b, Vector2 c)
{
    float pab = Cross2D(p - a, b - a);
    float pbc = Cross2D(p - b, c - b);
    // If P left of one of AB and BC and right of the other, not inside triangle
    if (!SameSign(pab, pbc)) return false;
    float pca = Cross2D(p - c, a - c);
    // If P left of one of AB and CA and right of the other, not inside triangle
    if (!SameSign(pab, pca)) return false;
    // P left or right of all edges, so must be in (or on) the triangle
    return true;
}
#endif

        /// <summary>
        /// Test if Vector3 p inside polyhedron given as the intersection volume of n halfspaces
        /// </summary>
        public static bool TestPointPolyhedron(Vector3 p, Plane[] h, int n)
        {
            for (int i = 0; i < n; i++)
            {
                // Exit with no containment if p ever found outside a halfspace
                if (DistPointPlane(p, h[i]) > 0.0f) return false;
            }
            // p inside all halfspaces, so p must be inside intersection volume
            return true;
        }

        /// <summary>
        /// Given planes p1 and p2, compute line L = p+t*d of their intersection.
        /// Return 0 if no such line exists
        /// </summary>
        public static bool IntersectPlanes(Plane p1, Plane p2, ref Vector3 p, ref Vector3 d)
        {
            // Compute direction of intersection line
            d = Vector3.Cross(p1.n, p2.n);

            // If d is (near) zero, the planes are parallel (and separated)
            // or coincident, so theyre not considered intersecting
            float denom = Vector3.Dot(d, d);
            if (denom < EPSILON) return false;

            // Compute Vector3 on intersection line
            p = Vector3.Cross(p1.d * p2.n - p2.d * p1.n, d) / denom;
            return true;
        }

        /// <summary>
        /// Compute the Vector3 p at which the three planes p1, p2, and p3 intersect (if at all)
        /// </summary>
        public static bool IntersectPlanes(Plane p1, Plane p2, Plane p3, ref Vector3 p)
        {
            Vector3 u = Vector3.Cross(p2.n, p3.n);
            float denom = Vector3.Dot(p1.n, u);
            if (Math.Abs(denom) < EPSILON) return false; // Planes do not intersect in a Vector3
            p = (p1.d * u + Vector3.Cross(p1.n, p3.d * p2.n - p2.d * p3.n)) / denom;
            return true;
        }

        /// <summary>
        /// Test whether sphere and capsule intersect
        /// </summary>
        public static bool TestSphereCapsule(Sphere s, Capsule capsule)
        {
            // Compute (squared) distance between sphere center and capsule line segment
            float dist2 = SqDistPointSegment(capsule.a, capsule.b, s.c);

            // If (squared) distance smaller than (squared) sum of radii, they collide
            float radius = s.r + capsule.r;
            return dist2 <= radius * radius;
        }

        /// <summary>
        /// Test whether two capsules intersect
        /// </summary>
        public static bool TestCapsuleCapsule(Capsule capsule1, Capsule capsule2)
        {
            // Compute (squared) distance between the inner structures of the capsules
            float s, t;
            Vector3 c1, c2;
            float dist2 = ClosestPtSegmentSegment(capsule1.a, capsule1.b,
                                                  capsule2.a, capsule2.b,
                                                  out s, out t, out c1, out c2);

            // If (squared) distance smaller than (squared) sum of radii, they collide
            float radius = capsule1.r + capsule2.r;
            return dist2 <= radius * radius;
        }

        /// <summary>
        /// Test collision between two capsules.  Returns true if capsules intersect.
        /// Also returns contact points and penetration depth if contact.  Otherwise,
        /// contact and depth variables are undefined.
        /// </summary>
        public static bool TestCollision(Capsule capsule1, Capsule capsule2,
                                            out Vector3 capsule1Contact, out Vector3 capsule2Contact, out Vector3 contactNormal, out float penetrationDepth)
        {
            // Compute (squared) distance between the inner structures of the capsules
            float s, t;
            float dist2 = ClosestPtSegmentSegment(capsule1.a, capsule1.b,
                                                  capsule2.a, capsule2.b,
                                                  out s, out t,
                                                  out capsule1Contact, out capsule2Contact);

            // If (squared) distance smaller than (squared) sum of radii, they collide
            float radius = capsule1.r + capsule2.r;
            if (dist2 <= radius * radius)
            {
                penetrationDepth = (radius - (float)Math.Sqrt(dist2))*0.5f;
                contactNormal = capsule2Contact - capsule1Contact;
                MathUtil.SafeNormalize(ref contactNormal);
                capsule1Contact += capsule1.r * contactNormal;
                capsule2Contact -= capsule2.r * contactNormal;

                return true;
            }
            else
            {
                contactNormal = capsule1Contact = capsule2Contact = Vector3.Zero;
                penetrationDepth = 0;
                return false;
            }
        }

        /// <summary>
        /// Test collision between two spheres
        /// </summary>
        public static bool TestCollision(Sphere s1, Sphere s2,
                                         out Vector3 s1Contact, out Vector3 s2Contact, out Vector3 contactNormal,
                                         out float penetrationDepth)
        {
            Vector3 offset = s2.c-s1.c;
            float dist = offset.Length();
            float clearance = dist - (s1.r + s2.r);

            if (clearance <= 0)
            {
                if (dist < EPSILON)
                {
                    s1Contact = s1.c;
                    s2Contact = s2.c;
                    contactNormal = Vector3.Zero;
                }
                else
                {
                    contactNormal = offset / dist;
                    s1Contact = s1.c + contactNormal*s1.r;
                    s2Contact = s2.c - contactNormal*s2.r;
                }
                penetrationDepth = clearance*(-0.5f);
                return true;
            }
            else
            {
                contactNormal = s1Contact = s2Contact = Vector3.Zero;
                penetrationDepth = 0;
                return false;
            }
        }

        /// <summary>
        /// Test collision between a sphere and a capsule
        /// </summary>
        public static bool TestCollision(Sphere s, Capsule c,
                                         out Vector3 sContact, out Vector3 cContact, out Vector3 contactNormal,
                                         out float penetrationDepth)
        {
            float t;  // not used
            Vector3 cp;
            ClosestPtPointSegment(s.c, c.a, c.b, out t, out cp);

            Vector3 offset = cp - s.c;
            float dist = offset.Length();
            float clearance = dist - (s.r + c.r);

            if (clearance <= 0)
            {
                if (dist < EPSILON)
                {
                    sContact = s.c;
                    cContact = cp;
                    contactNormal = Vector3.Zero;
                }
                else
                {
                    contactNormal = offset / dist;
                    sContact = s.c + contactNormal * s.r;
                    cContact = cp - contactNormal * c.r;
                }
                penetrationDepth = clearance * (-0.5f);
                //Debug.Assert((contactNormal - Vector3.UnitY).LengthSquared() > 0.01, "Ball knocked upward");
                return true;
            }
            else
            {
                contactNormal = sContact = cContact = Vector3.Zero;
                penetrationDepth = 0;
                return false;
            }
        }

#if NotYetPorted

    === Section 5.5.1: =============================================================

// Intersect sphere s0 moving in direction d over time interval t0 <= t <= t1, against
// a stationary sphere s1. If found intersecting, return time t of collision
int TestMovingSphereSphere(Sphere s0, Vector d, float t0, float t1, Sphere s1, float &t)
{
    // Compute sphere bounding motion of s0 during time interval from t0 to t1
    Sphere b;
    float mid = (t0 + t1) * 0.5f;
    b.c = s0.c + d * mid;
    b.r = (mid - t0) * Length(d) + s0.r;
    // If bounding sphere not overlapping s1, then no collision in this interval
    if (!TestSphereSphere(b, s1)) return 0;

    // Cannot rule collision out: recurse for more accurate testing. To terminate the
    // recursion, collision is assumed when time interval becomes sufficiently small
    if (t1 - t0 < INTERVAL_EPSILON) {
        t = t0;
        return 1;
    }

    // Recursively test first half of interval; return collision if detected
    if (TestMovingSphereSphere(s0, d, t0, mid, s1, t)) return 1;

    // Recursively test second half of interval
    return TestMovingSphereSphere(s0, d, mid, t1, s1, t);
}

--------------------------------------------------------------------------------

// Test collision between objects a and b moving over the time interval
// [startTime, endTime]. When colliding, time of collision is returned in hitTime
int IntervalCollision(Object a, Object b, float startTime, float endTime, float &hitTime)
{
    // Compute the maximum distance objects a and b move over the time interval
    float maxMoveA = MaximumObjectMovementOverTime(a, startTime, endTime);
    float maxMoveB = MaximumObjectMovementOverTime(b, startTime, endTime);
    float maxMoveDistSum = maxMoveA + maxMoveB;
    // Exit if distance between a and b at start larger than sum of max movements
    float minDistStart = MinimumObjectDistanceAtTime(a, b, startTime);
    if (minDistStart > maxMoveDistSum) return 0;
    // Exit if distance between a and b at end larger than sum of max movements
    float minDistEnd = MinimumObjectDistanceAtTime(a, b, endTime);
    if (minDistEnd > maxMoveDistSum) return 0;

    // Cannot rule collision out: recurse for more accurate testing. To terminate the
    // recursion, collision is assumed when time interval becomes sufficiently small
    if (endTime - startTime < INTERVAL_EPSILON) {
        hitTime = startTime;
        return 1;
    }
    // Recursively test first half of interval; return collision if detected
    float midTime = (startTime + endTime) * 0.5f;
    if (IntervalCollision(a, b, startTime, midTime, hitTime)) return 1;
    // Recursively test second half of interval
    return IntervalCollision(a, b, midTime, endTime, hitTime);
}

=== Section 5.5.3: =============================================================

// Intersect sphere s with movement vector v with plane p. If intersecting
// return time t of collision and Vector3 q at which sphere hits plane
int IntersectMovingSpherePlane(Sphere s, Vector v, Plane p, float &t, Vector3 &q)
{
    // Compute distance of sphere center to plane
    float dist = Vector3.Dot(p.n, s.c) - p.d;
    if (Math.Abs(dist) <= s.r) {
        // The sphere is already overlapping the plane. Set time of
        // intersection to zero and q to sphere center
        t = 0.0f;
        q = s.c;
        return 1;
    } else {
        float denom = Vector3.Dot(p.n, v);
        if (denom * dist >= 0.0f) {
            // No intersection as sphere moving parallel to or away from plane
            return 0;
        } else {
            // Sphere is moving towards the plane

            // Use +r in computations if sphere in front of plane, else -r
            float r = dist > 0.0f ? s.r : -s.r;
            t = (r - dist) / denom;
            q = s.c + t * v - r * p.n;
            return 1;
        }
    }
}

--------------------------------------------------------------------------------

// Test if sphere with radius r moving from a to b intersects with plane p
int TestMovingSpherePlane(Vector3 a, Vector3 b, float r, Plane p)
{
    // Get the distance for both a and b from plane p
    float adist = Vector3.Dot(a, p.n) - p.d;
    float bdist = Vector3.Dot(b, p.n) - p.d;
    // Intersects if on different sides of plane (distances have different signs)
    if (adist * bdist < 0.0f) return 1;
    // Intersects if start or end position within radius from plane
    if (Math.Abs(adist) <= r || Math.Abs(bdist) <= r) return 1;
    // No intersection
    return 0;
}

=== Section 5.5.5: =============================================================

int TestMovingSphereSphere(Sphere s0, Sphere s1, Vector v0, Vector v1, float &t)
{
    Vector s = s1.c - s0.c;      // Vector between sphere centers
    Vector v = v1 - v0;          // Relative motion of s1 with respect to stationary s0
    float r = s1.r + s0.r;       // Sum of sphere radii
    float c = Vector3.Dot(s, s) - r * r;
    if (c < 0.0f) {
        // Spheres initially overlapping so exit directly
        t = 0.0f;
        return 1;
    }
    float a = Vector3.Dot(v, v);
    if (a < EPSILON) return 0; // Spheres not moving relative each other
    float b = Vector3.Dot(v, s);
    if (b >= 0.0f) return 0;   // Spheres not moving towards each other
    float d = b * b - a * c;
    if (d < 0.0f) return 0;    // No real-valued root, spheres do not intersect

    t = (-b - Sqrt(d)) / a;
    return 1;
}

--------------------------------------------------------------------------------

int TestMovingSphereSphere(Sphere s0, Sphere s1, Vector v0, Vector v1, float &t)
{
    // Expand sphere s1 by the radius of s0
    s1.r += s0.r;
    // Subtract movement of s1 from both s0 and s1, making s1 stationary
    Vector v = v0 - v1;
    // Can now test directed segment s = s0.c + tv, v = (v0-v1)/||v0-v1|| against
    // the expanded sphere for intersection
    Vector3 q;
    float vlen = Length(v);
    if (IntersectRaySphere(s0.c, v / vlen, s1, t, q)) {
        return t <= vlen;
    }
    return 0;
}

=== Section 5.5.7: =============================================================

int IntersectMovingSphereAABB(Sphere s, Vector d, AABB b, float &t)
{
    // Compute the AABB resulting from expanding b by sphere radius r
    AABB e = b;
    e.min.x -= s.r; e.min.y -= s.r; e.min.z -= s.r;
    e.max.x += s.r; e.max.y += s.r; e.max.z += s.r;

    // Intersect ray against expanded AABB e. Exit with no intersection if ray
    // misses e, else get intersection Vector3 p and time t as result
    Vector3 p;
    if (!IntersectRayAABB(s.c, d, e, t, p) || t > 1.0f)
        return 0;

    // Compute which min and max faces of b the intersection Vector3 p lies
    // outside of. Note, u and v cannot have the same bits set and
    // they must have at least one bit set amongst them
    int u = 0, v = 0;
    if (p.x < b.min.x) u |= 1;
    if (p.x > b.max.x) v |= 1;
    if (p.y < b.min.y) u |= 2;
    if (p.y > b.max.y) v |= 2;
    if (p.z < b.min.z) u |= 4;
    if (p.z > b.max.z) v |= 4;

    // Or all set bits together into a bit mask (note: here u + v == u | v)
    int m = u + v;

    // Define line segment [c, c+d] specified by the sphere movement
    Segment seg(s.c, s.c + d);

    // If all 3 bits set (m == 7) then p is in a vertex region
    if (m == 7) {
        // Must now intersect segment [c, c+d] against the capsules of the three
        // edges meeting at the vertex and return the best time, if one or more hit
        float tmin = FLT_MAX;
        if (IntersectSegmentCapsule(seg, Corner(b, v), Corner(b, v ^ 1), s.r, &t))
            tmin = Min(t, tmin);
        if (IntersectSegmentCapsule(seg, Corner(b, v), Corner(b, v ^ 2), s.r, &t))
            tmin = Min(t, tmin);
        if (IntersectSegmentCapsule(seg, Corner(b, v), Corner(b, v ^ 4), s.r, &t))
            tmin = Min(t, tmin);
        if (tmin == FLT_MAX) return 0; // No intersection
        t = tmin;
        return 1; // Intersection at time t == tmin
    }
    // If only one bit set in m, then p is in a face region
    if ((m & (m - 1)) == 0) {
        // Do nothing. Time t from intersection with
        // expanded box is correct intersection time
        return 1;
    }
    // p is in an edge region. Intersect against the capsule at the edge
    return IntersectSegmentCapsule(seg, Corner(b, u ^ 7), Corner(b, v), s.r, &t);
}

// Support function that returns the AABB vertex with index n
Vector3 Corner(AABB b, int n)
{
    Vector3 p;
    p.x = ((n & 1) ? b.max.x : b.min.x);
    p.y = ((n & 2) ? b.max.y : b.min.y);
    p.z = ((n & 4) ? b.max.z : b.min.z);
    return p;
}

=== Section 5.5.8: =============================================================

// Intersect AABBs a and b moving with constant velocities va and vb.
// On intersection, return time of first and last contact in tfirst and tlast
int IntersectMovingAABBAABB(AABB a, AABB b, Vector va, Vector vb, float &tfirst, float &tlast)
{
    // Exit early if a and b initially overlapping
    if (TestAABBAABB(a, b)) {
        tfirst = tlast = 0.0f;
        return 1;
    }

    // Use relative velocity; effectively treating 'a' as stationary
    Vector v = vb - va;

    // Initialize times of first and last contact
    tfirst = 0.0f;
    tlast = 1.0f;

    // For each axis, determine times of first and last contact, if any
    for (int i = 0; i < 3; i++) {
        if (v[i] < 0.0f) {
            if (b.max[i] < a.min[i]) return 0; // Nonintersecting and moving apart
            if (a.max[i] < b.min[i]) tfirst = Max((a.max[i] - b.min[i]) / v[i], tfirst);
            if (b.max[i] > a.min[i]) tlast  = Min((a.min[i] - b.max[i]) / v[i], tlast);
        }
        if (v[i] > 0.0f) {
            if (b.min[i] > a.max[i]) return 0; // Nonintersecting and moving apart
            if (b.max[i] < a.min[i]) tfirst = Max((a.min[i] - b.max[i]) / v[i], tfirst);
            if (a.max[i] > b.min[i]) tlast = Min((a.max[i] - b.min[i]) / v[i], tlast);
        }

        // No overlap possible if time of first contact occurs after time of last contact
        if (tfirst > tlast) return 0;
    }

    return 1;
}
#endif
    }
}