// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace Twig
{
    /// <summary>
    /// Object that implements a particular heads-up display.
    /// </summary>
    public abstract class HUD : DrawableGameComponent, IDrawable
    {
        TwigGame twigGame;

        /// <summary>
        /// TwigGame object to which this HUD belongs
        /// </summary>
        public TwigGame TwigGame
        {
            get { return twigGame; }
            set { twigGame = value; }
        }

        /// <summary>
        /// Default draworder for heads-up displays
        /// </summary>
        public const int HUDDrawOrder = 9000;
        int IDrawable.DrawOrder
        {
            get
            {
                return HUDDrawOrder;
            }
        }

        /// <summary>
        /// GraphicsDevice of this object's game
        /// </summary>
        protected GraphicsDevice Device
        {
            get
            {
                return TwigGame.Device;
            }
        }

        /// <summary>
        /// BasicEffect object used by this HUD's game for displaying HUDs.
        /// </summary>
        protected BasicEffect HUDEffect {
            get
            {
                return TwigGame.HUDEffect;
            }
        }
        /// <summary>
        /// SpriteBatch to use in rendering HUDs
        /// </summary>
        protected SpriteBatch HUDSprites
        {
            get
            {
                return TwigGame.HUDSprites;
            }
        }

        /// <summary>
        /// Font to use when rendering HUDS
        /// </summary>
        protected SpriteFont HUDFont
        {
            get
            {
                return TwigGame.HUDFont;
            }
        }

        /// <summary>
        /// Creates a HUD and adds it to game.
        /// </summary>
        /// <param name="game"></param>
        public HUD(TwigGame game) : base(game)
        {
            twigGame = game;
        }

        /// <summary>
        /// Creates a HUD for the specified game with the specified hotkey and initial value for the Displayed attribute.
        /// </summary>
        public HUD(TwigGame game, Keys hotKey, string keyDocumentation, bool visible)
            : this(game)
        {
            TwigGame.BindKey(hotKey, keyDocumentation, delegate(Keys ignore)
                {
                    Visible = !Visible;
                });
            Visible = visible;
        }

        /// <summary>
        /// Draws a line on the HUD with the specified endpoints and color.
        /// </summary>
        protected void DrawLine(float xs, float ys, float xe, float ye, Color color)
        {
            TwigGame.DrawLine(xs, ys, xe, ye, color);
        }
    }
}
