// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

//#define DebugFight

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using System.Diagnostics;

using SlotSet = Twig.Mammal.AttentionBuffer.SlotSet;
using Slot = Twig.Mammal.AttentionBuffer.Slot;

#pragma warning disable 1591

namespace Twig
{
    //
    // The Mammal class extends Ragdoll with posture control
    // and basic locomotion and manipulation behaviors
    //
    // See the files Posture.cs, Walking.cs, and Hugging.cs for actual
    // behavior implementations
    //
    // See Person class for affect and motivation systems
    //
    public abstract partial class Mammal : Ragdoll
    {
        /// <summary>Mark up objects on screen with local coordinate frames, etc.</summary>
        public static bool showDebugLines = false;

        /// <summary>
        /// Activation threshold for approach behaviors, beneath which the behavior
        /// won't fire even if it's a winner.
        /// </summary>
        public float actionThreshold = 0.5f;
        /// <summary>
        /// Maximum speed the creature can walk at.
        /// </summary>
        public float maxSpeed;

        DistanceLowerBound spineConstraint;
        float height;

        /// <summary>
        /// Height of creature for purposes of camera framing.
        /// </summary>
        public float Height
        {
            get { return height; }
            set { height = value; }
        }

        /// <summary>
        /// Height of eye level of character for purposes of camera framing.
        /// </summary>
        public float EyeLevel
        {
            get { return head.MidPoint.Y; }
            //set { eyeLevel = value; }
        }


        #region State variables maintained by Update()
        // Midpoints of feet, pelvis, shoulders
        /// <summary>
        /// Midpoint of the two feet.
        /// A useful definition of "where the creature is standing".
        /// </summary>
        public Vector3 footMidpoint;
        /// <summary>
        /// Midpoint of the pelvis.
        /// </summary>
        public Vector3 pelvis;
        /// <summary>
        /// Midpoint of the shoulders.
        /// </summary>
        public Vector3 torso;
        // Shoulder-centered coordinate system
        /// <summary>
        /// The rough "up" direction of the torso.
        /// Defined as torso center - pelvis center.
        /// </summary>
        public Vector3 bodyUp;
        /// <summary>
        /// Points "right" in a torso-centered coordinate system.
        /// I.e. the direction of the right shoulder relative to the left.
        /// </summary>
        public Vector3 bodyRight;
        /// <summary>
        /// Direction the torso is facing, i.e. the normal of the front of the chest.
        /// </summary>
        public Vector3 bodyForward;
        // Pelvis-centered

        /// <summary>
        /// Points "right" in a pelvis-centered coordinate system.
        /// I.e. the direction of the right hip relative to the left.
        /// </summary>
        public Vector3 pelvisRight;
        /// <summary>
        /// Direction the pelvis is facing, i.e. the normal of the front of the pelvis.
        /// </summary>
        public Vector3 pelvisForward;

        // Foot-centered
        /// <summary>
        /// Forward direction for feet.
        /// Normal both to the line connecting the feet and also to bodyUp.
        /// </summary>
        public Vector3 footForward;
        // Head-centered
        /// <summary>
        /// Gaze direction relative to head.
        /// </summary>
        public Vector3 lookForward;
        /// <summary>
        /// Rightward direction in head-centered coordinate system.
        /// </summary>
        public Vector3 lookRight;
        /// <summary>
        /// Up direction in head-centered coordinate system.
        /// </summary>
        public Vector3 lookUp;
        #endregion

        #region Functions and predicates
        /// <summary>
        /// True if the slot in the attention buffer actually corresponds to a real object.
        /// False if the slot is empty.
        /// </summary>
        public Predicate boundp;
        /// <summary>
        /// Whether the character is facing toward the object
        /// </summary>
        public Predicate facingToward;
        /// <summary>
        /// Whether the character is facing away from the object
        /// </summary>
        public Predicate facingAwayFrom;
        /// <summary>
        /// Whether the character is looking at the object.
        /// </summary>
        public Predicate lookingAt;
        /// <summary>
        /// Distance of object from character
        /// </summary>
        public Function<float> distance;
        /// <summary>
        /// Distance from docking point
        /// </summary>
        public Function<float> dockingDistance;
        /// <summary>
        /// Cosine of the angle between the front direction of the character's body
        /// and the object.
        /// </summary>
        public Function<float> orientationCosine;
        #endregion

        /// <summary>
        /// Base constructor for all mammalian characters
        /// </summary>
        public Mammal(TwigGame g, string name, float spineLength, float spineRadius, 
            float armSpacing, float armRadius, float upperArm, float foreArm,
            float pelvis, float legRadius, float femur, float shin, float headHeight)
            : base(g, name, spineLength, spineRadius, armSpacing, armRadius, upperArm, foreArm, pelvis, legRadius, femur, shin, headHeight)
        {
            spineConstraint = new DistanceLowerBound(name+" spine constraint", this, spineTop, spineBottom, spineLength, 0f);

            // Normalize knee bend
            targetKneeBend *= spineLength;
            standGain /= spineLength;

            Attention = new AttentionBuffer(this);

            maxSpeed = 2f * legs.left.length;

            approach = new ApproachBehavior(this);
            painWithdrawalReflex = new PainWithdrawalReflex(this);
            freezeLocomotion = new FreezeLocomotion();
            // Switching locomotion behaviors takes 200ms
            locomotionController = new MaxBehavior<Vector3>(0.2f, new Behavior<Vector3>[]
                                        { freezeLocomotion, painWithdrawalReflex, approach });

            approachIdle = new ApproachNothing();
            fightBehavior = new FightBehavior(this);
            exploreBehavior = new ExploreBehavior(this);
            attachmentBehavior = new AttachmentBehavior(this);
            approachActionBehavior = new ApproachActionBehavior(this);
            // Can't switch approach behaviors more over than every 0.75s
            approachController = new MaxBehavior<ApproachParameters>(0.75f, new Behavior<ApproachParameters>[]
                                        { approachIdle, approachActionBehavior, attachmentBehavior, fightBehavior, exploreBehavior });
            head.painSensitivity = 2f;

            boundp = new Predicate("boundp", this,
                delegate(TwigObject o)
                {
                    return o != null;
                });
            distance = new Function<float>("distance", this,
                delegate(TwigObject o) {
                    if (o == null)
                        return float.PositiveInfinity;
                    else
                        return MathUtil.Distance(Position, o.Position);
                });
            dockingDistance = new Function<float>("distance", this,
                delegate(TwigObject o)
                {
                    if (o == null)
                        return float.PositiveInfinity;
                    else
                        return MathUtil.Distance(Position, o.DockingTarget);
                });

            orientationCosine = new Function<float>("orientationCosine", this, new Function<float>.ObjectUpdateProcedure(this.OrientationCosine));
            facingToward = new Predicate("facing", this,
                delegate()
                {
                    float[] orientations = orientationCosine.ValueArray;
                    byte mask = 0;
                    for (int i=0; i < orientations.Length; i++)
                        if (orientations[i] > 0.75f)
                            mask = (byte)(mask | (1 << i));
                    return new SlotSet(mask);
                });
            facingAwayFrom = new Predicate("facingAwayFrom", this,
                delegate()
                {
                    float[] orientations = orientationCosine.ValueArray;
                    byte mask = 0;
                    for (int i=0; i < orientations.Length; i++)
                        if (orientations[i] < -0.75f)
                            mask = (byte)(mask | (1 << i));
                    return new SlotSet(mask);
                });
            lookingAt = new Predicate("lookingAt", this,
                delegate() {
                    if (gazeTarget==null)
                        return SlotSet.Empty;
                    else
                        return new SlotSet(Attention.ObjectSlot(gazeTarget));
                });

            InitializeSitting();
            InitializeGrasping();
        }

        #region Useful geometric information
        public float StandingHeight
        {
            get
            {
                return headTop.Position.Y;
            }
        }

        float PelvisAzmuth(Vector3 v)
        {
            if (v == Vector3.Zero)
                return 0f;
            else
                return (float)Math.Atan2(Vector3.Dot(v, pelvisRight), Vector3.Dot(v, pelvisForward));
        }

        public override Vector3 Position
        {
            get {
                Vector3 position = spineBottom.Position;
                position.Y = GroundLevel;
                return position;
            }
        }

        public float DistanceTo(TwigObject o)
        {
            return (Position - o.Position).Length();
        }

        public abstract float GroundLevel
        {
            get;
        }
        #endregion

        #region Update logic
        public override void Constrain()
        {
#if constrainKneeSeparation
            // kneeDist needs to be positive
            // i.e. the right knee needs to be to the right
            // of the left knee.
            const float minKneeDist = 0.2f;
            float kneeDist = Vector3.Dot(pelvisRight, legs.right.joint.Position - legs.left.joint.Position);
            if (kneeDist < minKneeDist)
            {
                Vector3 motion = pelvisRight * (minKneeDist - kneeDist) * 0.5f;
                legs.right.joint.Position += motion;
                legs.left.joint.Position -= motion;
            }
#endif
            // footDist needs to be positive
            // i.e. the right foot needs to be to the right
            // of the left foo.
            const float minFootDist = 0.2f;
            float footDist = Vector3.Dot(pelvisRight, legs.right.end.Position - legs.left.end.Position);
            if (footDist < minFootDist)
            {
                Vector3 motion = pelvisRight * (minFootDist - footDist) * 0.5f;
                legs.right.end.Position += motion;
                legs.left.end.Position -= motion;
            }

            //
            // Constrain knees to lie in plane formed by hip-foot vector and "forward" direction
            //

            // Left knee
            leftLegDirection = legs.left.root.Position - legs.left.end.Position;
            leftLegForward = Vector3.Cross(pelvisRight, leftLegDirection);
            leftLegPlaneNormal = Vector3.Cross(leftLegForward, leftLegDirection);
            leftLegPlaneNormal.Normalize();
            // Distance knee is offset from the plane relative to foot
            float leftKneeError = Vector3.Dot(leftLegPlaneNormal, legs.left.joint.Position - legs.left.end.Position);
            Vector3 loffset = 0.5f * leftKneeError * leftLegPlaneNormal;
            legs.left.joint.Position -= loffset;
            legs.left.end.Position += loffset;
            leftKneeForward = Vector3.Cross(leftLegPlaneNormal, leftLegDirection);
            leftKneeForward.Normalize();
            float leftKneeBend = Vector3.Dot(leftKneeForward, legs.left.joint.Position - legs.left.end.Position);
            if (leftKneeBend < targetKneeBend)
            {
                Vector3 lkbCorrection = 0.5f * (leftKneeBend-targetKneeBend) * leftKneeForward;
                legs.left.joint.Position -= lkbCorrection;
                legs.left.end.Position += lkbCorrection;
            }

            // Right knee
            rightLegDirection = legs.right.root.Position - legs.right.end.Position;
            rightLegForward = Vector3.Cross(rightLegDirection, pelvisRight);
            rightLegPlaneNormal = Vector3.Cross(rightLegForward, rightLegDirection);
            rightLegPlaneNormal.Normalize();
            // Distance knee is offset from the plane relative to foot
            float rightKneeError = Vector3.Dot(rightLegPlaneNormal, legs.right.joint.Position - legs.right.end.Position);
            Vector3 roffset = 0.5f * rightKneeError * rightLegPlaneNormal;
            legs.right.joint.Position -= roffset;
            legs.right.end.Position += roffset;

            // Note reversal of cross product relative to left leg.
            rightKneeForward = Vector3.Cross(rightLegDirection, rightLegPlaneNormal);
            rightKneeForward.Normalize();
            float rightKneeBend = Vector3.Dot(rightKneeForward, legs.right.joint.Position - legs.right.end.Position);
            if (rightKneeBend < targetKneeBend)
            {
                Vector3 rkbCorrection = 0.5f * (rightKneeBend-targetKneeBend) * rightKneeForward;
                legs.right.joint.Position -= rkbCorrection;
                legs.right.end.Position += rkbCorrection;
            }

#if CONSTRAIN_ELBOWS
            //
            // Now the elbows
            //

            const float targetElbowBend = 0.2f;

            // Left elbow
            Vector3 leftArmDirection = arms.left.root.Position - arms.left.end.Position;
            Vector3 leftArmForward = Vector3.Cross(leftArmDirection, pelvisRight);
            Vector3 leftArmPlaneNormal = Vector3.Cross(leftArmForward, leftArmDirection);
            leftArmPlaneNormal.Normalize();
            if (!hug)
            {
                // Distance Elbow is offset from the plane relative to wrist
                float leftElbowError = Vector3.Dot(leftArmPlaneNormal, arms.left.joint.Position - arms.left.end.Position);
                Vector3 leoffset = 0.5f * leftElbowError * leftArmPlaneNormal;
                arms.left.joint.Position -= leoffset;
                arms.left.end.Position += leoffset;
            }
            Vector3 leftElbowForward = Vector3.Cross(leftArmPlaneNormal, leftArmDirection);
            leftElbowForward.Normalize();
            float leftElbowBend = Vector3.Dot(leftElbowForward, arms.left.joint.Position - arms.left.end.Position);
            if (leftElbowBend > targetElbowBend)
            {
                Vector3 lebCorrection = 0.5f * (leftElbowBend - targetElbowBend) * leftElbowForward;
                arms.left.joint.Position -= lebCorrection;
                arms.left.end.Position += lebCorrection;
            }

            // Right elbow
            Vector3 rightArmDirection = arms.right.root.Position - arms.right.end.Position;
            Vector3 rightArmForward = Vector3.Cross(rightArmDirection, pelvisRight);
            Vector3 rightArmPlaneNormal = Vector3.Cross(rightArmForward, rightArmDirection);
            rightArmPlaneNormal.Normalize();
            if (!hug)
            {
                // Distance Elbow is offset from the plane relative to wrist
                float rightElbowError = Vector3.Dot(rightArmPlaneNormal, arms.right.joint.Position - arms.right.end.Position);
                Vector3 leoffset = 0.5f * rightElbowError * rightArmPlaneNormal;
                arms.right.joint.Position -= leoffset;
                arms.right.end.Position += leoffset;
            }
            // Note reversal of cross product relative to left arm.
            Vector3 rightElbowForward = Vector3.Cross(rightArmDirection, rightArmPlaneNormal);
            rightElbowForward.Normalize();
            float rightElbowBend = Vector3.Dot(rightElbowForward, arms.right.joint.Position - arms.right.end.Position);
            if (rightElbowBend > targetElbowBend)
            {
                Vector3 lebCorrection = 0.5f * (rightElbowBend - targetElbowBend) * rightElbowForward;
                arms.right.joint.Position -= lebCorrection;
                arms.right.end.Position += lebCorrection;
            }
#endif

            base.Constrain();
            if (sitUp || standUp)
                spineJoint.Position = pelvis + Vector3.Up * (spineLength / 3f)+leanForward*0.5f;
            spineConstraint.Constrain();
        }

        Vector3 Centroid()
        {
            Vector3 sum = Vector3.Zero;
            foreach (Node n in nodes)
                sum += n.Position;
            return sum / nodes.Length;
        }

        public override void  Update(GameTime gameTime)
        {
#if DEBUG
            // Remove any wait conditions marked from the previous clock tick.
            waitList.Clear();
#endif

            SensePain();
            Attention.Update();
            UpdateInference();
            UpdateGaze();
            UpdateSecurity();

            // Precompute walking parameters that are used
            // by posture control.
            // Note: this will cause locomotionController to run the locomotion behaviors,
            // which will cause approachController to run all the approach-related behaviors
            ComputeWalkingParameters();
            UpdateEgocentricBasisVectors();
            PostureControl();
            GaitControl();
            StaticBalance();
            HugControl();
            SittingControl();
            GraspControl();
            if (standUp && walkVector == Vector3.Zero && (footForward - pelvisForward).LengthSquared() < 0.1f)
            {
                legs.left.end.ControlMode = ControlMode.FallAndStick;
                legs.right.end.ControlMode = ControlMode.FallAndStick;
            }
            else
            {
                legs.left.end.ControlMode = ControlMode.Free;
                legs.right.end.ControlMode = ControlMode.Free;
            }


#if DebugFight
            Chart("FocusAppraisal", Appraisal(Attention.Focus).valence, 2);
            Chart("Anxiety", Anxiety, 2);
            Chart("Pain", pain, 10000);
#endif
            base.Update(gameTime);
        }

#if OldCollision
        public override bool TestCollisionWithRagdoll(Ragdoll h)
        {
            Vector3 offset = h.Position - Position;
            bool contact = false;
            // Don't bother if they're too far apart
            if (offset.Length() < 2 * (legLength + spineLength))
            {
                // Ick.  Collide every link with every other link
                foreach (Link l in Links)
                {
                    foreach (Link hl in h.Links)
                        if (l.CheckCollision(hl, h))
                            contact = true;
                }
            }
            if (hugging && arms.left.grip != null)
                foreach (Link hl in h.Links)
                    if (arms.left.grip.CheckCollision(hl, this))
                        contact = true;

            return contact;
        }
#endif


        void UpdateEgocentricBasisVectors()
        {
            // Compute basis vectors for egocentric coordinate systems
            footMidpoint = (legs.left.end.Position + legs.right.end.Position) * .5f;
            Debug.Assert(MathUtil.IsProperVector(footMidpoint), "Foot midpoint is an invalid vector");
            footMidpoint.Y = 0;  // Just in case a foot is lifted for a step
            pelvis = (legs.left.root.Position + legs.right.root.Position) * 0.5f;
            Debug.Assert(MathUtil.IsProperVector(pelvis), "Pelvis midpoint is an invalid vector");
            torso = (arms.left.root.Position + arms.right.root.Position) * 0.5f;
            Debug.Assert(MathUtil.IsProperVector(torso), "Shoulder midpoint is an invalid vector");

            // Set up an egocentric coordinate system
            bodyUp = torso - pelvis;
            bodyUp.Normalize();
            Debug.Assert(MathUtil.IsProperVector(bodyUp), "Body up vector is an invalid vector");
            bodyRight = arms.right.root.Position - arms.left.root.Position;
            bodyRight.Normalize();
            Debug.Assert(MathUtil.IsProperVector(bodyRight), "Body right vector is an invalid vector");
            bodyForward = Vector3.Cross(bodyUp, bodyRight);
            bodyForward.Normalize();
            Debug.Assert(MathUtil.IsProperVector(bodyForward), "Body forward vector is an invalid vector");
            pelvisRight = legs.right.root.Position - legs.left.root.Position;
            pelvisRight.Normalize();
            Debug.Assert(MathUtil.IsProperVector(pelvisRight), "Pelvis right vector is an invalid vector");
            pelvisForward = Vector3.Cross(bodyUp, pelvisRight);
            pelvisForward.Normalize();
            Debug.Assert(MathUtil.IsProperVector(pelvisForward), "Pelvis forward vector is an invalid vector");
            footForward = Vector3.Cross(bodyUp, legs.right.end.Position - legs.left.end.Position);
            footForward.Normalize();
            Debug.Assert(MathUtil.IsProperVector(footForward), "Foot forward vector is an invalid vector");

            // Head-centered
            lookForward = (locomotionController.Activation>0 && walkDirection.LengthSquared()>0.1f)?walkDirection:bodyForward;
            if (gazeTarget != null)
            {
                Vector3 newTarget = gazeTarget.LookPoint - spineTop.Position;
                //if (gazeTarget is Marker)
                //    newTarget.Y = 0;
                float targetDistance = (gazeTarget.Position - Position).Length();
                if (hug && targetDistance < 4f)
                    // Move the face down toward the leg when approaching for a hug
                    newTarget.Y *= targetDistance / 4f;
                MathUtil.SafeNormalize(ref newTarget);
                if (newTarget == Vector3.Zero)
                    newTarget = bodyForward;
                //lookForward = 0.75f * lookForward + 0.25f * newTarget;
                //lookForward.Normalize();
                lookForward = newTarget;
            }
            Debug.Assert(MathUtil.IsProperVector(lookForward), "Look forward vector is an invalid vector");
            lookUp = headTop.Position - spineTop.Position;
            lookUp.Normalize();
            Debug.Assert(MathUtil.IsProperVector(lookUp), "Look up vector is an invalid vector");
            lookRight = Vector3.Cross(lookForward, lookUp);
            lookRight.Normalize();
            Debug.Assert(MathUtil.IsProperVector(lookRight), "Look right vector is an invalid vector");
        }
        #endregion

        /// <summary>
        /// Move character to specified location and orientation
        /// </summary>
        public override void PoseAt(Vector3 position, Vector3 facingDirection)
        {
            base.PoseAt(position, facingDirection);
            sitUp = standUp = true;
            UpdateEgocentricBasisVectors();
        }

        /// <summary>
        /// Pick up object and hold in preparation for use.  Uses default hand.
        /// </summary>
        public void HoldForUse(TwigObject o)
        {
            HoldForUse(o, Grasp.Any);
        }

        /// <summary>
        /// Pick up object and hold in preparation for use.
        /// </summary>
        public virtual void HoldForUse(TwigObject o, Grasp grasp)
        {
            o.SetupHoldForUse(this, grasp);
        }

        /// <summary>
        /// Whether the character is right-handed
        /// </summary>
        public bool RightHanded = true;

        /// <summary>
        /// Left hand is not grasping anything
        /// </summary>
        public bool LeftHandFree
        {
            get
            {
                return !arms.left.Gripping;
            }
        }

        /// <summary>
        /// Right hand is not grasping anything
        /// </summary>
        public bool RightHandFree
        {
            get
            {
                return !arms.right.Gripping;
            }
        }

        /// <summary>
        /// True if the left hand is acceptable for this grasp.
        /// </summary>
        public bool AcceptLeftHand(Grasp g)
        {
            return ((g&Grasp.LeftHanded)>0) || ((g&Grasp.NonDominantHand)>0 &&RightHanded)
                || ((g&Grasp.DominantHand)>0&&!RightHanded);
        }

        /// <summary>
        /// True if the right hand is acceptable for this grasp.
        /// </summary>
        public bool AcceptRightHand(Grasp g)
        {
            return ((g & Grasp.RightHanded) > 0) || ((g & Grasp.DominantHand) > 0 && RightHanded)
                || ((g & Grasp.NonDominantHand) > 0 && !RightHanded);
        }

        /// <summary>
        /// Given a set of grasp preferences and the current state of the hands,
        /// return either Grasp.LeftHanded, Grasp.RightHanded, or Grasp.TwoHanded.
        /// </summary>
        public Grasp ChooseHand(Grasp g)
        {
            if (RightHanded)
            {
                if (RightHandFree & AcceptRightHand(g))
                    return Grasp.RightHanded;
                else if (LeftHandFree & AcceptLeftHand(g))
                    return Grasp.LeftHanded;
                else if ((g & Grasp.TwoHanded) > 0)
                    return g & Grasp.TwoHanded;
                else
                    throw new NoFreeHandException();
            }
            else
            {
                if (LeftHandFree & AcceptLeftHand(g))
                    return Grasp.LeftHanded;
                else if (RightHandFree & AcceptRightHand(g))
                    return Grasp.RightHanded;
                else if ((g & Grasp.TwoHanded) > 0)
                    return g & Grasp.TwoHanded;
                else
                    throw new NoFreeHandException();
            }
        }

        /// <summary>
        /// Given a set of grasp preferences and the current state of the hands,
        /// return either Grasp.LeftHanded, Grasp.RightHanded, or Grasp.TwoHanded.
        /// </summary>
        public void ChooseHand(ref Grasp g)
        {
            g = ChooseHand(g);
        }

        /// <summary>
        /// Hold the specified object however we can
        /// </summary>
        public virtual void Hold(TwigObject o)
        {
            o.SetupHold(this, Grasp.Any);
        }

        /// <summary>
        /// Grasp the specified object using the specified node and hand.
        /// </summary>
        public void Hold(Node grip, TwigObject o, Grasp g)
        {
            Hold(grip, grip, o, g);
        }

        /// <summary>
        /// Grasp specified object using the specified hand and whichever node is appropriate
        /// </summary>
        public virtual void Hold(Node leftHandGrip, Node rightHandGrip, TwigObject o, Grasp g)
        {
            if (arms.left.GraspedObject == o || arms.right.GraspedObject == o)
                return;

            ChooseHand(ref g);
            if ((g & Twig.Grasp.TwoHanded) > 0)
            {
                arms.left.Grab(leftHandGrip, o);
                arms.right.Grab(rightHandGrip, o);
            }
            else if ((g & Twig.Grasp.LeftHanded) > 0)
                arms.left.Grab(leftHandGrip, o);
            else
                arms.right.Grab(rightHandGrip, o);
        }

        public virtual void SelectGrasp(Node leftHandGrip, Node rightHandGrip, TwigObject o, Grasp graspRequest,
                                        out Grasp selectedGrasp, out Node leftTarget, out Node rightTarget)
        {
            if (arms.left.GraspedObject == o || arms.right.GraspedObject == o)
            {
                selectedGrasp = Grasp.None;
                leftTarget = leftTarget = null;
            }
            selectedGrasp = graspRequest;
            ChooseHand(ref selectedGrasp);
            if ((selectedGrasp & Twig.Grasp.TwoHanded) > 0)
            {
                leftTarget = leftHandGrip;
                rightTarget = rightHandGrip;
            }
            else if ((selectedGrasp & Twig.Grasp.LeftHanded) > 0)
            {
                leftTarget = leftHandGrip;
                rightTarget = null;
            }
            else {
                leftTarget = null;
                rightTarget = rightHandGrip;
            }
        }


        /// <summary>
        /// Drop object being held.
        /// </summary>
        /// <param name="o"></param>
        public void Drop(TwigObject o)
        {
            if (arms.left.GraspedObject == o)
                arms.left.Ungrab();
            if (arms.right.GraspedObject == o)
                arms.right.Ungrab();
        }

        /// <summary>
        /// True if character is writing with a pen.
        /// </summary>
        public bool IsManipulating
        {
            get
            {
                return (arms.left.GraspedObject != null && arms.left.GraspedObject.IsBeingManipulated)
                        || (arms.right.GraspedObject != null && arms.right.GraspedObject.IsBeingManipulated);
            }
        }

        public override void Draw(GameTime gameTime)
        {
#if DEFER_DEBUGLINES
            if (showDebugLines)
            {
                d.RenderState.Lighting = false;
                Link.DrawLine(d, legs.left.root.Position, legs.left.root.Position + 3.0f * pelvisForward, new ColorValue(0.0f, 1.0f, 0.0f));
                Link.DrawLine(d, arms.left.root.Position, arms.left.root.Position + 3.0f * bodyForward, new ColorValue(0.0f, 1.0f, 0.0f));
                Link.DrawLine(d, legs.left.joint.Position, legs.left.joint.Position + leftKneeForward, new ColorValue(0.0f, 0.0f, 1.0f));
                Link.DrawLine(d, legs.right.joint.Position, legs.right.joint.Position + rightKneeForward, new ColorValue(0.0f, 0.0f, 1.0f));
                Link.DrawLine(d, torso, torso + 2.0f * walkVector, new ColorValue(1.0f, 1.0f, 0.0f));
                d.RenderState.Lighting = true;
            }
            if (showWalkVector)
            {
                d.RenderState.Lighting = false;
                Link.DrawLine(d, Position + Vector3.Up * 3f * spineLength, Position + Vector3.Up * 3f * spineLength + walkVector,
                    new ColorValue(255, 255, 255));
                d.RenderState.Lighting = true;
            }
#endif
            base.Draw(gameTime);
        }

        #region Dialog
        /// <summary>
        /// Event handler for dialog notifications
        /// </summary>
        public delegate void DialogEventHandler(TwigObject speaker, string speech);

        /// <summary>
        /// Signaled when character speaks a new utterance.
        /// </summary>
        public event DialogEventHandler NewUtterance;

        public void Say(string speech)
        {
            if (NewUtterance != null)
                NewUtterance(this, speech);

            DialogBalloonManager.RemoveDialog(this);
            DialogBalloonManager.Say(this, new string[] { speech });
        }

        public Action Say(string speech, TwigObject addressee, bool interrupt, float reactionTime)
        {
            return new SayAction(this, speech, addressee, interrupt, reactionTime);
        }

        public Action Say(string speech, TwigObject addressee)
        {
            return Say(speech, addressee, false, 0.2f);
        }

        /// <summary>
        /// True if this character is speaking
        /// </summary>
        public bool IsSpeaking
        {
            get
            {
                return DialogBalloonManager.IsSpeaking(this);
            }
        }

        /// <summary>
        /// Force character to look at target
        /// </summary>
        public void LookAt(TwigObject target)
        {
            // Need to assign gazeTarget here for the case where target is null
            // because the normal gazeTarget search will never set it back to null.
            gazeTarget = lookAtCommandTarget = target;
        }

        /// <summary>
        /// Screen position of the top of the character's head
        /// </summary>
        public Point HeadScreenPosition
        {
            get
            {
                Vector3 pos = Device.Viewport.Project(headTop.Position, TwigGame.ProjectionMatrix, TwigGame.ViewMatrix, Matrix.Identity);
                return new Point((int)pos.X, (int)pos.Y);
            }
        }

        public void Fight(Mammal target)
        {
            specificObjectAppraisal[target] = new ObjectAppraisal(-1);
            Attention.Focus = target;
        }

        public void StopFight(Mammal target)
        {
            specificObjectAppraisal[target] = new ObjectAppraisal(0);
            Goto(null).Update();
        }

        #endregion

        #region Debugging support
        public override string DebugString
        {
            get
            {
                return String.Format("{0}\n\nAnxiety: {1:0.00}     Pain: {10:0.##}\nFocus: {3} ({4:0.00})\nGaze: {5} ({6:0.00})\n{7}: {8} ({9:0.##})\nLocomotion: {11}\nWalk: {12}\nAB: {13}, target={14}\ndist={15}, direction={16}\nApproach: {19}, activation={17}\nmotor output={18}\nrepulsion={20}",
                    base.DebugString, Anxiety,
                    attachmentBehavior.Activation,
                    BriefName(Attention.Focus), Appraisal(Attention.Focus).valence,
                    BriefName(gazeTarget), Appraisal(gazeTarget).valence,
                    (approachController.maxBehavior==null)?"Idle":approachController.maxBehavior.Name,
                    BriefName(approachController.MotorOutput.target),
                    approachController.Activation,
                    pain,
                    (locomotionController.maxBehavior==null)?"null":locomotionController.maxBehavior.Name,
                    walk,
                    approachController.maxBehavior.Name, BriefName(approachController.MotorOutput.target),
                    approachController.MotorOutput.approachDistance, approachController.MotorOutput.approachOffset,
                    approach.Activation,
                    approach.MotorOutput,
                    approachController.maxBehavior.Name,
                    approach.repulsion);
            }
        }

        string BriefName(TwigObject o)
        {
            if (o == null)
                return "null";
            else
                return o.Name;
        }
        public void LogMotion()
        {
#if NodeLogging
            foreach (Node n in Nodes)
                n.EnableLogging();
#endif
        }

#if DEBUG
        /// <summary>
        /// List of conditions that different control systems within the character are waiting for.
        /// </summary>
        public List<string> waitList = new List<string>();
#endif

        /// <summary>
        /// Asserts that some part of the character's control system is waiting for the specified condition
        /// (debugging instrumentation)
        /// </summary>
        [Conditional("DEBUG"), DebuggerStepThrough()]
        public void WaitingFor(string description)
        {
#if DEBUG
            if (!waitList.Contains(description))
                waitList.Add(description);
#endif
        }
        #endregion
    }
}
