// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;

namespace Twig
{
    /// <summary>
    /// Definitions for useful mathematical operations.
    /// </summary>
    public static class MathUtil
    {
        /// <summary>
        /// True if number is neither infinite nor NaN
        /// </summary>
        public static bool IsProperNumber(this float x)
        {
            return !(Single.IsInfinity(x) || Single.IsNaN(x));
        }

        /// <summary>
        /// True if number is neither infinite nor NaN
        /// </summary>
        public static bool IsProperNumber(this double x)
        {
            return !(Double.IsInfinity(x) || Double.IsNaN(x));
        }

        /// <summary>
        /// Vector has no infinities or NaNs in it.
        /// </summary>
        public static bool IsProperVector(this Vector3 v)
        {
            return IsProperNumber(v.X) && IsProperNumber(v.Y) && IsProperNumber(v.Z);
        }

        /// <summary>
        /// Returns a vector with identical X and Z coordinates but which rests on the ground plane.
        /// </summary>
        /// <param name="v"></param>
        /// <returns></returns>
        public static Vector3 ProjectToGroundPlane(this Vector3 v)
        {
#if GroundAtZero
            v.Y = 0;
            return v;
#endif
        }

        /// <summary>
        /// Clips input in the range +/- max.
        /// </summary>
        public static float Saturate(float input, float max)
        {
            if (input < -max)
                return -max;
            else if (input > max)
                return max;
            else
                return input;
        }

        /// <summary>
        /// Truncates vector v to be no longer than maxLength.
        /// </summary>
        public static Vector3 Saturate(Vector3 v, float maxLength)
        {
            float len = v.Length();
            if (len > maxLength)
                return v * (maxLength / len);
            else
                return v;
        }

        /// <summary>
        /// Returns the component of v in the specified direction.
        /// Direction must be a unit vector.
        /// </summary>
        public static Vector3 ComponentInDirection(this Vector3 v, Vector3 directionUnitVector)
        {
            Debug.AssertUnitVector(directionUnitVector, "ComponentInDirection called with non-unit vector");
            return Vector3.Dot(v, directionUnitVector) * directionUnitVector;
        }

        /// <summary>
        /// Component of v in specified direction.  Direction need not be a unit vector.
        /// </summary>
        public static Vector3 ComponentInDirectionNonUnit(this Vector3 v, Vector3 directionVector)
        {
            return directionVector * Vector3.Dot(v, directionVector) * (1f / directionVector.LengthSquared());
        }

        /// <summary>
        /// Vector triple product
        /// </summary>
        public static Vector3 VTriple(Vector3 u, Vector3 v, Vector3 w)
        {
            return Vector3.Cross(u, Vector3.Cross(v, w));
        }

        /// <summary>
        /// Scalar triple product
        /// </summary>
        public static float STriple(Vector3 u, Vector3 v, Vector3 w)
        {
            return Vector3.Dot(u, Vector3.Cross(v, w));
        }

        static Random rand = new Random();
        /// <summary>
        /// Returns a random vector whose components are in the range +/- magnitude/2.
        /// </summary>
        public static Vector3 NoiseVector(float magnitude)
        {
            return new Vector3(magnitude * ((float)rand.NextDouble() - 0.5f), 0f, magnitude * ((float)rand.NextDouble() - 0.5f));
        }

        /// <summary>
        /// Normalize vector if non-zero.
        /// If vector is almost zero, it is set to 0.
        /// </summary>
        public static void SafeNormalize(ref Vector3 v)
        {
            float lenSq = v.LengthSquared();
            if (lenSq < 0.001f)
                v = Vector3.Zero;
            else
                v *= (float)(1 / Math.Sqrt(lenSq));
        }

        /// <summary>
        /// Normalize vector if non-zero.
        /// If vector is almost zero, it is set to 0.
        /// </summary>
        public static Vector3 SafeNormalize(this Vector3 v)
        {
            float lenSq = v.LengthSquared();
            if (lenSq < 0.001f)
                return Vector3.Zero;
            else
                return v * (float)(1 / Math.Sqrt(lenSq));
        }

        /// <summary>
        /// x squared
        /// </summary>
        public static float Square(float x)
        {
            Debug.AssertProper(x, "Square of a non-proper number");
            return x * x;
        }

        /// <summary>
        /// Difference between floats x and y
        /// </summary>
        public static float Difference(float x, float y)
        {
            return Math.Abs(x - y);
        }

        /// <summary>
        /// Returns a vector parallel to v of the specified length
        /// </summary>
        public static Vector3 SetLength(this Vector3 v, float length)
        {
            Debug.AssertProper(v, "Setting length of non-proper vector");
            return v * (length / v.Length());
        }

        /// <summary>
        /// A vector parallel to v with length no longer than max.
        /// </summary>
        public static Vector3 ClampLength(this Vector3 v, float max)
        {
            float l = v.Length();
            if (l <= max)
                return v;
            else
                return v * (max / l);
        }

        /// <summary>
        /// Distance between two vectors
        /// </summary>
        public static float Distance(Vector3 x, Vector3 y)
        {
            return (x - y).Length();
        }

        /// <summary>
        /// Square of the distance between two vectors
        /// </summary>
        public static float DistanceSquared(Vector3 x, Vector3 y)
        {
            return (x - y).LengthSquared();
        }

        /// <summary>
        /// Test whether the distance between the two vectors is less than the threshold
        /// </summary>
        public static bool DistanceLessThan(Vector3 x, Vector3 y, float threshold)
        {
            return DistanceSquared(x, y) < threshold * threshold;
        }

        /// <summary>
        /// Test whether the distance between the two vectors is greater than the threshold
        /// </summary>
        public static bool DistanceGreaterThan(Vector3 x, Vector3 y, float threshold)
        {
            return DistanceSquared(x, y) > threshold * threshold;
        }

        /// <summary>
        /// Creates a local->global transform matrix given the unit vectors
        /// of the object's local X, Y, and Z axes in global coordinates.
        /// </summary>
        public static Matrix ChangeOfBasis(Vector3 localX, Vector3 localY, Vector3 localZ)
        {
            Matrix m = Matrix.Identity;
            m.M11 = localX.X;
            m.M12 = localX.Y;
            m.M13 = localX.Z;
            m.M21 = localY.X;
            m.M22 = localY.Y;
            m.M23 = localY.Z;
            m.M31 = localZ.X;
            m.M32 = localZ.Y;
            m.M33 = localZ.Z;
            return m;
        }
        /// <summary>
        /// Creates a local->global transform matrix given the unit vectors
        /// of the object's local X, Y, and Z axes in global coordinates, and
        /// its origin in global coordinates
        /// </summary>
        public static Matrix ChangeOfBasis(Vector3 localX, Vector3 localY, Vector3 localZ, Vector3 origin)
        {
            Matrix m = Matrix.CreateTranslation(origin);
            m.M11 = localX.X;
            m.M12 = localX.Y;
            m.M13 = localX.Z;
            m.M21 = localY.X;
            m.M22 = localY.Y;
            m.M23 = localY.Z;
            m.M31 = localZ.X;
            m.M32 = localZ.Y;
            m.M33 = localZ.Z;
            return m;
        }

        /// <summary>
        /// Returns the index of the maximal element of array
        /// </summary>
        public static int ArgMax(this float[] a)
        {
            float max = a[0];
            int arg = 0;
            for (int i = 1; i<a.Length; i++)
                if (a[i] > max)
                {
                    max = a[i];
                    arg = i;
                }
            return arg;
        }

        /// <summary>
        /// Returns the index of the minimal element of array
        /// </summary>
        public static int ArgMin(this float[] a)
        {
            float min = a[0];
            int arg = 0;
            for (int i = 1; i < a.Length; i++)
                if (a[i] < min)
                {
                    min = a[i];
                    arg = i;
                }
            return arg;
        }
    }
}
