// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Twig
{
    /// <summary>
    /// A sheet of paper
    /// </summary>
    public class Paper : BoxModel
    {
        /// <summary>
        /// Makes a TwigObject representing a sheet of paper
        /// </summary>
        public Paper(TwigGame g, string name, string model, Vector3 position)
            : base(g, name, g.Content.Load<Model>(model), position, true)
        {
            Atlas["front"] = Atlas["top"];
        }

        Node LowerLeftCorner
        {
            get
            {
                return nodes[0];
            }
        }

        Node LowerRightCorner
        {
            get
            {
                return nodes[1];
            }
        }

        /// <summary>
        /// The normal to the reading side of the paper
        /// </summary>
        protected override Vector3 LocalY
        {
            get
            {
                // Paper is so thin the nodes on the front and back sides
                // can get pushed behind one another, so the standard way
                // of computing LocalY isn't reliable.  We compute it
                // instead from the X and Z axes.
                Vector3 up = Vector3.Cross(LocalZ, LocalX);
                up.Normalize();
                return up;
            }
        }

        /// <summary>
        /// Tells paper to position itself for reading by reader; also tells reader to grab paper.
        /// </summary>
        public override void SetupHoldForUse(Mammal reader, Grasp grasp)
        {
            SetupHold(reader, grasp);
            Place(reader.Atlas["torso"], new Vector3(0, 0.6f, 0.8f), "front", reader.headTop);
        }


        /// <summary>
        /// Tells paper to position itself for reading by reader; also tells reader to grab paper.
        /// </summary>
        public override void SetupHold(Mammal reader, Grasp grasp)
        {
            reader.Hold(LowerLeftCorner, LowerRightCorner, this, grasp);
            Place(null, Vector3.Zero, null, null);
        }

        /// <summary>
        /// Select hands and nodes for grasping.  Will choose a lower-corner.
        /// </summary>
        public override void SelectGrasp(Mammal grasper, Grasp graspRequest,
                                         out Grasp selectedGrasp, out Node leftTarget, out Node rightTarget)
        {
            grasper.SelectGrasp(LowerLeftCorner, LowerRightCorner, this, graspRequest,
                                out selectedGrasp, out leftTarget, out rightTarget);
        }

        /// <summary>
        /// Characters can walk over the paper.
        /// </summary>
        public override bool IsObstacle
        {
            get
            {
                return false;
            }
        }
    }
}
