﻿// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Slot = Twig.Mammal.AttentionBuffer.Slot;
using SlotSet = Twig.Mammal.AttentionBuffer.SlotSet;

namespace Twig
{
    /// <summary>
    /// Represents a unary predicate over objects in AttentionBuffer.
    /// </summary>
    [System.Diagnostics.DebuggerDisplay("{DebugString}")]
    public class Predicate
    {
        string name;

        /// <summary>
        /// Name of predicate for debugging purposes.
        /// </summary>
        public string Name
        {
            get { return name; }
        }

        SlotSet extension;

        /// <summary>
        /// The extension of the predicate as a SlotSet.
        /// </summary>
        public SlotSet Extension
        {
            get { return extension; }
        }

        GoalAssertion goal;

        /// <summary>
        /// GoalAssertion for this predicate, i.e. how strongly the system wants to achieve this
        /// predicate for each object in the attention buffer.
        /// </summary>
        public GoalAssertion Goal
        {
            get
            {
                if (goal == null)
                    goal = new GoalAssertion(name+".goal", this, owner);
                return goal;
            }
        }

        /// <summary>
        /// Sets the preconditions for whatever behavior(s) achieves this goal.
        /// </summary>
        public void DeclarePreconditions(params Predicate[] preconditions)
        {
            Goal.SetPreconditions(preconditions);
        }

        Mammal owner;

        /// <summary>
        /// Tests whether predicate is true of an object given its slot.
        /// </summary>
        public bool this[Slot s]
        {
            get
            {
                return extension[s];
            }
        }

        /// <summary>
        /// Tests predicate on this object.
        /// </summary>
        public bool this[TwigObject o]
        {
            get
            {
                return extension[owner.Attention.ObjectSlotNumber(o)];
            }
        }

        /// <summary>
        /// A procedure that can compute the extension (a slotset) of a predicate)
        /// </summary>
        /// <returns>slotset of the predicate's extension</returns>
        public delegate SlotSet UpdateThunk();

        /// <summary>
        /// A procedure that can test individual objects in working memory.
        /// </summary>
        public delegate bool TestProcedure(TwigObject o);

        /// <summary>
        /// A procedure that can test individual objects in working memory given their slots.
        /// </summary>
        public delegate bool SlotTestProcedure(Slot s);

        /// <summary>
        /// Creates a predicate given a procedure that can compute its extension
        /// </summary>
        public Predicate(string name, Mammal owner, UpdateThunk u)
        {
            this.name = name;
            this.owner = owner;
            updateThunk = u;
            owner.RegisterUpdate(new Mammal.InferenceThunk(Update));
        }

        /// <summary>
        /// Creates a predicate given a procedure to testing individual elements in working memory.
        /// </summary>
        public Predicate(string name, Mammal m, TestProcedure t)
            : this(name, m, delegate()
            {
                return m.Attention.TestSlots(t);
            })
        { }

        /// <summary>
        /// Creates a predicate given a procedure to testing individual elements givne their slots.
        /// </summary>
        public Predicate(string name, Mammal m, SlotTestProcedure t)
            : this(name, m, delegate()
            {
                int mask=0;
                for (int i = 0; i < Mammal.AttentionBuffer.AttentionCapacity; i++)
                    if (t(new Slot((byte)i)))
                        mask |= 1 << i;
                return new SlotSet((byte)mask);
            })
        { }

        UpdateThunk updateThunk;

        /// <summary>
        /// Forces recomputation of the predicate's extension.
        /// </summary>
        public void Update()
        {
            extension = updateThunk();
        }

        /// <summary>
        /// The negation of the predicate
        /// </summary>
        public static SlotSet operator ~(Predicate p)
        {
            return ~p.extension;
        }

        /// <summary>
        /// Conjunction of two predicates/slotsets
        /// </summary>
        public static SlotSet operator &(Predicate a, Predicate b)
        {
            return a.extension & b.extension;
        }

        /// <summary>
        /// Conjunction of two predicates/slotsets
        /// </summary>
        public static SlotSet operator &(Predicate a, SlotSet b)
        {
            return a.extension & b;
        }

        /// <summary>
        /// Conjunction of two predicates/slotsets
        /// </summary>
        public static SlotSet operator &(SlotSet a, Predicate b)
        {
            return a & b.extension;
        }

        /// <summary>
        /// Disjunction of two predicates/slotsets
        /// </summary>
        public static SlotSet operator |(Predicate a, Predicate b)
        {
            return a.extension | b.extension;
        }

        /// <summary>
        /// Disjunction of two predicates/slotsets
        /// </summary>
        public static SlotSet operator |(Predicate a, SlotSet b)
        {
            return a.extension | b;
        }

        /// <summary>
        /// Disjunction of two predicates/slotsets
        /// </summary>
        public static SlotSet operator |(SlotSet a, Predicate b)
        {
            return a | b.extension;
        }

        /// <summary>
        /// Renders the extension of the predicate as a string
        /// </summary>
        public string ExtensionAsString
        {
            get
            {
                StringBuilder b = new StringBuilder();
                bool gotone = false;
                for (int i = 0; i < Mammal.AttentionBuffer.AttentionCapacity; i++)
                {
                    Slot s = new Slot((byte)i);
                    if (extension[s])
                    {
                        TwigObject o = owner.Attention.SlotObject(s);
                        if (o != null)
                        {
                            if (gotone)
                                b.Append(", ");
                            else
                                b.Append('{');
                            b.Append(o.Name);
                            gotone = true;
                        }
                    }
                }
                if (gotone)
                {
                    b.Append('}');
                    return b.ToString();
                }
                else
                    return "{}";
            }
        }

        /// <summary>
        /// Shows value as a string.
        /// </summary>
        public string DebugString
        {
            get
            {
                return name + "=" + ExtensionAsString;
            }
        }
    }
}
