// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

#pragma warning disable 1591

namespace Twig
{
    /// <summary>
    /// The Ragdoll class implements the basic ragdoll physics
    /// for jointed Mammal figures, without any behavior
    /// or any angular constraints on joints.
    ///
    /// See the Mammal class for basic posture, locomotion and
    /// joint constraints.
    /// </summary>
    [Serializable]
    public class Ragdoll : TwigObject
    {
        public LimbPair arms, legs;
        public Node spineTop, spineJoint, spineBottom, headTop;
        public Link upperSpine, lowerSpine, head;
        public float legLength;
        public float armLength;
        public float spineLength;
        public float pelvisWidth;
        public float shoulderWidth;
        public float headHeight;
        float m_mass;

        /// <summary>
        /// TwigObject that is currently touching this Ragdoll
        /// If multiple contacts, no guarantees are made as to which object will be returned.
        /// </summary>
        public TwigObject contactObject = null;

        public Ragdoll(TwigGame g, string name, float spineLength, float spineRadius,
            float armSpacing, float armRadius, float upperArm, float foreArm,
            float pelvis, float legRadius, float femur, float shin, float headHeight)
            : base(g, name, true)
        {
            this.spineLength = spineLength;
            pelvisWidth = pelvis;
            spineTop = new Node(this, name+" spine top");
            spineBottom = new Node(this, name+" spine bottom");
            headTop = new Node(this, name+" head top");
            this.headHeight = headHeight;
            head = new Link(name+" head", this, spineTop, headTop, headHeight, 2f * armRadius);
            spineJoint = new Node(this, name+" spine joint");
            upperSpine = new Link(name+" upper spine", this, spineTop, spineJoint, spineLength*(2/3f), spineRadius);
            lowerSpine = new Link(name+" lower spine", this, spineJoint, spineBottom, spineLength/3f, spineRadius);
            arms = new LimbPair(this as Mammal, name+" arms", this, spineTop, armSpacing, armRadius, upperArm, foreArm);
            legs = new LimbPair(this as Mammal, name+" legs", this, spineBottom, pelvis, legRadius, femur, shin);
            armLength = upperArm + foreArm;
            legLength = femur + shin;
            shoulderWidth = armSpacing;

            arms.left.root.BodyReference = legs.left.root;
            arms.left.root.BodyCenteredDamping = 0.5f;
            arms.right.root.BodyReference = legs.right.root;
            arms.right.root.BodyCenteredDamping = 0.5f;

            // The elegant thing would be to use enumerators
            // to enumerate links, but the links get enumerated
            // a lot, and so if we call the enumerator directly
            // allocate about 1MB/sec just on the enumerators.
            // Since the set of links is constant, we enumerate it
            // once at object creation time and cache it in an
            // array so that for-each will generate a simple loop
            // with no malloc'ing.
            List<Link> links = new List<Link>();
            foreach (Link l in LinksInternal)
                links.Add(l);
            this.links = links.ToArray();

            // Construct node list from links list.
            List<Node> nodes = new List<Node>();
            foreach (Link l in Links)
            {
                if (!nodes.Contains(l.node1))
                    nodes.Add(l.node1);
                if (!nodes.Contains(l.node2))
                    nodes.Add(l.node2);
            }
            this.nodes = nodes.ToArray();

            List<CollisionElement> ces = new List<CollisionElement>();
            foreach (Link l in Links)
            {
                ces.Add(l.CollisionElement);
            }
            collisionElements = ces.ToArray();

            ComputeMasses();
        }

#if MDX
        public override Material DefaultMaterial
        {
            get
            {
                Material m = new Material();
                m.DiffuseColor = new ColorValue(100, 0, 100);
                m.AmbientColor = m.DiffuseColor; // new ColorValue(100, 0, 0);
                //m.Ambient = m.Diffuse;
                //m.AmbientColor = m.DiffuseColor;
                return m;
                
            }
        }
#endif

        /// <summary>
        /// Updates the masses of the nodes by computing the masses of the links under the
        /// assumption they have the density of water, then distributing that mass among the nodes.
        /// </summary>
        void ComputeMasses()
        {
            float totalMass = 0;

            //  Nodes default to unit mass.
            foreach (Node n in Nodes)
            {
                n.Mass = 0;
            }

            foreach (Link l in Links)
            {
                // Volume of link in cubic meters
                float linkVolume = l.Length * 2 * (float)Math.PI * l.radius * l.radius;
                // Mass of link in kilograms, assuming it's made of water
                // The multiplications convert from m^3 to cm^3;  Water density is roughly 1g/cm, so don't
                // need a conversion to get to grams.  We then divide by 1000 to get KG.
                float linkMass = (linkVolume * 100 * 100 * 100) / 1000.0f;

                totalMass += linkMass;

                // Divide the link's mass evently between its two nodes
                l.node1.Mass += linkMass/2f;
                l.node2.Mass += linkMass/2f;
            }

            m_mass = totalMass;
#if DEBUG
            float recomputedTotalMass=0;
            foreach (Node n in Nodes)
                recomputedTotalMass += n.Mass;
            System.Diagnostics.Debug.Assert(Math.Abs(totalMass-recomputedTotalMass)<1f, "Mismatch between actual mass and computed total mass");
#endif
        }

        public override void Draw(GameTime gameTime)
        {
            //TwigGame.DrawDebugMarker(CenterOfMass + headTop.Position.Y * Vector3.UnitY, Color.Green);
            ConfigureMaterial();
            BasicEffect.Begin();
            foreach (EffectPass pass in BasicEffect.CurrentTechnique.Passes)
            {
                pass.Begin();

                head.Draw();
                arms.Draw();
                upperSpine.Draw();
                lowerSpine.Draw();
                legs.Draw();
                base.Draw(gameTime);
                pass.End();
            }
            BasicEffect.End();
        }

        public override void  Update(GameTime gameTime)
        {
            arms.Update();
            legs.Update();
            spineBottom.Update();
            spineBottom.Update();
            spineJoint.Update();
            upperSpine.Update();
            lowerSpine.Update();
            head.Update();
            base.Update(gameTime);
        }

        public override void Constrain()
        {
            for (int i = 0; i < 2; i++)
            {
                arms.Constrain();
                legs.Constrain();
                upperSpine.Constrain();
                lowerSpine.Constrain();
                head.Constrain();
            }
            UpdateCollisionRadius();
#if OldCollisions
            CheckCollisions();
#endif
            // Do a limited check for self-collisions
            Link.TestCollision(upperSpine, arms.left.lower);
            Link.TestCollision(upperSpine, arms.right.lower);
            Link.TestCollision(lowerSpine, arms.left.lower);
            Link.TestCollision(lowerSpine, arms.right.lower);
            //Link.TestCollision(spine, arms.left.upper);
            //Link.TestCollision(spine, arms.right.upper);
            Link.TestCollision(arms.left.lower, legs.left.upper);
            Link.TestCollision(arms.right.lower, legs.right.upper);
            //Link.TestCollision(arms.left.lower, legs.right.upper);
            //Link.TestCollision(arms.right.lower, legs.left.upper);

        }

#if OldCollision
        void CheckCollisions()
        {
            contactObject = null;
            foreach (TwigObject o in TwigObject.AllTwigObjects)
                if (o != this)
                    if (o.TestCollisionWithRagdoll(this))
                        contactObject = o;
        }


        public override bool TestCollisionWithRagdoll(Ragdoll h)
        {
            Vector3 offset = h.Position - Position;
            bool contact = false;
            // Don't bother if they're too far apart
            if (offset.Length() < 2 * (legLength + spineLength))
            {
                // Ick.  Collide every link with every other link
                foreach (Link l in Links)
                {
                    foreach (Link hl in h.Links)
                        if (l.CheckCollision(hl, h))
                            contact = true;
                }
            }
            return contact;
        }
#endif

        /// <summary>
        /// Enumerates all the Link objects in the Ragdoll.
        /// Used only to build the real Links array.
        /// </summary>
        protected IEnumerable<Link> LinksInternal
        {
            get
            {
                foreach (Link l in arms.Links)
                    yield return l;
                foreach (Link l in legs.Links)
                    yield return l;
                yield return upperSpine;
                yield return lowerSpine;
                yield return head;
            }
        }

        // The elegant thing would be to use enumerators
        // to enumerate links, but the links get enumerated
        // a lot, and so if we call the enumerator directly
        // allocate about 1MB/sec just on the enumerators.
        // Since the set of links is constant, we enumerate it
        // once at object creation time and cache it in an
        // array so that for-each will generate a simple loop
        // with no malloc'ing.
            
        Link[] links;
        /// <summary>
        /// Enumerates all the (solid) Link objects in the Ragdoll
        /// for purposes of collision detection
        /// </summary>
        public Link[] Links
        {
            get { return links; }
        }

        public override Vector3 Position
        {
            get { return spineTop.Position; }
        }

        /// <summary>
        /// Center of Mass in global coordinates
        /// </summary>
        public override Vector3 CenterOfMass
        {
            get {
                Vector3 sum = Vector3.Zero;
                foreach (Node n in Nodes)
                    sum += n.Mass*n.Position;
                Vector3 com = sum / Mass;
#if DEBUG
                Debug.AssertProper(com, "Center of Mass is not a proper vector");
                if (Physics.IsSettled)
                    Debug.Assert((com - Position).LengthSquared() < 16, "Impossibly large difference between position and center of mass");
#endif
                return com;
            }
        }

        public override float Mass
        {
            get { return m_mass; }
        }

        public override Vector3 LookPoint
        {
            get { return (headTop.Position+spineTop.Position)*0.5f; }
        }

#if DEFER_TELEPORT
        public void Teleport(Vector3 position)
        {
            Vector3 offset = new Vector3(0, 0, pelvisWidth / 2);
            legs.left.end.Position = position - offset;
            legs.left.end.ControlMode = ControlMode.Fixed;
            legs.right.end.Position = position + offset;
            legs.right.end.ControlMode = ControlMode.Fixed;
            for (int i = 0; i < 100; i++)
            {
                Update();
                Constrain();
            }
            legs.left.end.ControlMode = ControlMode.Free;
            legs.right.end.ControlMode = ControlMode.Free;
        }
#endif

        void PlaceNode(Node n, Vector3 position)
        {
            n.ControlMode = ControlMode.Free;
            n.Position = position;
            n.Velocity = Vector3.Zero;
        }

        void PlaceLimb(Limb l, Vector3 position)
        {
            PlaceNode(l.end, position);
            position += l.lower.Length*Vector3.Up;
            PlaceNode(l.joint, position);
            position += l.upper.Length * Vector3.Up;
            PlaceNode(l.root, position);
        }

        public virtual void PoseAt(Vector3 position, Vector3 facingDirection)
        {
            Vector3 bodyRight = Vector3.Cross(Vector3.UnitY, facingDirection);
            bodyRight.Normalize();
            PlaceLimb(legs.left, position - (bodyRight * pelvisWidth / 2));
            PlaceLimb(legs.right, position + (bodyRight * pelvisWidth / 2));
            position += legs.left.length * Vector3.Up;
            PlaceNode(spineBottom, position);
            PlaceNode(spineJoint, position + (0.3f*spineLength * Vector3.Up));
            position += spineLength * Vector3.Up;
            PlaceNode(spineTop, position);
            PlaceLimb(arms.left, (position - arms.left.length*Vector3.Up) - (bodyRight * shoulderWidth / 2));
            PlaceLimb(arms.right, (position - arms.left.length * Vector3.Up) + (bodyRight * shoulderWidth / 2));
            position += headHeight * Vector3.Up;
            PlaceNode(headTop, position);

            // The above places the hands inside the femurs, so we need to spread them out slightly
            arms.left.end.Position = arms.left.end.Position - (arms.left.lower.radius + legs.left.upper.radius) * 0.6f * bodyRight;
            arms.left.end.Velocity = Vector3.Zero;
            arms.right.end.Position = arms.right.end.Position + (arms.right.lower.radius + legs.right.upper.radius) * 0.6f * bodyRight;
            arms.right.end.Velocity = Vector3.Zero;
        }


        /// <summary>
        /// Fix Ragdoll at specified position, letting the rest of the body hang from there.
        /// </summary>
        /// <param name="position">Position for base of neck</param>
        public void HangAt(Vector3 position)
        {
            arms.left.root.Position = position + new Vector3(-arms.spacing / 2.0f, 0, 0);
            arms.left.root.ControlMode = ControlMode.Fixed;
            arms.right.root.Position = position + new Vector3(arms.spacing / 2.0f, 0, 0);
            arms.right.root.ControlMode = ControlMode.Fixed;
        }

        /// <summary>
        /// Cancels effects of HangAt()
        /// </summary>
        public void UnHang()
        {
            arms.left.root.ControlMode = ControlMode.Free;
            arms.right.root.ControlMode = ControlMode.Free;
        }

        float collisionRadius = 1;
        public void UpdateCollisionRadius()
        {
            float maxSquared = 0;
            Vector3 position = Position;
            foreach (Node n in Nodes)
            {
                float squared = MathUtil.ProjectToGroundPlane(n.Position - position).LengthSquared();
                if (squared > maxSquared)
                    maxSquared = squared;
            }
            collisionRadius = (float)Math.Sqrt(maxSquared);
        }
        public override float CollisionRadius
        {
            get { return collisionRadius; }
        }

        /// <summary>
        /// The character's pose
        /// </summary>
        public Pose Pose
        {
            get
            {
                return new Pose(this);
            }
            set
            {
                for (int i = 0; i < nodes.Length; i++)
                {
                    nodes[i].Position = value.NodePositions[i];
                    nodes[i].Velocity = Vector3.Zero;
                }
            }
        }
    }
}