// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using Microsoft.Xna.Framework;
using System.Reflection;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;

namespace Twig
{
    /// <summary>
    /// Contains definitions for executing commands in ScriptScript
    /// </summary>
    public class ScriptScriptInterpreter : GameComponent
    {
        /// <summary>
        /// The TwigGame this interpreter is running for.
        /// </summary>
        protected TwigGame game;

        /// <summary>
        /// MessageControl delegate for any Action objects that are executed
        /// This is filled in by the TwigServer constructor, or else left null.
        /// </summary>
        protected Action.MessageControl messageControl = null;

        /// <summary>
        /// RequestID of the current Execute() call, if running in a server.
        /// </summary>
        protected int currentRequestID = -1;

        /// <summary>
        /// Constructs a new interpreter
        /// </summary>
        /// <param name="g"></param>
        protected ScriptScriptInterpreter(TwigGame g) : base(g)
        {
            game = g;
        }

        class ScriptException : Exception
        {
            public ScriptException(string message) : base(message) { }
        }

        /// <summary>
        /// Makes a new instance of a script interpreter
        /// </summary>
        public ScriptScriptInterpreter(TwigGame g, string assetName) : base(g)
        {
            game = g;
            script = new System.IO.StringReader(g.Content.Load<string>(assetName));
        }

        TextReader script;

        /// <summary>
        /// Output stream for status responses
        /// </summary>
        protected TextWriter output;
        /// <summary>
        /// The Action on which the script is waiting
        /// </summary>
        public Action blockingAction;
        List<Action> pendingActions = new List<Action>();

        /// <summary>
        /// Execute further commands, if appropriate
        /// </summary>
        public override void Update(GameTime gameTime)
        {
            pendingActions.RemoveAll(delegate(Action a)
            {
                return a.Status != ActionStatus.Running;
            });
            foreach (Action a in pendingActions)
                a.Update();

            if (blockingAction != null)
            {
                switch (blockingAction.Status)
                {
                    case ActionStatus.Running:
                        return;

                    case ActionStatus.Succeed:
                        blockingAction = null;
                        break;

                    case ActionStatus.Fail:
                        throw new ScriptException("Action failed");
                }
            }

            if (blockingAction == null)
            {
                SkipWhiteSpace(script);
                string command = script.ReadLine();
                if (command != null && command != "")
                    Execute(command);
            }
        }

        /// <summary>
        /// Execute a command in ScriptScript
        /// </summary>
        public void Execute(TextReader reader)
        {
            SkipWhiteSpace(reader); 
            string first = ReadName(reader);
            SkipWhiteSpace(reader);
            if (reader.Peek() == (int)':')
            {
                reader.Read();   // Skip colon
                ExecuteCharacterCommand(first, reader);
            }
            else
                ExecuteSystemCommand(first, reader);
        }

        /// <summary>
        /// Execute a command in ScriptScript
        /// </summary>
        public void Execute(string command)
        {
            Execute(new StringReader(command));
        }

        void ExecuteCharacterCommand(string characterName, TextReader reader)
        {
            object character = ObjectNamed(characterName);
            SkipWhiteSpace(reader);
            string command = ((char)reader.Peek()=='"')?"say":ReadName(reader);
            SkipWhiteSpace(reader);
            if (reader.Peek() == (int)'=')
            {
                reader.Read();
                SkipWhiteSpace(reader);
                // field or property assignment
                List<object> values = ReadArglist(reader);
                if (command == "color" && values[0] is Vector3)
                    values[0] = new Color((byte)((Vector3)values[0]).X, (byte)((Vector3)values[0]).Y, (byte)((Vector3)values[0]).Z);
                if (values.Count != 1)
                    throw new ScriptException("Extra values included in = command");
                if (character is Type)
                    ((Type)character).InvokeMember(command,
                        BindingFlags.Static | BindingFlags.IgnoreCase
                        | BindingFlags.SetField | BindingFlags.SetProperty | BindingFlags.Public | BindingFlags.NonPublic,
                        null, character, new object[1] { values[0] });
                else
                    character.GetType().InvokeMember(command,
                        BindingFlags.Instance | BindingFlags.IgnoreCase
                        | BindingFlags.SetField | BindingFlags.SetProperty | BindingFlags.Public | BindingFlags.NonPublic,
                        null, character, new object[1] { values[0] });
            }
            else
            {
                // Method call
                List<object> args = ReadArglist(reader);
                object result = character.GetType().InvokeMember(command, BindingFlags.InvokeMethod | BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.IgnoreCase,
                    null, character, args.ToArray());
                if (result != null && result is Action)
                {
                    Action a = result as Action;
                    a.messageControl = this.messageControl;
                    a.requestID = currentRequestID;
                    a.Start();
                    pendingActions.Add(a);
                    // Block on the action unless the line ends with an &.
                    if (reader.Peek() != (int)'&')
                        blockingAction = a;
                }
            }
        }

        void ExecuteSystemCommand(string command, TextReader reader)
        {
            switch (command)
            {
                case "new":
                    ExecuteNew(reader);
                    break;

                case "pause":
                    ExecutePause(reader);
                    break;

                case "exit":
                    game.Exit();
                    break;

                case "ls":
                    ExecuteLS();
                    break;

                default:
                    throw new ScriptException("Unknown system command: " + command);
            }
        }

        void ExecuteNew(TextReader r)
        {
            Type t = ReadType(r);
            List<object> args = ReadArglist(r);
            args.Insert(0, game);      

            object o = t.InvokeMember(null, System.Reflection.BindingFlags.CreateInstance, null, null, args.ToArray());
            GameComponent c = o as GameComponent;
            if (c!=null)
                game.Components.Add(c);
        }

        void ExecutePause(TextReader r)
        {
            List<object> args = ReadArglist(r);

            if (args.Count == 1 && args[0] is float)
            {
                blockingAction = new Sleep((float)args[0]);
                blockingAction.Start();
                pendingActions.Add(blockingAction);
            }
            else
                throw new ScriptException("Invalid argument list for pause (should have one argument, a number)");
        }

        void ExecuteLS()
        {
            foreach (TwigObject o in TwigObject.AllTwigObjects)
            {
                output.Write(o.Name);
                output.Write(',');
            }
            output.WriteLine();
            output.Flush();
        }

        string ReadName(TextReader r)
        {
            StringBuilder b = new StringBuilder();
            while (char.IsLetterOrDigit((char)r.Peek()))
            {
                b.Append((char)r.Read());
            }
            return b.ToString();
        }

        object ReadNamedObject(TextReader r)
        {
            return ObjectNamed(ReadName(r));
        }

        object ObjectNamed(string name)
        {
            switch (name)
            {
                case "null":
                    return null;

                case "true":
                    return true;

                case "false":
                    return false;


                default:
                    if (game.NamedObjects.ContainsKey(name))
                        return game.NamedObjects[name];
                    else
                    {
                        // It's limiting to prefix it with Twig. but if we don't then someone could
                        // conceivably alter a script file to say 'System.Directory: delete "C:\" true'
                        // which would be um, uh, unfortunate ...
                        Type probe = Type.GetType("Twig."+name);
                        if (probe != null)
                            return probe;
                        else
                            throw new ScriptException("Unknown object: \"" + name + "\"");
                    }
            }
        }

        string ReadQuotedString(TextReader r)
        {
            StringBuilder b = new StringBuilder();

            Debug.Assert(r.Peek() == '"', "Reading quoted string, but it doesn't start with a quote");
            r.Read();
            while (r.Peek() > 0 && r.Peek() != '"')
            {
                char c = (char)r.Read();
                if (c == '\\')
                {
                    c = (char)r.Read();
                    switch (c)
                    {
                        case '\\':
                            break;
                        case 'n':
                            c = '\n';
                            break;

                        case '\"':
                            c = '\"';
                            break;

                        default:
                            throw new ScriptException("Illegal escape sequence: \\" + c);
                    }
                }
                b.Append(c);
            }
            // Skip close quote
            r.Read();
            return b.ToString();
        }

        /// <summary>
        /// Reads an ASCII decimal number from r.
        /// </summary>
        protected float ReadNumber(TextReader r)
        {
            double sign = 1;
            double divisor = 1;
            double undivided = 0;
            bool seenDecimalPoint = false;
            while (char.IsDigit((char)r.Peek()) || r.Peek() == (int)'-' || r.Peek() == (int)'.')
            {
                char c = (char)r.Read();
                if (char.IsDigit(c))
                {
                    undivided = undivided * 10 + c - '0';
                    if (seenDecimalPoint)
                        divisor *= 10;
                }
                else if (c == '-')
                    sign = -1;
                else if (c == '.')
                    seenDecimalPoint = true;
                else
                    Debug.Assert(false, "Impossible character inside ScriptScript.ReadNumber()");
            }
            return (float)(sign * undivided / divisor);
        }

        Type ReadType(TextReader r)
        {
            SkipWhiteSpace(r);
            string name = ReadName(r);
            Type t = Type.GetType("Twig."+name);
            if (t == null)
                throw new ScriptException("Unknown type: Twig." + t);
            else
                return t;
        }

        object ReadVector(TextReader r)
        {
            List<float> numbers = new List<float>();
            Debug.Assert((char)r.Peek() == '(', "Reading vector, but it doesn't start with an '('");
            // Skip open paren
            r.Read();
            SkipWhiteSpace(r);
            while (r.Peek()=='-' || char.IsDigit((char)r.Peek()) || r.Peek()==(int)'.')
            {
                numbers.Add(ReadNumber(r));
                if (numbers.Count > 3)
                {
                    int next = r.Peek();
                    char c = (char)next;
                    throw new ScriptException("Vector seems to have too many components");
                }
                SkipWhiteSpace(r);
            }
            if (r.Peek() == ')')
            {
                r.Read(); // Skip over close paren
                switch (numbers.Count)
                {
                    case 2:
                        return new Vector2(numbers[0], numbers[1]);
                        
                    case 3:
                        return new Vector3(numbers[0], numbers[1], numbers[2]);

                    default:
                        throw new ScriptException("Vector had an incorrect number of components in it: " + numbers.Count.ToString());
                }
            }
            else
                throw new ScriptException("Vector in script didn't end with a ')'");
        }

        List<object> ReadArglist(TextReader r)
        {
            List<object> list = new List<object>();

            SkipWhiteSpace(r);
            while (r.Peek() > 0)
            {
                char c = (char)r.Peek();
                if (char.IsDigit(c) || c == '-')
                    list.Add(ReadNumber(r));
                else if (c == '(')
                    list.Add(ReadVector(r));
                else if (char.IsLetterOrDigit(c))
                    list.Add(ReadNamedObject(r));
                else if (c == '"')
                    list.Add(ReadQuotedString(r));
                else if (c == '&')
                    // There's an & on the end of the command
                    break;
                else
                    throw new ScriptException("Bad object in script argument list");
                SkipWhiteSpace(r);
            }
            return list;
        }

        /// <summary>
        /// Skips over any whitespace in r
        /// </summary>
        protected void SkipWhiteSpace(TextReader r)
        {
            while (char.IsWhiteSpace((char)r.Peek()))
                r.Read();
        }
    }
}
