// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Twig
{
    /// <summary>
    /// Triangle mesh consisting of a single triangle list, strip, or fan.
    /// </summary>
    public class SimpleMesh
    {
        /// <summary>
        /// Buffer for the mesh's vertices
        /// </summary>
        protected VertexBuffer vertexBuffer;
        /// <summary>
        /// Number of triangles in mesh
        /// </summary>
        protected int count;
        /// <summary>
        /// Mesh format (TriangleList, TriangleStrip, TriangleFan).
        /// </summary>
        protected PrimitiveType type;

        static VertexDeclaration vertexDeclaration;

        /// <summary>
        /// Creates a SimpleMesh
        /// </summary>
        public SimpleMesh(GraphicsDevice device, PrimitiveType type, VertexPositionNormalTexture[] verticies, int count) {
            this.count = count;
            this.type = type;
            vertexBuffer = new VertexBuffer(device, typeof(VertexPositionNormalTexture), verticies.Length, BufferUsage.WriteOnly);
            vertexBuffer.SetData<VertexPositionNormalTexture>(verticies);
            vertexDeclaration = new VertexDeclaration(device,VertexPositionNormalTexture.VertexElements);
        }

        /// <summary>
        /// Makes a SimpleMesh from a MeshBuilder
        /// </summary>
        /// <param name="device"></param>
        /// <param name="builder"></param>
        public SimpleMesh(GraphicsDevice device, MeshBuilder<VertexPositionNormalTexture> builder)
        {
            this.count = builder.Count-2;
            this.type = PrimitiveType.TriangleStrip;
            vertexBuffer = builder.MakeVertexBuffer(device, BufferUsage.WriteOnly);
            vertexDeclaration = new VertexDeclaration(device, VertexPositionNormalTexture.VertexElements);
        }


        /// <summary>
        /// Draws the mesh on its device.
        /// </summary>
        public void Draw() {
            GraphicsDevice device = vertexBuffer.GraphicsDevice;
            // Uncomment to disable backface culling to test meshes.
            //device.RenderState.CullMode = CullMode.None;
            device.VertexDeclaration = vertexDeclaration;
            device.Vertices[0].SetSource(vertexBuffer, 0, VertexPositionNormalTexture.SizeInBytes);
            device.DrawPrimitives(type, 0, count);
        }

        /// <summary>
        /// Makes a triangle fan mesh.
        /// </summary>
        public static SimpleMesh TriangleFan(GraphicsDevice device, VertexPositionNormalTexture[] verticies)
        {
            return new SimpleMesh(device, PrimitiveType.TriangleFan, verticies, verticies.Length-2);
        }

        /// <summary>
        /// Makes a triangle strip mesh.
        /// </summary>
        public static SimpleMesh TriangleStrip(GraphicsDevice device, VertexPositionNormalTexture[] verticies)
        {
            return new SimpleMesh(device, PrimitiveType.TriangleStrip, verticies, verticies.Length-2);
        }

        static Vector3 XPLUS = new Vector3(1,  0,  0);	// X
        static Vector3 XMIN  = new Vector3(-1,  0,  0);	// -X
        static Vector3 YPLUS = new Vector3(0,  1,  0);	//  Y
        static Vector3 YMIN  = new Vector3(0, -1,  0);	// -Y
        static Vector3 ZPLUS = new Vector3(0,  0,  1);	//  Z
        static Vector3 ZMIN  = new Vector3(0,  0, -1);	// -Z


        /// <summary>
        /// Creates a SimpleMesh containing a sphere.
        /// </summary>
        public static SimpleMesh Sphere(GraphicsDevice device, float radius, int level, float normalNoiseLevel)
        {
            List<VertexPositionNormalTexture> vertices = new List<VertexPositionNormalTexture>();
            // Vertices of a unit octahedron
            Triangulate(vertices, radius, level, XPLUS, ZPLUS, YPLUS, normalNoiseLevel);
            Triangulate(vertices, radius, level, YPLUS, ZPLUS, XMIN, normalNoiseLevel);
            Triangulate(vertices, radius, level, XMIN, ZPLUS, YMIN, normalNoiseLevel);
            Triangulate(vertices, radius, level, YMIN, ZPLUS, XPLUS, normalNoiseLevel);
            Triangulate(vertices, radius, level, XPLUS, YPLUS, ZMIN, normalNoiseLevel);
            Triangulate(vertices, radius, level, YPLUS, XMIN, ZMIN, normalNoiseLevel);
            Triangulate(vertices, radius, level, XMIN, YMIN, ZMIN, normalNoiseLevel);
            Triangulate(vertices, radius, level, YMIN, XPLUS, ZMIN, normalNoiseLevel);

            return new SimpleMesh(device, PrimitiveType.TriangleList, vertices.ToArray(), vertices.Count / 3);
        }

        static void Triangulate(List<VertexPositionNormalTexture> vertices, float radius, int level, Vector3 p1, Vector3 p2, Vector3 p3, float normalNoiseLevel)
        {
            p1 *= radius / p1.Length();
            p2 *= radius / p2.Length();
            p3 *= radius / p3.Length();

            if (level == 0)
            {
                Vector3 n1 = p1 + MathUtil.NoiseVector(radius*normalNoiseLevel);
                Vector3 n2 = p1 + MathUtil.NoiseVector(radius * normalNoiseLevel);
                Vector3 n3 = p1 + MathUtil.NoiseVector(radius * normalNoiseLevel);
                n1.Normalize();
                n2.Normalize();
                n3.Normalize();

                vertices.Add(new VertexPositionNormalTexture(p1, n1, Vector2.Zero));
                vertices.Add(new VertexPositionNormalTexture(p2, n2, Vector2.Zero));
                vertices.Add(new VertexPositionNormalTexture(p3, n3, Vector2.Zero));
            }
            else
            {
                Vector3 m12 = (p1 + p2) * 0.5f;
                Vector3 m23 = (p2 + p3) * 0.5f;
                Vector3 m31 = (p3 + p1) * 0.5f;

                Triangulate(vertices, radius, level - 1, p1, m12, m31, normalNoiseLevel);
                Triangulate(vertices, radius, level - 1, m12, p2, m23, normalNoiseLevel);
                Triangulate(vertices, radius, level - 1, m31, m23, p3, normalNoiseLevel);
                Triangulate(vertices, radius, level - 1, m12, m23, m31, normalNoiseLevel);
            }
        }
    }
}
