// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace Twig
{
    /// <summary>
    /// Displays performance information and debug string for current DebugObject
    /// </summary>
    public class StandardHUD : HUD
    {
        /// <summary>
        /// Displays performance information and debug string for current DebugObject
        /// </summary>
        public StandardHUD(TwigGame g) : this(g, Keys.None, false) { }

        /// <summary>
        /// Displays performance information and debug string for current DebugObject
        /// </summary>
        public StandardHUD(TwigGame game, Keys hotKey, bool display)
            : base(game, hotKey, "Toggle display of debugging information on selected character", display)
        {
#if DEBUG
            age = new ushort[20];
            for (int i = 0; i < age.Length; i++)
                age[i] = ushort.MaxValue;
#endif
        }

#if DEBUG
        ushort[] age;
        Dictionary<string,int> slots = new Dictionary<string,int>();

        void NewFrame()
        {
            for (int i = 0; i < age.Length; i++)
                if (age[i] < ushort.MaxValue)
                    age[i]++;
        }

        int GetScreenSlot(string message)
        {
            if (slots.ContainsKey(message))
            {
                int s = slots[message];
                if (age[s]==0)
                    // It's already in use for this frame.
                    return AllocateSlot(message);
                else
                {
                    age[s] = 0;
                    return s;
                }
            }
            else
                return AllocateSlot(message);
        }

        int AllocateSlot(string message) {
            int oldest = 0;
            int oldestAge = age[oldest];
            for (int i = 0; i < age.Length; i++)
            {
                if (age[i] > oldestAge)
                {
                    oldest = i;
                    oldestAge = age[i];
                }
            }
            age[oldest] = 0;
            slots[message] = oldest;
            return oldest;
        }
#endif

        /// <summary>
        /// Draws the HUD using the game's HUDEffect.
        /// </summary>
        public override void Draw(GameTime gameTime)
        {
            HUDEffect.Begin();
            foreach (EffectPass pass in HUDEffect.CurrentTechnique.Passes)
            {
                pass.Begin();
                HUDSprites.Begin();
                HUDSprites.DrawString(HUDFont,
                String.Format("{0}frame {1}, {2:0.0}ms update, {3:0.0}ms render",
                Physics.SlowMo ? "(SlowMo) " : "", Physics.tickCount,
                TwigGame.UpdateTime * 1000, TwigGame.RenderTime * 1000),
                new Vector2(550, 10), Color.White);
                if (TwigGame.DebugObject != null)
                    HUDSprites.DrawString(HUDFont, TwigGame.DebugObject.DebugString, new Vector2(550, 25), Color.White);
            //HUDSprites.DrawString(HUDFont, device.Viewport.Height.ToString(), new Vector2(600, 25), Color.White);

                // Draw the waitList information for the character, if any
                // This list is generated during Update(), so it should be complete by the time draw is
                // run, provided we aren't multithreaded.
                
#if DEBUG
                Mammal character = TwigGame.DebugObject as Mammal;
                if (character != null && character.waitList.Count > 0)
                {
                    NewFrame();
                    HUDSprites.DrawString(HUDFont, "Waiting for:",
                            new Vector2(0, 0), Color.White);
                    for (int i = 0; i < character.waitList.Count; i++)
                        HUDSprites.DrawString(HUDFont, character.waitList[i],
                            new Vector2(0, (GetScreenSlot(character.waitList[i]) + 1) * HUDFont.LineSpacing), Color.White);
                }
#endif
                HUDSprites.End();
                pass.End();
            }
            HUDEffect.End();
        }
    }
}
