// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Input;
using System.Security.Permissions;
using System.Diagnostics;
//[assembly:CLSCompliant(false)]
//[assembly:FileIOPermission(SecurityAction.RequestMinimum)]

namespace Twig
{
    /// <summary>
    /// Bucky bits (shift, control, alt)
    /// </summary>
    [Flags]
    public enum KeyModifiers
    {
        /// <summary>
        /// No keys held down
        /// </summary>
        None = 0,
        /// <summary>
        /// The shift key (left or right) held down.
        /// </summary>
        Shift = 1,
        /// <summary>
        /// Control key (left or right) held down
        /// </summary>
        Control = 2,
        /// <summary>
        /// Alt key held down (left or right)
        /// </summary>
        Alt = 4
    }

    /// <summary>
    /// Base class for games that use the Twig engine.  Specializes Microsoft.Xna.Framework.Game.
    /// </summary>
    public abstract partial class TwigGame : Game
    {
        #region Instance variables
        /// <summary>
        /// Font for use in the heads-up display.
        /// </summary>
        public SpriteFont HUDFont;
        /// <summary>
        /// SpriteBatch used by the HUD
        /// </summary>
        public SpriteBatch HUDSprites;

        /// <summary>
        /// Font to use for Dialog
        /// </summary>
        public SpriteFont DialogFont;
        /// <summary>
        /// Used for drawing DialogBalloons
        /// </summary>
        public SpriteBatch DialogSprites;

        /// <summary>
        /// Graphics manager
        /// </summary>
        protected GraphicsDeviceManager graphics;
        /// <summary>
        /// GraphicsDevice used by the game
        /// </summary>
        private GraphicsDevice device;

        /// <summary>
        /// GraphicsDevice for use in rendering game
        /// </summary>
        public GraphicsDevice Device
        {
            get { return device; }
            set { device = value; }
        }
        /// <summary>
        /// View matrix for the current viewpoint.
        /// </summary>
        private Matrix viewMatrix;

        /// <summary>
        /// The view matrix used for rendering.
        /// Updated automatically from CameraPosition, CameraTargetPosition, CameraUp, and CameraFOV.
        /// </summary>
        public Matrix ViewMatrix
        {
            get { return viewMatrix; }
            //set { viewMatrix = value; }
        }
        /// <summary>
        /// Projection matrix
        /// </summary>
        private Matrix projectionMatrix;

        /// <summary>
        /// The projection matrix used for rendering.
        /// Updated automatically from CameraPosition, CameraTargetPosition, CameraUp, and CameraFOV.
        /// </summary>
        public Matrix ProjectionMatrix
        {
            get { return projectionMatrix; }
            //set { projectionMatrix = value; }
        }
        /// <summary>
        /// BasicEffect used for rendering cylinders in models
        /// </summary>
        private BasicEffect basicEffect;

        /// <summary>
        /// The BasicEffect to be used in rendering TwigObjects.
        /// </summary>
        public BasicEffect BasicEffect
        {
            get { return basicEffect; }
            set { basicEffect = value; }
        }
        /// <summary>
        /// BasicEffect used for rendering the HUD
        /// </summary>
        public BasicEffect HUDEffect;
        float mFillLevel = 0.3f;

        DialogBalloon.DialogBalloonManager dialogBalloonManager;

        /// <summary>
        /// Manager for Dialog Balloons in this game.
        /// </summary>
        internal DialogBalloon.DialogBalloonManager DialogBalloonManager
        {
            get { return dialogBalloonManager; }
        }

        bool pause=false;
        /// <summary>
        /// Indicates whether TwigObject update has been paused.
        /// </summary>
        public bool Pause
        {
            get { return pause; }
            set {
                pause = value;
#if NodeLogging
                if (!pause)
                    Node.PlaybackLog = false;
#endif
            }
        }
        bool singleStep = false;

        /// <summary>
        /// If true, game will pause after this update cycle.
        /// </summary>
        public bool SingleStep
        {
            get { return singleStep; }
            set { singleStep = value; }
        }

        /// <summary>
        /// Proportion of overall light level devoted to fill light as opposed to key light
        /// </summary>
        public float FillLevel
        {
            get { return mFillLevel; }
            set { mFillLevel = value; }
        }

        private float renderTime = 0;   // not yet supported.

        /// <summary>
        /// Average time to render a frame (seconds)
        /// </summary>
        public float RenderTime
        {
            get { return renderTime; }
        }
        int totalRenderingTicks = 0;
        int renderedFrames;

        private float updateTime = 0;

        /// <summary>
        /// Average time to update objects per frame (seconds)
        /// </summary>
        public float UpdateTime
        {
            get { return updateTime; }
        }

        Vector3 cameraPosition = new Vector3(0, 1f, 10f);

        /// <summary>
        /// Location of camera
        /// </summary>
        public Vector3 CameraPosition
        {
            get { return cameraPosition; }
            set { cameraPosition = value; }
        }

        Vector3 cameraTargetLocation = new Vector3(0, 1f, 0);

        /// <summary>
        /// Direction of camera
        /// </summary>
        public Vector3 CameraTargetLocation
        {
            get { return cameraTargetLocation; }
            set { cameraTargetLocation = value; }
        }

        Vector3 cameraUp = Vector3.Up;

        /// <summary>
        /// Up direction of camera, to determine tilt;
        /// </summary>
        public Vector3 CameraUp
        {
            get { return cameraUp; }
            set { cameraUp = value; }
        }

        float cameraFOV = MathHelper.ToRadians(45);

        /// <summary>
        /// Camera's field of view
        /// </summary>
        public float CameraFOV
        {
            get { return cameraFOV; }
            set { cameraFOV = value; }
        }
        #endregion

        /// <summary>
        /// Constructor for TwigGame.
        /// Creates graphics and content managers.
        /// </summary>
        public TwigGame()
            : base()
        {
            Debug.debugGame = this;
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
            this.IsMouseVisible = true;
            BindKey(Keys.Space, "Remove last dialog item", delegate(Keys ignore)
            {
                DialogBalloonManager.RemoveLastDialogBalloon();
                Pause = false;
            });
            BindKey(Keys.Pause, "Toggle automatic pausing on dialog", delegate(Keys ignore)
            {
                DialogBalloonManager.PauseOnDialog = !DialogBalloonManager.PauseOnDialog;
            });
            dialogBalloonManager = new DialogBalloon.DialogBalloonManager(this);
        }

        #region Initialization
        /// <summary>
        /// Initializes Twig engine, including creation of device and effects, then calls InitializeTwigObjects.
        /// </summary>
        protected sealed override void Initialize()
        {
            SetUpXNADevice();
            InitializeEffects();
#if DEBUG
            debugSphere = SimpleMesh.Sphere(device, 0.1f, 2, 0f);
            debugLineVertexDeclaration = new VertexDeclaration(device, VertexPositionColor.VertexElements);
#endif
            InitializeTwigObjects();
            base.Initialize();
        }

        /// <summary>
        /// Create in-game Twig objects and perform any other desired initialization.  Called automatically after TwigGame is initialized.
        /// </summary>
        protected abstract void InitializeTwigObjects();

        /// <summary>
        /// Configures the GraphicsDevice
        /// </summary>
        protected void SetUpXNADevice()
        {
            device = graphics.GraphicsDevice;

            graphics.PreferredBackBufferWidth = 800;
            graphics.PreferredBackBufferHeight = 300;
            graphics.IsFullScreen = false;
            graphics.PreferMultiSampling = true;
            graphics.ApplyChanges();
            //device.RenderState.CullMode = CullMode.None;
            Window.Title = "Twig";
            basicEffect = new BasicEffect(device, null);

#if Windows
            Mouse.WindowHandle = Window.Handle;
#endif
        }

        /// <summary>
        /// Creates and configures Effects objects for rendering
        /// </summary>
        protected virtual void InitializeEffects()
        {
            basicEffect = new BasicEffect(graphics.GraphicsDevice, null);
            basicEffect.DiffuseColor = new Vector3(1.0f, 1.0f, 1.0f);
            basicEffect.SpecularColor = Color.Black.ToVector3();
            basicEffect.PreferPerPixelLighting = true;

            basicEffect.LightingEnabled = true;
            basicEffect.EnableDefaultLighting();

            basicEffect.View = viewMatrix;
            basicEffect.Projection = projectionMatrix;
            basicEffect.World = Matrix.Identity;

            HUDEffect = new BasicEffect(device, null);
            //HUDEffect.VertexColorEnabled = true;
            HUDEffect.Projection = Matrix.CreateOrthographicOffCenter(
                0, device.Viewport.Width, device.Viewport.Height, 0, 0, 1);
            HUDEffect.View = HUDEffect.World = Matrix.Identity;
        }
        #endregion

        /// <summary>
        /// Creates and initializes View and Projection matrices
        /// </summary>
        protected void UpdateViewProjection()
        {
            basicEffect.View = viewMatrix = Matrix.CreateLookAt(
                CameraPosition,
                CameraTargetLocation,
                CameraUp
                );

            basicEffect.Projection = projectionMatrix = Matrix.CreatePerspectiveFieldOfView(
                CameraFOV,
                (float)graphics.GraphicsDevice.Viewport.Width /
                (float)graphics.GraphicsDevice.Viewport.Height,
                1.0f, 100.0f);
        }

        #region Standard methods
        /// <summary>
        /// Loads media assets required by Twig engine.
        /// </summary>
        protected override void LoadContent()
        {
            HUDFont = Content.Load<SpriteFont>("HUD Font");
            HUDSprites = new SpriteBatch(device);
            DialogFont = Content.Load<SpriteFont>("DialogFont");
            DialogSprites = new SpriteBatch(device);
            base.LoadContent();
        }

        /// <summary>
        /// Calls Update methods on GameComponents and resolves position constraints among Nodes.
        /// </summary>
        /// <param name="gameTime"></param>
        protected override void Update(GameTime gameTime)
        {
            int start = Environment.TickCount;
            HandleKeyboardEvents();
#if DEBUG
            if (SingleStep || !Pause)
            {
                for (int i = 0; i < TwigObject.AllTwigObjects.Count; i++)
                    TwigObject.AllTwigObjects[i].DebugOutput.Clear();
            }
#endif

            base.Update(gameTime);
            if (SingleStep || !Pause)
            {
                Physics.Update(gameTime);
                updateTime = (99 * updateTime + (Environment.TickCount - start) * 0.001f) * 0.01f;
            }
            Pause |= SingleStep;
            SingleStep = false;
#if NodeLogging
            Node.LogNodes();
#endif
        }

        /// <summary>
        /// Draws game components and HUD.
        /// </summary>
        /// <param name="gameTime"></param>
        protected override void Draw(GameTime gameTime)
        {
            int start = Environment.TickCount;
            //graphics.GraphicsDevice.Clear(ClearOptions.DepthBuffer | ClearOptions.Target, Color.Blue, 1.0f, 0);
            graphics.GraphicsDevice.Clear(new Color(0,0, 190));
            device.RenderState.DepthBufferEnable = true;   // This is getting turned off somehow, probably by the SpriteBatch code
            UpdateViewProjection();
#if DEBUG
            DrawDebugMarkers();
            DrawDebugLines();
#endif
            base.Draw(gameTime);
            DialogBalloonManager.DrawBalloons(DialogSprites, DialogFont, HUDEffect);
            renderedFrames++;
            totalRenderingTicks += (Environment.TickCount - start);
            renderTime = totalRenderingTicks * 0.001f / renderedFrames;
#if NodeLogging
            Node.AdvanceLog();
#endif
        }

        /// <summary>
        /// Display dialog string for character.
        /// </summary>
        public void DrawDialog(Vector2 position, string blablabla)
        {
            DialogSprites.DrawString(DialogFont, blablabla, position, Color.White);
        }
        #endregion

        #region Keyboard handling
        KeyboardState previousKeyboardState = Keyboard.GetState();
        KeyboardState currentKeyboardState = Keyboard.GetState();

        private KeyModifiers keyModifiers;

        /// <summary>
        /// Keyboard modifiers currently in effect
        /// </summary>
        public KeyModifiers KeyModifiers
        {
            get { return keyModifiers; }
        }

        void HandleKeyboardEvents()
        {
            currentKeyboardState = Keyboard.GetState();

            keyModifiers = KeyModifiers.None;
            if (currentKeyboardState.IsKeyDown(Keys.LeftShift) || currentKeyboardState.IsKeyDown(Keys.RightShift))
                keyModifiers |= KeyModifiers.Shift;
            if (currentKeyboardState.IsKeyDown(Keys.LeftControl) || currentKeyboardState.IsKeyDown(Keys.RightControl))
                keyModifiers |= KeyModifiers.Control;
            if (currentKeyboardState.IsKeyDown(Keys.LeftAlt) || currentKeyboardState.IsKeyDown(Keys.RightAlt))
                keyModifiers |= KeyModifiers.Alt;

            for (int i = 0; i < keyBindings.Count; i++)
            {
                KeyBinding b = keyBindings[i];
                if (b.handler == null)
                    continue;

                if (currentKeyboardState.IsKeyDown(b.key) && previousKeyboardState.IsKeyUp(b.key) && b.modifiers==keyModifiers)
                    b.handler(b.key);
            }
            previousKeyboardState = currentKeyboardState;
        }

        /// <summary>
        /// True if key is down.
        /// </summary>
        public bool IsKeyDown(Keys k)
        {
            return currentKeyboardState.IsKeyDown(k);
        }

        /// <summary>
        /// Handler for a keyboard key.
        /// </summary>
        public delegate void KeyHandler(Keys key);

        /// <summary>
        /// Command key binding
        /// </summary>
        public class KeyBinding
        {
            /// <summary>
            /// Create a new key binding object without adding it to the keybinding list
            /// </summary>
            public KeyBinding(KeyModifiers m, Keys k, string doc, KeyHandler h)
            {
                modifiers = m;
                key = k;
                documentation = doc;
                handler = h;
            }
            /// <summary>
            /// Key being bound
            /// </summary>
            public Keys key;
            /// <summary>
            /// Modifiers that must be held down to activate this keyboard command
            /// </summary>
            public KeyModifiers modifiers;
            /// <summary>
            /// Description for display on the screen.
            /// </summary>
            public string documentation;
            /// <summary>
            /// Closure for handling the key.
            /// </summary>
            public KeyHandler handler;
        }

        List<KeyBinding> keyBindings = new List<KeyBinding>();

        /// <summary>
        /// List of all key bindings in effect.
        /// </summary>
        public List<KeyBinding> KeyBindings
        {
            get { return keyBindings; }
        }

        /// <summary>
        /// Binds a key to a handler.
        /// </summary>
        public void BindKey(Keys k, string documentation, KeyHandler handler)
        {
            BindKey(KeyModifiers.None, k, documentation, handler);
        }
            
        /// <summary>
        /// Binds a key to a handler.
        /// </summary>
        public void BindKey(KeyModifiers m, Keys k, string documentation, KeyHandler handler)
        {
            if (k != Keys.None)
            {
                foreach (KeyBinding b in KeyBindings)
                    if (k == b.key && m == b.modifiers)
                    {
                        throw new Exception("binding previously bound key");
                    }

                KeyBindings.Add(new KeyBinding(m, k, documentation, handler));
                KeyBindings.Sort(delegate(KeyBinding a, KeyBinding b)
                {
                    if (a.key < b.key)
                        return -1;
                    else if (a.key > b.key)
                        return 1;
                    else
                        return 0;
                });
            }
        }

        /// <summary>
        /// Provide documentation on key function to appear in the help display, without providing a handle.
        /// </summary>
        /// <param name="k"></param>
        /// <param name="documentation"></param>
        public void DocumentKey(Keys k, string documentation)
        {
            BindKey(k, documentation, null);
        }
        #endregion

        #region HUD code
        //
        // Utility for draing lines
        //
        VertexPositionColor[] linebuffer = new VertexPositionColor[2];
        /// <summary>
        /// Draws a single line on the screen.
        /// NB: assumes it's being called inside of DrawHUD()
        /// </summary>
        public void DrawLine(float xs, float ys, float xe, float ye, Color color)
        {
            Vector3 targetColor = color.ToVector3();
            if (HUDEffect.DiffuseColor != targetColor)
            {
                HUDEffect.DiffuseColor = targetColor;
                HUDEffect.CommitChanges();
            }
            linebuffer[0] = new VertexPositionColor(new Vector3(xs, ys, 0), color);
            linebuffer[1] = new VertexPositionColor(new Vector3(xe, ye, 0), color);
            device.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.LineList, linebuffer, 0, 1);
        }

        ChartRecorder chartRecorder;

        /// <summary>
        /// The ChartRecorder that Chart() calls should be logged to, if any.
        /// </summary>
        public ChartRecorder ChartRecorder
        {
            get { return chartRecorder; }
            set { chartRecorder = value; }
        }

        /// <summary>
        /// Should only be called from TwigObject.Chart
        /// </summary>
        public void Chart(int section, string label, float v, float maxvalue)
        {
            if (ChartRecorder != null)
                ChartRecorder.Chart(section, label, v, maxvalue);
        }

        static TwigObject m_debugObject;

        /// <summary>
        /// Object to display debugging information for in the HUD
        /// </summary>
        public TwigObject DebugObject
        {
            get
            {
                return m_debugObject;
            }
            set
            {
                m_debugObject = value;
            }
        }
        #endregion HUD code

#if DEBUG
        SimpleMesh debugSphere;
        struct DebugMarker
        {
            public DebugMarker(Vector3 position, Color color)
            {
                this.position = position;
                this.color = color;
            }
            public Vector3 position;
            public Color color;
        }

        List<DebugMarker> debugMarkers = new List<DebugMarker>();

        void DrawDebugMarkers()
        {
            if (debugMarkers.Count > 0)
            {
                BasicEffect.Begin();
                foreach (EffectPass pass in BasicEffect.CurrentTechnique.Passes)
                {
                    pass.Begin();
                    for (int i = 0; i < debugMarkers.Count; i++)
                    {
                        DebugMarker m = debugMarkers[i];
                        basicEffect.DiffuseColor = m.color.ToVector3();
                        basicEffect.World = Matrix.CreateTranslation(m.position);
                        basicEffect.CommitChanges();
                        debugSphere.Draw();
                    }
                    pass.End();
                }
                BasicEffect.End();
                debugMarkers.Clear();
            }
        }

        const int maxDebugLines = 500;
        VertexPositionColor[] debugLineVertices = new VertexPositionColor[maxDebugLines*2];
        int debugLineCount=0;
        VertexDeclaration debugLineVertexDeclaration;

        void DrawDebugLines() {
            if (debugLineCount == 0)
                return;

            Device.VertexDeclaration = debugLineVertexDeclaration;
            BasicEffect.World = Matrix.Identity;
            BasicEffect.CommitChanges();
            BasicEffect.Begin();
            foreach (EffectPass p in BasicEffect.CurrentTechnique.Passes) {
                p.Begin();
                Device.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.LineList, debugLineVertices, 0, debugLineCount);
                p.End();
            }
            BasicEffect.End();
            debugLineCount = 0;
        }
#endif

        /// <summary>
        /// Draws a sphere in the specified color at the specified location.
        /// </summary>
        [Conditional("DEBUG")]
        public void DrawDebugMarker(Vector3 position, Color c)
        {
#if DEBUG
            debugMarkers.Add(new DebugMarker(position, c));
#endif
        }

        /// <summary>
        /// Draws a sphere in the specified color at the specified location.
        /// </summary>
        [Conditional("DEBUG")]
        public void DrawDebugMarker(Node n, Color c)
        {
            DrawDebugMarker(n.Position, c);
        }

        /// <summary>
        /// Draw a line as a debugging annotation
        /// </summary>
        [Conditional("DEBUG")]
        public void DrawDebugLine(Vector3 start, Vector3 end, Color c) {
#if DEBUG
            int i = debugLineCount*2;
            debugLineVertices[i++] = new VertexPositionColor(start, c);
            debugLineVertices[i] = new VertexPositionColor(end, c);
            debugLineCount++;
#endif
        }

        /// <summary>
        /// Installs the typical debugging HUDs, etc.
        /// </summary>
        public void InstallStandardDebuggingComponents()
        {
            Components.Add(new StandardHUD(this, Keys.H, false));
            Components.Add(new ChartRecorder(this, Keys.R, false));
            Components.Add(new DebugPicker(this, Keys.D, Keys.Tab));
            Components.Add(new HelpDisplay(this, Keys.F1));
            Components.Add(new ConsolePrintf(this));
            PauseControl.InstallKeyHandlers(this, Keys.Enter, Keys.S, Keys.L);
        }

        #region Named components
        /// <summary>
        /// Named objects within the game
        /// </summary>
        public Dictionary<string,GameComponent> NamedObjects = new Dictionary<string,GameComponent>();
        #endregion
    }
}
