// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Diagnostics;

namespace Twig
{
    /// <summary>
    /// Base class for all in-game objects managed by Twig physics.
    /// </summary>
    [Serializable]
    [DebuggerDisplay("{name}")]
    public abstract class TwigObject : DrawableGameComponent, IDrawable, IUpdateable
    {
        #region Constructor
        /// <summary>
        /// Base initializer for all Twig objects
        /// </summary>
        public TwigObject(TwigGame g, string name, bool dynamic)
            : base(g)
        {
            this.name = name;
            tGame = g;
            AllTwigObjects.Add(this);
            SortTwigObjectsListByDynamicness();
            isDynamic = dynamic;
            if (dynamic)
                DynamicTwigObjects.Add(this);
            if (name != null)
                g.NamedObjects[name] = this;
        }
        #endregion

        #region Object tracking
        /// <summary>
        /// All existing Twig objects
        /// </summary>
        public static List<TwigObject> AllTwigObjects = new List<TwigObject>();
        /// <summary>
        /// List of all TwigObjects that are dynamic
        /// </summary>
        public static List<TwigObject> DynamicTwigObjects = new List<TwigObject>();

        static void SortTwigObjectsListByDynamicness()
        {
            AllTwigObjects.Sort(new Comparison<TwigObject>(CompareTwigObjectsForDynamicness));
        }

        static int CompareTwigObjectsForDynamicness(TwigObject o1, TwigObject o2)
        {
            if (o1.isDynamic == o2.isDynamic)
                return 0;
            if (o1.isDynamic)
                return -1;
            return 1;
        }
        #endregion

        #region Globals
        /// <summary>
        /// Setting this to true tells all characters to freeze.  It is checked by FreezeLocomotion.Update().
        /// </summary>
        public static bool AllCharactersFreeze = false;
        #endregion

        #region Position information
        /// <summary>
        /// Position of center of bottom of object.  This will generally lie in the ground plane.
        /// </summary>
        public abstract Vector3 Position
        {
            get;
        }

        /// <summary>
        /// Moves object to new position, without changing its orientation
        /// </summary>
        public virtual void TeleportTo(Vector3 newPosition)
        {
            Vector3 delta = newPosition - Position;
            foreach (Node n in nodes)
            {
                ControlMode save = n.ControlMode;
                n.ControlMode = ControlMode.Free;
                n.Position = n.Position + delta;
                n.Velocity = Vector3.Zero;
                n.ControlMode = save;
            }
        }

        Vector3 previousPosition, velocity;
        /// <summary>
        /// Current velocity vector for the object
        /// </summary>
        public virtual Vector3 Velocity
        {
            get
            {
                return velocity;
            }
        }
        void UpdateVelocity()
        {
            velocity = (Position - previousPosition) * (1 / Physics.deltaT);
            previousPosition = Position;
        }

        /// <summary>
        /// The object's total mass
        /// </summary>
        public abstract float Mass { get; }

        /// <summary>
        /// The object's current center of mass in world coordinates
        /// </summary>
        public abstract Vector3 CenterOfMass { get; }

        /// <summary>
        /// 3D location a creature should fixate when "looking at" this object.
        /// </summary>
        public abstract Vector3 LookPoint { get; }

        /// <summary>
        /// Screen position of the object (as defined by its Position property)
        /// </summary>
        public Point ScreenPosition
        {
            get
            {
                Vector3 pos = Device.Viewport.Project(Position, TwigGame.ProjectionMatrix, TwigGame.ViewMatrix, Matrix.Identity);
                return new Point((int)(pos.X + 0.5f), (int)(pos.Y + 0.5f));
            }
        }
        #endregion

        #region Nodes, links, and collision elements
        /// <summary>
        /// Set of nodes contained within this object
        /// </summary>
        protected Node[] nodes = new Node[] { };

        /// <summary>
        /// Add new nodes to the nodes list of this object
        /// </summary>
        protected void AddNodes(params Node[] newNodes)
        {
            Node[] updatedNodeList = new Node[nodes.Length + newNodes.Length];
            nodes.CopyTo(updatedNodeList, 0);
            newNodes.CopyTo(updatedNodeList, nodes.Length);
            nodes = updatedNodeList;
        }

        /// <summary>
        /// Set of nodes contained within this object
        /// </summary>
        public Node[] Nodes
        {
            get { return nodes; }
        }

        /// <summary>
        /// Elements of this object to be checked for collisions
        /// </summary>
        protected CollisionElement[] collisionElements;

        /// <summary>
        /// Add new nodes to the CollisionElements list of this object
        /// </summary>
        protected void AddCollisionElements(params CollisionElement[] newCEs)
        {
            CollisionElement[] updatedCEList = new CollisionElement[collisionElements.Length + newCEs.Length];
            collisionElements.CopyTo(updatedCEList, 0);
            newCEs.CopyTo(updatedCEList, collisionElements.Length);
            collisionElements = updatedCEList;
        }

        /// <summary>
        /// Collision elements to be tested when checking object for collisions
        /// </summary>
        public CollisionElement[] CollisionElements
        {
            get { return collisionElements; }
        }
        #endregion

        #region Other instance variables and properties
        /// <summary>
        /// Name string (for debug purposes only)
        /// </summary>
        private string name;

        /// <summary>
        /// String naming the object.
        /// </summary>
        public string Name
        {
            get
            {
                if (name == null)
                    return GetType().Name;
                else
                    return name;
            }
            set { name = value; }
        }

        Color mColor = Color.Red;
        /// <summary>
        /// Principle color in which render the object.
        /// </summary>
        public Color Color
        {
            get { return mColor; }
            set { mColor = value; }
        }

        [NonSerialized]
        TwigGame tGame;
        /// <summary>
        /// Game to which this object belongs
        /// </summary>
        public TwigGame TwigGame {
            get
            {
                return tGame;
            }
        }

        /// <summary>
        /// Gets the DialogBalloonManager for the TwigGame
        /// </summary>
        public DialogBalloon.DialogBalloonManager DialogBalloonManager
        {
            get
            {
                return TwigGame.DialogBalloonManager;
            }
        }

        bool IUpdateable.Enabled
        {
            get
            {
                return TwigGame.SingleStep || !TwigGame.Pause;
            }
        }

        bool isDynamic;

        /// <summary>
        /// True if the object's pose or configuration changes.  False if the object is asleep.
        /// </summary>
        public bool Dynamic
        {
            get { return isDynamic; }
            set
            {
                if (isDynamic != value)
                {
                    if (value)
                        DynamicTwigObjects.Add(this);
                    else
                        DynamicTwigObjects.Remove(this);
                    isDynamic = value;
                }
                SortTwigObjectsListByDynamicness();
            }
        }

        bool cloaked;

        /// <summary>
        /// If true, other characters will ignore this object, as if it didn't exist.
        /// </summary>
        public bool Cloaked
        {
            get { return cloaked; }
            set { cloaked = value; }
        }

        /// <summary>
        /// True if this object should be manuvered around when in the way.
        /// False means the character will walk over it.
        /// </summary>
        public virtual bool IsObstacle
        {
            get
            {
                return !Cloaked;
            }
        }
        #endregion

        #region Update
        /// <summary>
        /// Used for tagging objects with thunks to run when updating.
        /// </summary>
        public delegate void UpdateThunk();

        /// <summary>
        /// Called when object updates.
        /// </summary>
        public event UpdateThunk UpdateThunks;

        /// <summary>
        /// Run update thunks, if any.
        /// </summary>
        /// <param name="gameTime"></param>
        public override void Update(GameTime gameTime)
        {
            if (UpdateThunks!=null)
                UpdateThunks();
            base.Update(gameTime);
        }
        #endregion

        #region Constraint
        /// <summary>
        /// Enforce constraints on object's geometry.
        /// </summary>
        public abstract void Constrain();

        /// <summary>
        /// Resolve geometric constraints on all Twig objects.
        /// </summary>
        public static void ConstrainAllTwigObjects()
        {
            foreach (TwigObject o in AllTwigObjects)
            {
                o.Constrain();
            }
        }
        #endregion

        #region Graphics
        /// <summary>
        /// The GraphicsDevice of the game corresponding to this TwigObject
        /// </summary>
        public GraphicsDevice Device
        {
            get
            {
                return TwigGame.Device;
            }
        }

        /// <summary>
        /// The BasicEffect to use in rendering this object
        /// </summary>
        public BasicEffect BasicEffect
        {
            get
            {
                return TwigGame.BasicEffect;
            }
        }

        /// <summary>
        /// Configures BasicEffect for the Color of this object, but does not CommitChanges().
        /// </summary>
        public virtual void ConfigureMaterial()
        {
            BasicEffect.DiffuseColor = Color.ToVector3()*(1-TwigGame.FillLevel);
            BasicEffect.AmbientLightColor = Color.ToVector3()*TwigGame.FillLevel;
        }

        bool IDrawable.Visible
        {
            get
            {
                return Vector3.Transform(Position,TwigGame.ViewMatrix).Z<0;
            }
        }
        #endregion

        /// <summary>
        /// Check subject's gaze
        /// </summary>
        /// <param name="subject">TwigObject whose gaze to check</param>
        /// <param name="obj">TwigObject that might be being gazed at</param>
        /// <returns>True if subject is fixating object</returns>
        public static bool LookingAt(TwigObject subject, TwigObject obj)
        {
            if (subject == null)
                return false;
            else
                return subject.LookingAt(obj);
        }

        /// <summary>
        /// Check object's gaze.
        /// </summary>
        /// <param name="o">TwigObject that might be being gazed at</param>
        /// <returns>True if fixating o</returns>
        public virtual bool LookingAt(TwigObject o)
        {
            return false;  // Default is TwigObjects are inanimate and so don't look at anything.
        }

        /// <summary>
        /// Check object's destination
        /// </summary>
        /// <param name="target">Possible destination</param>
        /// <returns>True if object is attempting to approach target</returns>
        public virtual bool IsApproachTarget(TwigObject target)
        {
            return false;  // Default is TwigObjects are inanimate and so don't walk.
        }

        /// <summary>
        /// Radius of bounding cylinder (vertical, infinitely high) for approximate collision testing.
        /// </summary>
        public abstract float CollisionRadius
        {
            get;
        }

        /// <summary>
        /// True if this object is the supporting surface of n
        /// </summary>
        public virtual bool IsSupporting(Vector3 point)
        {
            return false;
        }

        /// <summary>
        /// True if this object is the supporting surface of n
        /// </summary>
        public virtual bool IsSupporting(Node n)
        {
            return n.SupportingSurface==this;
        }

        /// <summary>
        /// True if this object is supporting one of the nodes of o;
        /// </summary>
        public virtual bool IsSupporting(TwigObject o)
        {
            return false;
        }

        /// <summary>
        /// Default implementation of IsSupporting for supporting surfaces.
        /// </summary>
        protected bool SupportingSurfaceDefaultIsSupportingMethod(TwigObject o)
        {
            foreach (Node n in o.Nodes)
                if (n.SupportingSurface==this)
                    return true;
            return false;
        }

        /// <summary>
        /// Handler for when a node of the object moves over/off-of a suppporting surface.
        /// </summary>
        public delegate void MountEventHandler(Node n, TwigObject supportingSurface);

        /// <summary>
        /// Called when a new object becomes the supporting surface of one of this object's nodes.
        /// </summary>
        public event MountEventHandler Mount;
        /// <summary>
        /// Called when an object ceases to be the supporting surface of one of this object's nodes.
        /// </summary>
        public event MountEventHandler Dismount;

        /// <summary>
        /// Called whenever a node of the object mounts the supporting surface.
        /// </summary>
        public virtual void OnMount(Node n, TwigObject newSupportingSurface)
        {
            if (Mount != null)
                Mount(n, newSupportingSurface);
        }

        /// <summary>
        /// Called when a node of the object dismounts a supporting surface.
        /// Likely to be called many times when the object moves 
        /// </summary>
        public virtual void OnDismount(Node n, TwigObject oldSupportingSurface)
        {
            if (Dismount != null)
                Dismount(n, oldSupportingSurface);
        }



        /// <summary>
        /// Returns the height at which this object supports other objects, if this object functions as a supporting surface.
        /// </summary>
        public virtual float SupportingSurfaceLevel
        {
            get
            {
                return 0;
            }
        }

        /// <summary>
        /// True if this object can support others as a flat, horizontal surface.
        /// </summary>
        public virtual bool IsASupportingSurface
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// NodeFrame to use for rigidly attaching nodes to this supporting surface.
        /// </summary>
        public virtual NodeFrame NodeFrame
        {
            get
            {
                return null;
            }
        }

        /// <summary>
        /// The node a character should sit on, if any, when sitting on this object.
        /// </summary>
        public virtual Node SittingTarget
        {
            get { return null; }
        }

        /// <summary>
        /// Size of the sittable area of the object, centered on the SittingTarget node.
        /// </summary>
        public virtual float SittingTargetRadius
        {
            get { return 0; }
        }

        /// <summary>
        /// True if it's possible to sit on this object
        /// </summary>
        public virtual bool IsSittable {
            get { return SittingTarget != null; }
        }

        /// <summary>
        /// Direction from which to approach the object when trying to dock with it (e.g. sit on it).
        /// Zero means arbitrary direction.
        /// </summary>
        public virtual Vector3 DockingDirection
        {
            get
            {
                return Vector3.Zero;
            }
        }

        /// <summary>
        /// Distance to close to when docking.
        /// </summary>
        public virtual float DockingDistance
        {
            get
            {
                return 0.5f;
            }
        }

        /// <summary>
        /// Location to approach when docking with object
        /// </summary>
        public Vector3 DockingTarget
        {
            get
            {
                return Position + DockingDistance * DockingDirection;
            }
        }

        /// <summary>
        /// Provides a set of named, local coordinate systems for the object.
        /// To the extent this is reminiscent of the charts-and-atlas structure
        /// of differentiable manifolds, Atlas seems like a good name.
        /// However, it's something of an abuse of terminology since the surface of
        /// an object in 3-space, would actually be a 2-dimentional manifold
        /// in the mathematical sense, and so coordinates in the surface would be
        /// represented as 2-vectors.  However, it's convenient for us to use
        /// 3-vectors in which Y=0 means a point lying on the surface, and Y>0
        /// meaning a point above the surface.
        /// </summary>
        public readonly Dictionary<string, Chart> Atlas = new Dictionary<string, Chart>();

        #region Appraisal
        [NonSerialized]
        Mammal.ObjectAppraisal appraisalCache;
        [NonSerialized]
        int appraisalCacheTimestamp = -1;
        [NonSerialized]
        Mammal appraisalSource;

        /// <summary>
        /// Determine appraisal by appraiser
        /// </summary>
        /// <param name="appraiser">Mammal performing the appraisal</param>
        /// <returns>Appraisal structure</returns>
        public Mammal.ObjectAppraisal AppraisalFor(Mammal appraiser)
        {
            if (appraisalCacheTimestamp != Physics.tickCount || appraiser != appraisalSource)
            {
                appraisalCache = appraiser.Appraisal(this);
                appraisalCacheTimestamp = Physics.tickCount;
                appraisalSource = appraiser;
            }
            return appraisalCache;
        }
        #endregion

        #region Debugging aids
        /// <summary>
        /// Which section of the HUD chart this object's vital stats should appear on.
        /// Default is -1, meaning the object isn't included on the chart.
        /// </summary>
        [NonSerialized]
        public int chartSection = -1;

        /// <summary>
        /// Update this TwigObject's chart.
        /// </summary>
        /// <param name="label">Which channel to update</param>
        /// <param name="value">Current value</param>
        /// <param name="maxValue">Maximum possible value for scaling purposes</param>
        [Conditional("DEBUG")]
        public void Chart(string label, float value, float maxValue)
        {
            if (chartSection >= 0)
                TwigGame.Chart(chartSection, label, value, maxValue);
        }

        /// <summary>
        /// Useful information for optional display in the HUD
        /// </summary>
        public virtual string DebugString
        {
            get
            {
                return String.Format("{0}", (name == null) ? GetType().Name : name);
            }
        }

        List<string> debugOutput = new List<string>();

        /// <summary>
        /// List of debug strings to be displayed in the ConsolePrintf HUD
        /// </summary>
        public List<string> DebugOutput
        {
          get { return debugOutput; }
        }

        /// <summary>
        /// Add a line to the ConsolePrintf HUD
        /// </summary>
        [Conditional("DEBUG")]
        public void Printf(string s) {
            debugOutput.Add(s);
        }

        /// <summary>
        /// Add a line to the ConsolePrintf HUD
        /// </summary>
        [Conditional("DEBUG")]
        public void Printf(string s, object o1)
        {
            debugOutput.Add(String.Format(s, o1));
        }

        /// <summary>
        /// Add a line to the ConsolePrintf HUD
        /// </summary>
        [Conditional("DEBUG")]
        public void Printf(string s, object o1, object o2)
        {
            debugOutput.Add(String.Format(s, o1, o2));
        }

        /// <summary>
        /// Add a line to the ConsolePrintf HUD
        /// </summary>
        [Conditional("DEBUG")]
        public void Printf(string s, object o1, object o2, object o3)
        {
            debugOutput.Add(String.Format(s, o1, o2, o3));
        }

        /// <summary>
        /// Add a line to the ConsolePrintf HUD
        /// </summary>
        [Conditional("DEBUG")]
        public void Printf(string s, object o1, object o2, object o3, object o4)
        {
            debugOutput.Add(String.Format(s, o1, o2, o3, o4));
        }

        /// <summary>
        /// Add a line to the ConsolePrintf HUD
        /// </summary>
        [Conditional("DEBUG")]
        public void Printf(string s, object o1, object o2, object o3, object o4, object o5)
        {
            debugOutput.Add(String.Format(s, o1, o2, o3, o4, o5));
        }


        /// <summary>
        /// Assert that a new reading has taken place for the named value.  If its variation is over threshold, signal an error.
        /// </summary>
        /// <param name="name">Name of the signal being measured</param>
        /// <param name="newReading">New reading for this tick</param>
        /// <param name="threshold">threshold for time-averaged percentage inter-tick change, beyond which an error should be thrown</param>
        /// <param name="filterCoefficient">Weighting to be given to new tick's reading in computing time average</param>
        [Conditional("DEBUG")]
        public void CheckChatter(string name, float newReading, float threshold, float filterCoefficient)
        {
#if DEBUG
            FloatChatterDetector.Update(floatChatterDetectors, name, newReading, threshold, filterCoefficient);
#endif
        }

        /// <summary>
        /// Assert that a new reading has taken place for the named value.  If its variation is over threshold, signal an error.
        /// </summary>
        /// <param name="name">Name of the signal being measured</param>
        /// <param name="newReading">New reading for this tick</param>
        /// <param name="threshold">threshold for time-averaged percentage inter-tick change, beyond which an error should be thrown</param>
        /// <param name="filterCoefficient">Weighting to be given to new tick's reading in computing time average</param>
        [Conditional("DEBUG")]
        public void CheckChatter(string name, Vector3 newReading, float threshold, float filterCoefficient)
        {
#if DEBUG
            VectorChatterDetector.Update(vectorChatterDetectors, name, newReading, threshold, filterCoefficient);
#endif
        }

#if DEBUG
        Dictionary<string, FloatChatterDetector> floatChatterDetectors = new Dictionary<string, FloatChatterDetector>();
        Dictionary<string, VectorChatterDetector> vectorChatterDetectors = new Dictionary<string, VectorChatterDetector>();

        class FloatChatterDetector
        {
            float deltaAverage;
            float previousReading;
            float threshold;
            float filterCoefficient;
            string name;

            static public void Update(Dictionary<string, FloatChatterDetector> detectors, string name, float newReading, float threshold, float filterCoefficient)
            {
                FloatChatterDetector d;
                if (!detectors.TryGetValue(name, out d))
                {
                    d = new FloatChatterDetector(name, newReading, threshold, filterCoefficient);
                    detectors[name] = d;
                }
                d.Update(newReading);
            }

            FloatChatterDetector(string name, float firstReading, float threshold, float filterCoefficient)
            {
                this.name = name;
                this.previousReading = firstReading;
                this.threshold = threshold;
                this.filterCoefficient = filterCoefficient;
            }

            void Update(float newReading)
            {
                if (!Physics.IsSettled || newReading == 0)
                    return;

                deltaAverage = filterCoefficient * Math.Min(1f, Math.Abs((newReading - previousReading) / Math.Max(Math.Abs(newReading), Math.Abs(previousReading)))) + (1 - filterCoefficient) * deltaAverage;
                Debug.Assert(deltaAverage < threshold, "Chattering detected in " + name);
                previousReading = newReading;
            }
        }

        class VectorChatterDetector
        {
            float deltaAverage;
            Vector3 previousReading;
            float threshold;
            float filterCoefficient;
            string name;

            static public void Update(Dictionary<string, VectorChatterDetector> detectors, string name, Vector3 newReading, float threshold, float filterCoefficient)
            {
                VectorChatterDetector d;
                if (!detectors.TryGetValue(name, out d))
                {
                    d = new VectorChatterDetector(name, newReading, threshold, filterCoefficient);
                    detectors[name] = d;
                }
                d.Update(newReading);
            }

            VectorChatterDetector(string name, Vector3 firstReading, float threshold, float filterCoefficient)
            {
                this.name = name;
                this.previousReading = firstReading;
                this.threshold = threshold;
                this.filterCoefficient = filterCoefficient;
            }

            void Update(Vector3 newReading)
            {
                if (!Physics.IsSettled || newReading == Vector3.Zero)
                    return;
                deltaAverage = filterCoefficient * Math.Min(1f, (float)Math.Sqrt((newReading - previousReading).LengthSquared() / Math.Max(newReading.LengthSquared(), previousReading.LengthSquared()))) + (1 - filterCoefficient) * deltaAverage;
                Debug.Assert(deltaAverage < threshold, "Chattering detected in " + name);
                previousReading = newReading;
            }
        }
#endif
        #endregion

        #region Holding objects
        /// <summary>
        /// Configure object to be held by user.  Object also initiates grasp.
        /// </summary>
        public virtual void SetupHoldForUse(Mammal reader, Grasp grasp)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Configure object and grasper to grasp the object
        /// </summary>
        public virtual void SetupHold(Mammal reader, Grasp grasp)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Select which hand(s) to direct to which node(s) to grasp this object
        /// </summary>
        public virtual void SelectGrasp(Mammal grasper, Grasp graspRequest,
                                        out Grasp selectedGrasp, out Node leftTarget, out Node rightTarget)
        {
            selectedGrasp = Grasp.None;
            leftTarget = rightTarget = null;
        }

        Limb graspingLimb;

        /// <summary>
        /// Limb grasping this object, if any.
        /// </summary>
        public virtual Limb GraspingLimb
        {
            get { return graspingLimb; }
            set { graspingLimb = value; }
        }

        /// <summary>
        /// Whether the object is being grasped
        /// </summary>
        public bool IsGrasped
        {
            get
            {
                return GraspingLimb != null;
            }
        }

        /// <summary>
        /// The object (generally a mammal) grasping this object, if any
        /// </summary>
        public TwigObject Grasper
        {
            get
            {
                if (graspingLimb == null)
                    return null;
                else
                    return graspingLimb.TwigObject;
            }
        }

        /// <summary>
        /// True if this object is being manipulated by the holder, and so the holder should focus on it.
        /// </summary>
        public virtual bool IsBeingManipulated
        {
            get
            {
                return false;
            }
        }
        #endregion
    }
}