// Copyright 2007, 2008, 2009 Ian Horswill
// This file is part of Twig.
//
// Twig is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
// Twig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Twig.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Twig
{
    /// <summary>
    /// A TwigObject that takes its appearance from an XNA Model object.
    /// </summary>
    public abstract class TwigObjectWithModel : TwigObject
    {
        Model model;

        /// <summary>
        /// Creates a TwigObject with a specified XNA Model to represent it onscreen.
        /// </summary>
        public TwigObjectWithModel(TwigGame g, string name, Model m, bool dynamic)
            : base(g, name, dynamic)
        {
            this.model = m;
        }

        /// <summary>
        /// Render the object's model.
        /// </summary>
        public override void Draw(GameTime gameTime)
        {
            // Copy any parent transforms.
            Matrix[] transforms = new Matrix[model.Bones.Count];
            model.CopyAbsoluteBoneTransformsTo(transforms);

            // Draw the model. A model can have multiple meshes, so loop.
            foreach (ModelMesh mesh in model.Meshes)
            {
                // This is where the mesh orientation is set, as well as our camera and projection.
                foreach (BasicEffect effect in mesh.Effects)
                {
                    effect.EnableDefaultLighting();
                    effect.World = transforms[mesh.ParentBone.Index] * WorldMatrix;
                    effect.View = TwigGame.ViewMatrix;
                    effect.Projection = TwigGame.ProjectionMatrix;
                }
                // Draw the mesh, using the effects set above.
                mesh.Draw();
            }
            base.Draw(gameTime);
        }

        /// <summary>
        /// The global transformation matrix to be applied to all bones of the object's model.
        /// </summary>
        public abstract Matrix WorldMatrix { get; }
    }
}
